/**
 * Copyright (c) 2008, Freie Universitt Berlin, Tomasz Naumowicz
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:
 * 
 * Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
 * Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.
 * Neither the name of the Freie Universitt Berlin nor the names of its contributors may be used to endorse or promote products derived from this software without specific prior written permission.
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * 
 * Based on the ScatterWeb firmware from the Freie Universitaet Berlin under the BSD license.  
 */




/**
 * @file	ScatterWeb.Sd.c
 * @ingroup	libsd
 * @brief	MMC-/SD-Card library
 * 
 * @author	Michael Baar	<baar@inf.fu-berlin.de>
 * @date	Jan 2007
 * @version	0.2
 * 
 * Initialisation and basic functions for read and write access
 */

#include "ScatterWeb.Sd.internals.h"

volatile sd_state_t	sd_state;

sd_cache_t			_sd_cache;

/******************************************************************************
 * @name	Initialization and configuration
 * @{
 */

void Sd_init() {
	P5SEL |=  0x0E;	 // 00 00 11 10  -> Dout, Din, Clk = peripheral (now done in UART module)
	P5SEL &= ~0x01;  // 00 00 00 01  -> Cs = I/O
	P5OUT |=  0x01;  // 00 00 00 01  -> Cs = High
	P5DIR |=  0x0D;  // 00 00 11 01  -> Dout, Clk, Cs = output
	P5DIR &= ~0x02;  // 00 00 00 10  -> Din = Input
	P2SEL &= ~0x40;  // 11 00 00 00  -> protect, detect = I/O
	P2DIR &= ~0x40;  // 11 00 00 00  -> protect, detect = input	
	
	//Timers_Block(256);
	SD_Card_Init_Result = SD_INIT_NOCARD;	
}

enum sd_init_ret SD_Card_Init() { 
	
	Sd_init();
	
//enum sd_init_ret Sd_init_card(sd_cache_t* pCache) { //maybe the cache should also be hidden from the user, in the end there is no reason to give him access to it
	SD_Card_Init_Result = SD_INIT_NOCARD;
	enum sd_init_ret ret = SD_INIT_SUCCESS;
	sd_csd_t		csd;
	uint16_t			ccc = 0;
	sd_response_r3_t r3;

	memset((void*) &sd_state, 0x00, sizeof(sd_state));

	if( !Sd_detected() ) {
		SD_Card_Init_Result = SD_INIT_NOCARD;
		return SD_INIT_NOCARD;
	}

	/*
	if( Sd_protected() ) {
		SD_Card_Init_Result = SD_INIT_WRITEPROTECTED;
		return SD_INIT_WRITEPROTECTED;
	}
	*/

	// Wait for UART and switch to SPI mode
	if( !USART_LockWait(USART_MODE_SPI) ) { 
		SD_Card_Init_Result = SD_INIT_FAILED;
		return SD_INIT_FAILED;
	}

	// reset card
	if( !_Sd_reset() ) {
		ret = SD_INIT_FAILED;
		goto Sd_init_card_fail;
	}
	
	// Test for hardware compatibility
	if( !_Sd_send_cmd(SD_CMD_READ_OCR, SD_RESPONSE_TYPE_R3, NULL, &r3) ) {
		ret = SD_INIT_FAILED;
		goto Sd_init_card_fail;
	}
	if( (r3.ocr & SD_V_MASK) != SD_V_MASK ) {
		ret = SD_INIT_NOTSUPP;
		goto Sd_init_card_fail;
	}
	
	// Test for software compatibility
	if( !_Sd_read_register( &csd, SD_CMD_SEND_CSD, sizeof(sd_csd_t) ) ) {
		ret = SD_INIT_FAILED;
		goto Sd_init_card_fail;
	}
	ccc = SD_CSD_CCC(csd);
	if( (ccc & SD_DEFAULT_MINCCC) != SD_DEFAULT_MINCCC ) {
		ret = SD_INIT_NOTSUPP;
		goto Sd_init_card_fail;
	}
  
Sd_init_card_fail:
	USART_Unlock(USART_MODE_SPI); 
	if( ret != SD_INIT_SUCCESS ) {
		//if (ret == SD_INIT_NOTSUPP) {
//			printf("[X] SD INIT FAIL: %X\r\n", ccc);
		//}
		SD_Card_Init_Result = ret;
		return ret;
	}

	// state
	sd_state.MinBlockLen_bit = 9;
	sd_state.MaxBlockLen_bit = SD_CSD_READ_BL_LEN(csd);
	sd_state.Flags = 0;
	if( SD_CSD_READ_PARTIAL(csd) ) {
		sd_state.MinBlockLen_bit = 0;
		sd_state.Flags |= SD_READ_PARTIAL;
	}
	if( SD_CSD_WRITE_PARTIAL(csd) )
		sd_state.Flags |= SD_WRITE_PARTIAL;
	sd_state.BlockLen_bit = 9;
	sd_state.BlockLen = 1 << 9;

	#if SD_CACHE
		sd_state.Cache = &_sd_cache;
		_Sd_cache_init();
	#endif

	SD_Card_Init_Result = ret;
	return ret;
}

void Sd_flush() {
	if( USART_Lock(USART_MODE_SPI) ) {
		#if SD_WRITE && SD_CACHE
			Sd_cache_flush();
		#endif
		USART_Unlock(USART_MODE_SPI);
	}
}

void Sd_close() {
	Sd_flush();
}

uint8_t Sd_set_blocklength(const uint8_t blocklength_bit) {
	uint8_t	ret;
	uint8_t	arg[4];
	
	// test if already set
	if( blocklength_bit == sd_state.BlockLen_bit )
		return sd_state.BlockLen_bit;
	
	// Wait for UART and switch to SPI mode
	if( !USART_LockWait(USART_MODE_SPI) )
		return sd_state.BlockLen_bit;
	
	((uint16_t*)arg)[1] = 0;
	((uint16_t*)arg)[0] = 1 << blocklength_bit;
	
	// set blocklength command
	if( _Sd_send_cmd( SD_CMD_SET_BLOCKLENGTH, SD_RESPONSE_TYPE_R1, arg, NULL) ) {
		sd_state.BlockLen_bit = blocklength_bit;
		sd_state.BlockLen = ((uint16_t*)arg)[0];
		ret = blocklength_bit;
	} else {
		ret = SD_BLOCKLENGTH_INVALID;
	}
	
	// unlock uart
	USART_Unlock(USART_MODE_SPI);
	return ret;
}
//@}

///////////////////////////////////////////////////////////////////////////////
// Public functions, Reading
///////////////////////////////////////////////////////////////////////////////
uint16_t Sd_AlignAddress(uint32_t* pAddress) {
	uint16_t	blMask = sd_state.BlockLen - 1;
	uint16_t*	lw = (uint16_t*)pAddress;
	uint16_t	offset = *lw & blMask;
	*lw &= ~blMask;
	return offset;
}

uint16_t Sd_read_block(void (*const pBuffer), const uint32_t address) {

	if( !_Sd_read_start(SD_CMD_READ_SINGLE_BLOCK, address) )
		return false;
		
	SPI_read(pBuffer, sd_state.BlockLen, true);
	// receive CRC16 and finish
	_Sd_read_stop(2);	

	return sd_state.BlockLen;
}

#if SD_READ_BYTE
	bool Sd_read_byte(void* pBuffer, const uint32_t address) {
		
		if( Sd_set_blocklength(0) == 0 ) {
			return Sd_read_block(pBuffer, address);
		} else {
			uint32_t	blAdr = address;
			uint16_t	offset;			// bytes from aligned address to start of first byte to keep

			// align
			offset = Sd_AlignAddress(&blAdr);

			// start
			if( !_Sd_read_start(SD_CMD_READ_SINGLE_BLOCK, address) )
				return false;

			// read
			Spi_read(pBuffer, offset + 1, false);

			// done
			_Sd_read_stop( sd_state.BlockLen - offset - 1 );
		}
		return true;
	}
#endif

///////////////////////////////////////////////////////////////////////////////
// Public functions, Writing
///////////////////////////////////////////////////////////////////////////////
#if SD_WRITE
	enum sd_write_ret _Sd_write_finish() {
		uint16_t	r2;
		uint8_t	ret;
		enum sd_write_ret	result = SD_WRITE_STORE_ERR;
		uint16_t i;

		#if SPI_DMA_WRITE
			spi_dma_wait();
			spi_dma_lock = false;
		#endif
	
		// dummy crc
		SPI_idle(2);
		
		// receive data response (ZZS___ 3 bits crc response)
		for( i = 0 ; i < SD_TIMEOUT_NCR; i++ ) {
			ret = SPI_rx();
			if( (ret > 0) && (ret < 0xFF) ) {
				while( ret & 0x80 ) ret <<= 1;
				ret = ((ret & 0x70) == 0x20);
				break;
			}
		}
		// wait for data to be written
		_Sd_wait_standby();
		_Sd_unselect();
		if( ret ) {
			// data transfer to sd card buffer was successful
			// query for result of actual write operation
			ret = _Sd_send_cmd(SD_CMD_SEND_STATUS, SD_RESPONSE_TYPE_R2, NULL, &r2);
			if( ret && (r2 == 0) )
				result = SD_WRITE_SUCCESS;
		} else {
			// data transfer to sd card buffer failed
		}

		// unlock uart (locked from every write operation)
		USART_Unlock(USART_MODE_SPI);
		return result;
	}
	
	enum sd_write_ret Sd_write_flush() {
		#if SPI_DMA_WRITE
			if( !spi_dma_lock ) return 100;
			return _Sd_write_finish();
		#else
			return 100;
		#endif
	}	

	enum sd_write_ret _Sd_write_block(const uint32_t* pAddress, const void *pBuffer, bool incPtr) {
		uint8_t	r1, ret;

#if FATDEBUGGING
		printf("_Sd_write_block: writing to %lX\r\n", *pAddress);
#endif
		
		// block write-access on write protection
		if( Sd_protected() )
			return SD_WRITE_PROTECTED_ERR;

		// aquire uart
		if( !USART_LockWait(USART_MODE_SPI) )
			return SD_WRITE_INTERFACE_ERR;

		// start write
		STATUS_LED_ON;
		ret = _Sd_send_cmd(SD_CMD_WRITE_SINGLE_BLOCK, SD_RESPONSE_TYPE_R1, pAddress, &r1);
		if( !ret | r1 ) {
			USART_Unlock(USART_MODE_SPI);
			STATUS_LED_OFF;
			return SD_WRITE_COMMAND_ERR;
		}
		// write data
		_Sd_select();
		SPI_tx(0xFF);
		SPI_write( pBuffer, sd_state.BlockLen, SD_TOKEN_WRITE, incPtr );
		STATUS_LED_OFF;
		// finish write
		#if SPI_DMA_WRITE
			spi_dma_lock = true;
			return SD_WRITE_SUCCESS;
		#else
			return _Sd_write_finish();
		#endif
	}

	enum sd_write_ret Sd_set_block(const uint32_t address, const char (*const pChar)) {
		return _Sd_write_block(&address, pChar, false);
	}
	
	enum sd_write_ret Sd_write_block(const uint32_t address, void const (*const pBuffer)) {
		return _Sd_write_block(&address, pBuffer, true);
	}
#endif

///////////////////////////////////////////////////////////////////////////////
// Supporting functions
///////////////////////////////////////////////////////////////////////////////

/**
 * @brief Activate SD Card on SPI Bus
 * \internal
 */
void _Sd_select() {
	P5OUT &= ~0x01;              // Card Select
}

/**
 * @brief	Deactivate SD Card on SPI Bus
 * \internal
 */
void _Sd_unselect() {
	UART_WAIT_TXDONE();
	P5OUT |= 0x01;				// Card Deselect
	SPI_rx();
}

/**
 * @brief	Wait for the card to enter standby state
 * \internal
 */
bool _Sd_wait_standby() {
	sd_response_r1_t r1;
	bool	ret;
	int		i;
	
	SPI_waitToken(0xFF, SD_TIMEOUT_READ);
	
	for( i = 0; i < SD_TIMEOUT_IDLE; i++ ) {
		ret = _Sd_get_op_cond( &r1 );
		if( (ret) && (r1.r1.in_idle_state == 0) )
			return true;
	}
	return false;
}

/**
 * @brief 	Resets the card and (hopefully) returns with the card in standby state
 * \internal
 */
bool _Sd_reset() {
	int		i;
	bool	ret;
	sd_response_r1_t r1;
	

	_Sd_select();
	for( i = 0; i < 5; i++ )
		SPI_tx(0);
	ret = false;

	for( i = 0; i < 520; i++ ) {
		ret = ret || (SPI_rx() != 0xFF);
	}
	if( ret )
		_Sd_send_cmd(SD_CMD_STOP_TRANSMISSION, SD_RESPONSE_TYPE_R1, NULL, &r1);
	
 	for( i = 0; i < 4; i++ ) {
		ret = _Sd_send_cmd(SD_CMD_GO_IDLE_STATE, SD_RESPONSE_TYPE_R1, NULL, &r1);
		if( ret && r1.r1.illegal_cmd ) {
			_Sd_send_cmd(SD_CMD_STOP_TRANSMISSION, SD_RESPONSE_TYPE_R1, NULL, &r1);
			ret = _Sd_send_cmd(SD_CMD_GO_IDLE_STATE, SD_RESPONSE_TYPE_R1, NULL, &r1);
		}
		ret = _Sd_wait_standby();
		if( ret ) return true;
	}
	return false;
}

/**
 * @brief	Reads operating condition from SD or MMC card.
 * \internal
 * \Note	Should allow to find out the card type on first run if needed.
 */
bool _Sd_get_op_cond( sd_response_r1_t* pResponse )
{
	bool ret;
	
	// SD style
	ret = _Sd_send_cmd( SD_CMD_APP_SECIFIC_CMD, SD_RESPONSE_TYPE_R1, NULL, pResponse );
	if( ret )
		ret = _Sd_send_cmd( SD_ACMD_SEND_OP_COND, SD_RESPONSE_TYPE_R1, NULL, pResponse );
	
	// MMC style init
	if( !ret )
		ret = _Sd_send_cmd( SD_CMD_SEND_OP_COND, SD_RESPONSE_TYPE_R1, NULL, pResponse );
	
	if( *((uint8_t*)pResponse) & SD_R1_ERROR_MASK )
		return false; 
	
	return ret;
}

/**
 * @brief	Used to send all kinds of commands to the card and return the response.
 * \internal
 */
bool _Sd_send_cmd( const uint8_t command,
				   const uint8_t response_type,
				   const void* pArg,
				   void (*const pResponse)
				)
{
	uint8_t	data;	// rx buffer
	int		i;		// loop counter


	_Sd_select();

	// send command (1 byte)
	SPI_tx( 0x40 | command );
	// send argument (4 bytes)
	if( pArg == NULL ) {
		for( i = 0; i < 4; i++ )
			SPI_tx(0x00);
	} else {
		for( i = 3; i >= 0; i-- ) 
			SPI_tx( ((uint8_t*)pArg)[i] );
	}
	// send CRC matching CMD0 (1 byte)
	SPI_tx(0x95);
	
	// wait for start bit
	for( i = 0; i < SD_TIMEOUT_NCR; i++ ) {
		data = SPI_rx();
		if( (data & 0x80) == 0 )
			goto _Sd_send_cmd_response;
	}
	// timeout ( i >= SD_TIMEOUT_NCR )
	goto Sd_send_cmd_fail;
	
  _Sd_send_cmd_response:
	// start bit received, read response with size i
	i = response_type - 1;
	if( pResponse != NULL ) {
		// copy response to response buffer
		do {
			((uint8_t*)pResponse)[i] = data;
			if( i == 0 ) break;
			data = SPI_rx();
			i--;
		} while(1);
	} else {
		// receive and ignore response
		SPI_idle(i);
	}

	// done successfully
	_Sd_unselect();
	return true;

 Sd_send_cmd_fail:
	// failed
	_Sd_unselect();
	return false;
}

/**
 * @brief	Read Card Register
 * \internal
 */
uint16_t _Sd_read_register(void* pBuffer, uint8_t cmd, uint16_t size) {
	if( !_Sd_read_start(cmd, 0) )
		return false;
	SPI_read(pBuffer, size, true);
	_Sd_read_stop(2);

	return size;
}

/**
 * @brief	Begin block read operation
 * \internal
 */
bool _Sd_read_start(uint8_t cmd, uint32_t address) {
	uint8_t r1;
	uint8_t ret;
	
	// aquire uart
	if( !USART_LockWait(USART_MODE_SPI) )
		return false;	
	
	ret = _Sd_send_cmd(cmd, SD_RESPONSE_TYPE_R1, &address, &r1);
	if( !ret || r1 ) goto _Sd_read_start_fail;
	
	// Wait for start bit (0)
	ret = _Sd_read_wait();
	if( !ret ) goto _Sd_read_start_fail;

	return true;

  _Sd_read_start_fail:
	USART_Unlock(USART_MODE_SPI);
	return false;
}

/**
 * @brief	Wait for beginning of data
 * \internal
 */
bool _Sd_read_wait() {
	uint16_t	i;		// loop counter
	uint8_t	data;	// rx buffer

	_Sd_select();

	for( i = 0; i < SD_TIMEOUT_READ; i++ ) {
		data = SPI_rx();
		if( data == SD_TOKEN_READ ) {
			// token received, data bytes follow
			STATUS_LED_ON;
			return true;
		}
		/*
		Following code handles error tokens. Since these are currently not used in the
		application they can just be ignored. Anyway this is still useful when debugging.
		else if( (data != 0) && (data & SD_DATA_ERROR_TOKEN_MASK) == data ) {
			// data error token
			SPI_rx();
			break;
		} */
	}

	// error or timeout
	_Sd_unselect();
	return false;					
}

/**
 * @brief	Finished with reading, stop transfer
 * \internal
 */
void _Sd_read_stop(uint16_t count) {
	// finish block + crc
	if( count ) {
		uint8_t dump;
		SPI_read( &dump, count + 2, false );
		_Sd_unselect();
	}
	STATUS_LED_OFF;

	// wait for switch to standby mode
	if( !_Sd_wait_standby() )
		_Sd_reset();
		
	// unlock uart (locked from _Sd_read_start)
	USART_Unlock(USART_MODE_SPI);
}

/*
// read block
COMMAND(sdr, 0) {

char buffer[512];
uint32_t addr = String_parseLongInt(&str[4], NULL);
	Sd_read_block(buffer, addr);
		uint8_t	cs[2];
		uint16_t	i;

		printf("%u Bytes follow:\r\n", sd_state.BlockLen);
		for(i = 0; i < sd_state.BlockLen; ) {
			String_fromByteToHex(buffer[i], cs);
			printf("%.2s", cs);
			i++;
			if( (i % 2) == 0 ) printf(" ");
			if( (i % 16) == 0 ) printf("\r\n");
		}
}

*/
