/**
 * Copyright (c) 2008, Freie Universitt Berlin, Tomasz Naumowicz
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:
 * 
 * Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
 * Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.
 * Neither the name of the Freie Universitt Berlin nor the names of its contributors may be used to endorse or promote products derived from this software without specific prior written permission.
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * 
 * Based on the ScatterWeb firmware from the Freie Universitaet Berlin under the BSD license.  
 */


/**
 * @file	ScatterWeb.Sd.cache.c
 * @ingroup	libsd
 * @brief	MMC-/SD-Card library, cached read and write
 * 
 * @author	Michael Baar	<baar@inf.fu-berlin.de>
 * @date	Jan 2007
 * @version	0.2
 */

/**
 * @addtogroup	libsd
 * @{
 */
#include "ScatterWeb.Sd.internals.h"

#if SD_CACHE

void _Sd_cache_init() {
	uint32_t adr = 0;
	
	sd_state.Cache->address = 1; //ok, it's just a dummy value for now, it will be modified durint the initlization
	sd_state.Cache->state = 0;
	// pre-read first block
	Sd_cache_read_block(&adr);
	SD_FREE_LOCK(sd_state.Cache);
}

void Sd_cache_flush() {
	#if SD_WRITE
		SD_GET_LOCK(sd_state.Cache);
		if(sd_state.Cache->state & SD_CACHE_DIRTY) {		
			Sd_set_blocklength(SD_WRITE_BLOCKLENGTH_BIT);
			Sd_write_block(sd_state.Cache->address, sd_state.Cache->buffer);
			sd_state.Cache->state &= ~SD_CACHE_DIRTY;
		}
		SD_FREE_LOCK(sd_state.Cache);
	#endif
}

/*
void Sd_cache_flush() {
	_Sd_cache_flush();
}
*/

sd_cache_t* Sd_cache_read_block(const uint32_t* pblAdr) {	
#if FATDEBUGGING
	printf("[Sd_cache_read_block %lX]", *pblAdr);
#endif
	SD_GET_LOCK(sd_state.Cache);
	if( sd_state.Cache->address != *pblAdr ) { //cache miss, need to reload the data
#if FATDEBUGGING
			printf("[Sd_cache_read_block.MISS]");
#endif
			Sd_set_blocklength(SD_WRITE_BLOCKLENGTH_BIT);
		if(sd_state.Cache->state & SD_CACHE_DIRTY) { //the cache was modified --> store the content of the cache to the SD card
#if FATDEBUGGING
			printf("[Sd_cache_read_block.WASDIRTY.stored]");
#endif
			Sd_write_block(sd_state.Cache->address, sd_state.Cache->buffer);
			sd_state.Cache->state &= ~SD_CACHE_DIRTY;
		}
		sd_state.Cache->address = *pblAdr; //set the new address
		if( !Sd_read_block(sd_state.Cache->buffer, *pblAdr) ) { //read the data, if it fails, then
			SD_FREE_LOCK(sd_state.Cache);						//free the lock (THIS WAS MISSING IN THE PREVIOUS VERSION CODE)
			return false;										//and return false 
		}
#if FATDEBUGGING
		printf("[Sd_cache_read_block.cache updated]");
#endif
	}
#if FATDEBUGGING
	else {
			printf("[Sd_cache_read_block.HIT]");
	}
#endif
	SD_FREE_LOCK(sd_state.Cache); //success path --> free the lock, return the cache

	return sd_state.Cache;			
}

#if SD_READ_ANY
uint16_t Sd_read(
		void* pBuffer,
		uint32_t address,
		uint16_t size
	) {
		uint16_t	offset;			// bytes from aligned address to start of first byte to keep
		char*			p;				// pointer to current pos in receive buffer
		uint16_t	bytes_left;		// num bytes to read
		uint16_t	read_count;		// num bytes to read from current block
		
		//
		// parameter processing
		//
		p = (char*)pBuffer;
		bytes_left = size;
			
		// align to block
		//offset = Sd_AlignAddress(&address); //address changes here too ::::: this was an error, offset has to be computed every time in the do-while loop
		
		//
		// Data transfer
		//
		do {

		    offset = Sd_AlignAddress(&address); //address changes here too

			// calculate block
			if( (offset == 0) && (bytes_left >= sd_state.BlockLen) ) {
				read_count = sd_state.BlockLen;
				Sd_read_block(p, address); //cool, no need for cache, leave the old cache in memory
			} else {
				Sd_cache_read_block(&address);
				read_count = bytes_left + offset; //read_count is used as a temp varialbe here in order to see, if the data is contained in the current block (comparing with the size of the block)
				if( read_count > sd_state.BlockLen )
					read_count = sd_state.BlockLen - offset;
				else
					read_count = bytes_left;
			
				memcpy(p, sd_state.Cache->buffer + offset, read_count);
				SD_FREE_LOCK(sd_state.Cache);
			}
			
			bytes_left -= read_count;
			if( bytes_left == 0 )
				return size;			
			
			p+= read_count;
			address += sd_state.BlockLen;
		} while(1);
	}
#endif // SD_READ_ANY

#if SD_WRITE
	uint16_t Sd_write( //this function is able to write data _across_ block boundaries, checked, OK!
		uint32_t address,
		void* pBuffer,
		uint16_t size
	) {
		uint16_t	offset;			// bytes from aligned address to start of first byte to keep
		char*	p;						// pointer to current pos in receive buffer
		uint16_t	bytes_left;		// num bytes to read
		uint16_t	read_count;		// num bytes to read from current block
		
		//
		// parameter processing
		//
		p = (char*)pBuffer;
		bytes_left = size;
			
		

		
		Sd_set_blocklength(SD_WRITE_BLOCKLENGTH_BIT);
		
		//
		// Data transfer
		//
		do {
			// align to block
			offset = Sd_AlignAddress(&address);
		
			// calculate block
			if( (offset == 0) && (bytes_left >= sd_state.BlockLen) ) { //ok, the adress was aligned, and there is at least BlockLen of data
				read_count = sd_state.BlockLen;						   //so, this one can be just stored without preprocessing, the simple case
				Sd_write_block(address, p);
			} else {
				Sd_cache_read_block(&address); 
				read_count = bytes_left + offset;
				if( read_count > sd_state.BlockLen )
					read_count = sd_state.BlockLen - offset;
				else
					read_count = bytes_left;
			
				memcpy(sd_state.Cache->buffer + offset, p, read_count);
				sd_state.Cache->state |= SD_CACHE_DIRTY;

#if FATDEBUGGING
		printf("[Sd_write.DIRTY]");
#endif

				SD_FREE_LOCK(sd_state.Cache);
			}

			bytes_left -= read_count; 
			
			if( bytes_left == 0 ) //sth strange could happen here: assume the address was aligned (the first if in the loop)
				return size;				//bytes_left doesn't change, so we dont leave this loop now.
			
			p+= read_count;
			address += sd_state.BlockLen;   
		} while(1);
	}
#endif // SD_WRITE

#endif	// SD_CACHE

/** @} */

