/**
 * Copyright (c) 2008, Freie Universitt Berlin, Tomasz Naumowicz
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:
 * 
 * Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
 * Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.
 * Neither the name of the Freie Universitt Berlin nor the names of its contributors may be used to endorse or promote products derived from this software without specific prior written permission.
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * 
 * Based on the ScatterWeb firmware from the Freie Universitaet Berlin under the BSD license.  
 */



/**
 * @file	ScatterWeb.Sd.h
 * @ingroup	libsd
 * @brief	MMC-/SD-Card library, Public interface
 * 
 * @author	Michael Baar	<baar@inf.fu-berlin.de>
 * @date	Jan 2007
 * @version	0.2
 * 
 * Revised and modified by Tomasz Naumowicz naumowic@inf.fu-berlin.de
 * 
 */


#ifndef __SD_H__
#define __SD_H__

#define SD_BLOCKLENGTH_INVALID		0xFF
#define	SD_WRITE_BLOCKLENGTH_BIT	9
#define SD_WRITE_BLOCKLENGTH		0x200


#define SD_READ_PARTIAL		0x80
#define SD_WRITE_PARTIAL	0x40
/** @} */

#ifdef __TI_COMPILER_VERSION__
	#pragma pack(1)
#endif

/**
 * @brief	Card Identification Register
 */
#ifndef __TI_COMPILER_VERSION__
	__attribute__((packed))
#endif
typedef struct sd_cid {
	uint8_t		mid;
	char		oid[2];
	char		pnm[5];
	uint8_t		revision;
	uint32_t	psn;
	uint16_t				:  4;
	uint16_t	mdt			: 12;
	uint8_t		crc7		:  7;
	uint8_t					:  1;
} sd_cid_t;

/**
 * @name	Card Specific Data register
 * @{
 */
typedef struct sd_csd {
	uint8_t	raw[16];
} sd_csd_t;
#define SD_CSD_READ_BL_LEN(csd)		(( csd ).raw[5] & 0x0F)		// uchar : 4
#define SD_CSD_READ_PARTIAL(csd)	(( csd ).raw[ 6] & 0x80)	// bool
#define SD_CSD_CCC(csd)				( (( csd ).raw[4]<<4) | (( csd ).raw[5]>>4) )	// uchar
#define SD_CSD_WRITE_PARTIAL(csd)	(( csd ).raw[13] & 0x04)	// bool
#define SD_CSD_C_SIZE(csd)			( ((( csd ).raw[6] & 0x03)<<10) | (( csd ).raw[7]<<2) | (( csd ).raw[8]>>6) ) // uint : 12
#define SD_CSD_C_MULT(csd)			( ((( csd ).raw[9] & 0x03)<<1) | (( csd ).raw[10]>>7) )	// uchar : 4
/** @} */

#ifdef __TI_COMPILER_VERSION__
	#pragma pack()
#endif

/// Card access library state
#define SD_CACHE_LOCKED			0x01
#define SD_CACHE_DIRTY			0x02
typedef struct {
	char		buffer[SD_WRITE_BLOCKLENGTH];
	uint32_t	address;
	uint8_t		state;
} sd_cache_t;
typedef struct {
	uint16_t		MinBlockLen_bit : 4;	///< minimum supported blocklength
	uint16_t		MaxBlockLen_bit : 4;	///< maximum supported blocklength
	uint16_t		Flags			: 8;	///< feature flags
	uint8_t			BlockLen_bit;			///< currently selected blocklength as bit value (n where BlockLen is 2^n)
	uint16_t		BlockLen;				///< currently selected blocklength for reading and writing
   #if SD_CACHE
	sd_cache_t*	Cache;
   #endif
} sd_state_t;

extern volatile sd_state_t	sd_state;	///< Card access library state

/**
 * @brief	Setup ports for sd card communication
 * 
 * This function needs to be called once before any other library function.
 * It invokes ::Spi_enable to configure UART1 for SPI communication.
 * 
 * If you need to reconfigure the UART for other operations only call
 * ::Spi_enable to return to SPI mode. ::Sd_setup needs to be run only once.
 */
void Sd_init();

/**
 * @brief	Return value of ::Sd_init function
 */
enum sd_init_ret {
	SD_INIT_SUCCESS			= 0,
	SD_INIT_FAILED			= 1,
	SD_INIT_NOTSUPP			= 2,
	SD_INIT_NOCARD			= 3,
	SD_INIT_WRITEPROTECTED	= 4,
};

/**
 * @brief	Return value of write functions
 * @see ::Sd_write, ::Sd_write_block
 */
enum sd_write_ret {
	SD_WRITE_SUCCESS		= 0,		///< writing successfull
	SD_WRITE_PROTECTED_ERR	= 1,		///< card write protected
	SD_WRITE_INTERFACE_ERR	= 2,		///< error in UART SPI interface
	SD_WRITE_COMMAND_ERR	= 3,		///< error in write command or command arguments (e.g. target address)
	SD_WRITE_STORE_ERR		= 4			///< storing written data to persistant memory on card failed
};

enum sd_init_ret SD_Card_Init_Result;

/**
 * @brief	Initialize card and state
 * @return	one of sd_init_ret
 * 
 * Initializes the memory card, checks supported voltage range and
 * functionality. Initializes the global state struct sd_state.
 * Should be invoked once immediately after ::Sd_setup.
 */
enum sd_init_ret SD_Card_Init();

/**
 * @brief	Last operation to call when finished with using the card.
 */
void Sd_close();

/**
 * @brief SD Card physically present?
 * @return	if no card present returns 0
 */
#define Sd_detected()	((P2IN & 0x40) == 0)
 
/**
 * @brief SD Card locked by switch?
 * @return	if card is not locked returns 0
 */
#define	Sd_protected()	((P2IN & 0x80) != 0)

/**
 * @brief	Set blocklength to 2 ^ n
 * 
 * Tries to set the card's blocklength for reading and writing to any
 * 2^n value, where n shall be between 0 and 11.
 * Be aware that a card may or may not support different blocksizes.
 * 
 * Since all read and write operations have to use blockaligned addresses
 * and need to process complete blocks always try to use the optimal blocksize
 * and let ::Sd_read do the rest. If the blocklength is already set to the new
 * value nothing is done.
 * 
 * \Note
 * The default blocklength is 512 (n=9). SD Cards usually support partial blocks
 * with a blocklength of 1 to 512 (n=0 to n=9), MMC cards don't.
 * Large cards support blocklength up to 2048 (n=11).
 * The supported range of blocklengths can be read as n-value from 
 * sd_state.BlockLenMin_bit and sd_state.BlockLneMax_bit. The current value
 * is available as n-value in sd_state.Blocsd_state.BlockLen_bit and as byte-value
 * in sd_state.BlockLen.
 * 
 * @param[in]	blocklength_bit		blocklength as n-value of 2^n
 * 
 * @return	Returns blocklength_bit if set operation was successful or
 * 			SD_BLOCKLENGTH_INVALID otherwise.
 * 
 */
uint8_t Sd_set_blocklength(const uint8_t blocklength_bit);

/**
 * @brief			Align byte address to current blocklength
 * @param[in,out]	pAddress	address to align, will be modified to be block aligned
 * @return			Offset from aligned address to original address
 */
uint16_t Sd_AlignAddress(uint32_t* pAddress);

/**
 * @brief	Read one complete block from a block aligned address into buffer
 * 
 * This function reads a single block of the currently set blocklength
 * from a block aligned address.
 * 
 * \Note: If address is not block aligned the card will respond with an error
 * and no bytes will be read.
 * 
 * @param[out]	pBuffer		Pointer to a buffer to which data is read. It should be least
 * 							1 byte large
 * @param[in]	address		The address of the first byte that shall be read to pBuffer
 * 
 * @return		Number of bytes read (should always be = sd_state.BlockLen)
 */
uint16_t Sd_read_block(void (*const pBuffer), const uint32_t address);

#if SD_READ_BYTE
	/**
	 * @brief	Read one byte from any address
	 * This function reads a single byte from any address. It is optimized for best speed
	 * at any blocklength.
	 * \Note: blocklength is modified
	 * 
	 * @param[out]	pBuffer		Pointer to a buffer to which data is read. It should be least
	 * 							1 byte large
	 * @param[in]	address		The address of the byte that shall be read to pBuffer
	 * 
	 * @return	Number of bytes read (usually 1)
	 */	
	bool Sd_read_byte(void* pBuffer, const uint32_t address);
#endif

#if SD_WRITE
	/**
	 * @brief Write one complete block at a block aligned address from buffer to card
	 * 
	 * @param[in]	address		block aligned address to write to
	 * @param[in]	pBuffer		pointer to buffer with a block of data to write
	 * @return		result code (see enum #sd_write_ret)
	 * 
	 * \Note
	 * Only supported block size for writing is usually 512 bytes.
	 */
	enum sd_write_ret Sd_write_block(const uint32_t address, void const (*const pBuffer));

	/**
	 * @brief		Fill one complete block at a block aligned address with
	 * 				a single character.
	 * 
	 * @param[in]	address		block aligned address to write to
	 * @param[in]	pChar		pointer to buffer with a character to write
	 * @return		result code (see enum #sd_write_ret)
	 * 
	 * @note		Use this for settings blocks to 0x00. 
	 *				Only supported block size for writing is usually 512 bytes.
	 */	
	enum sd_write_ret Sd_set_block(const uint32_t address, const char (*const pChar));
	
	/**
	 * @brief		Flush the DMA write buffer
	 * 
	 * Wait for a running DMA write operation to finish
	 */
	enum sd_write_ret Sd_write_flush();
#endif

	void Sd_flush();

#if SD_CACHE
#if FATDEBUGGING
    #define	SD_WAIT_LOCK(x)	while( x ->state & SD_CACHE_LOCKED ) { _NOP(); printf("[w]"; P5OUT &=~0x80;); }
	#define SD_GET_LOCK(x)	do { while( x ->state & SD_CACHE_LOCKED ) { _NOP(); }; x ->state |= SD_CACHE_LOCKED; printf("[g]"); P5OUT &=~0x80;} while(0)
	#define SD_FREE_LOCK(x)	do { x ->state &= ~SD_CACHE_LOCKED; printf("[f]"); P5OUT |= 0x80; } while(0)
#else

    #define	SD_WAIT_LOCK(x)	while( x ->state & SD_CACHE_LOCKED ) { _NOP(); P5OUT &=~0x80;); }
	#define SD_GET_LOCK(x)	do { while( x ->state & SD_CACHE_LOCKED ) { _NOP(); }; x ->state |= SD_CACHE_LOCKED; P5OUT &=~0x80;} while(0)
	#define SD_FREE_LOCK(x)	do { x ->state &= ~SD_CACHE_LOCKED; P5OUT |= 0x80; } while(0)
#endif

	/**
	 * @brief	Flush the sd cache
	 * 
	 * Writes back the cache buffer, if it has been modified. Call this if
	 * a high level operation has finished and you want to store all data
	 * persistantly. The write back operation does not use timers.
	 */
	void Sd_cache_flush();
	
	/**
	 * @brief	Read a block into the cache buffer
	 * @internal
	 * 
	 * You won't usually need this operation.
	 */
	sd_cache_t* Sd_cache_read_block(const uint32_t* blAdr);

#if SD_WRITE
		uint16_t Sd_write(
		uint32_t address,
		void* pBuffer,
		uint16_t size);
#endif

#endif

/**
 * @brief		Erase (clear) blocks of data
 * @param[in]	address		Starting address of first block to erase
 * @param[in]	numBlocks	Number of blocks to erase (starting at address)
 * @return		true if successful
 * @note		Data can only be erased in whole blocks. All bytes will be set
 *				predifined character (see CSD). Common cards use 0xFF.
 */
bool Sd_erase_blocks(const uint32_t address, const uint16_t numBlocks);

/**
 * @brief		Read card identification register (CID)
 * @param[out]	pCID	Pointer to buffer for CID
 * @return		Number of bytes read (= size of CID) or 0 if failed
 *
 * @see			::sd_cid_t
 */
uint16_t Sd_read_cid(sd_cid_t* pCID);

/**
 * @brief	Read size of card
 * @return	Number of bytes available
 *
 * Reads the number of bytes in the card memory from the CSD register.
 */
uint32_t Sd_get_size();

#if SD_READ_ANY
	/**
	 * @brief	Read any number of bytes from any address into buffer
	 * 
	 * @param[out]	pBuffer		Pointer to a buffer to which data is read. It should be least
	 * 							size bytes large
	 * @param[in]	address		The address of the first byte that shall be read to pBuffer
	 * @param[in]	size		Number of bytes which shall be read starting at address
	 */
	uint16_t Sd_read(void* pBuffer, uint32_t address, uint16_t size);
#endif

#endif /*__SD_H__*/
/** @} */
