/**
 * Copyright (c) 2008, Freie Universitt Berlin, Tomasz Naumowicz
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:
 * 
 * Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
 * Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.
 * Neither the name of the Freie Universitt Berlin nor the names of its contributors may be used to endorse or promote products derived from this software without specific prior written permission.
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * 
 * Based on the ScatterWeb firmware from the Freie Universitaet Berlin under the BSD license.  
 */


/*
 * Limited usage of functions in order to save memory.
 * goto's instead of functions preferred in such scenario.
 */

#include "System.h"

/**
 * Application Header at a fixed location in memory
 * Used to determine the version and the name of the application.
 */
APP(APPH_LOADER, __APP_NAME, __APPVER_MAJOR, __APPVER_MINOR);

//////////////////////////////////////////////////////////////////
// Helper Funtions:
/**
 * Converts hexadecimal number formatted as two characters to byte
 * Increments the input pointer by 2
 */
unsigned char hex2byte(char** cs) {
	register unsigned int a;
	register unsigned int b;
	
	a = (*(*cs)++) - '0';
	if( a > 9 )
		a -= 7;	
	b = (*(*cs)++) - '0';
	if( b > 9 )
		b -= 7;
		
	a<<=1; a<<=1; a<<=1; a<<=1;		// smaller by 2 bytes from "a <<= 4;"...
	
	b |= a;
	return( b );
}

/**
 * Converts hexadecimal number formatted as four characters to word
 * Increments the input pointer by 4
 */
unsigned int hex2word(char** cs) {
	register unsigned int w;
	register unsigned int b;
	register unsigned int cnt = 3;
	
	w = (*(*cs)++) - '0';
	if( w > 9 )
		w -= 7;
	
 hex2word_halfbyte:
		b = (*(*cs)++) - '0';
		if( b > 9 )
			b -= 7;
		w<<=1; w<<=1; w<<=1; w<<=1;	// smaller by 2 bytes from "w <<= 4;"...
		w |= b;
		
		cnt--;
	if( cnt ) goto hex2word_halfbyte;
	
	return w;	
}



/**
 * used for visual feedback (blinking LED)
 */
void _blockExecution() {
	volatile uint16_t i;
	for( i = 0; i < 60000; i++ ) { _NOP(); _NOP(); }
}

/**
 * Visual Feedback to the user: blinking LED
 */
void VisualMessage(uint8_t n) {
	uint16_t j, k;

	
	STATUS_LED_OFF;
	for( k = 0; k < 4; k++ ) {
		_blockExecution();
		for( j = 0; j < n; j++ ) {
			STATUS_LED_ON;
			_blockExecution();
			STATUS_LED_OFF;
			_blockExecution();
		}
	}
	
	_blockExecution();
}

/**
 * Ereases one flash page (erease == set all bytes to 0xFF)
 */
void _Prog_Flash_clr( unsigned int* address ) {
	FCTL3 = FWKEY; 					// Lock = 0   

	FCTL1 = FWKEY | ERASE;			// ERASE = 1 
	*address = 0;					// erase Flash segment 

	FCTL3 = FWKEY | LOCK;			// Lock = 1 
}

/**
 * Writes ONE word to the flash memory
 * (writing == setting selected bytes to zero, that's why ereasing is required first)
 */
void _Prog_Flash_ww( unsigned int* address, unsigned int word) {//, uint8_t writeReset) { //added the 3. parameter to make sure the reset address will not be overwritten
	FCTL3 = FWKEY;					// Lock = 0 

	FCTL1 = FWKEY | WRT;			// WRT = 1 
	
	*address = word;				// program Flash word 
	
	FCTL1 = FWKEY;					// WRT = 0 
	FCTL3 = FWKEY | LOCK;           // Lock = 1 
	
	WDTCTL = WDTPW | WDTHOLD;		// stop watchdog timer
}

file_handle_t UpdateFile;
file_handle_t hRootDir;
struct fat_direntry	dirEntry;

#define INFOMEM_START		0x1000
#define INFOMEM_BLOCK_SIZE	128

//TODO: change defines to enums later
#define APP_FLASHED				1
#define APP_NOT_FLASHED			0
#define USING_GOLDENIMAGE		1
#define NOT_USING_GOLDENIMAGE	0

void INFOMEM_write_bootloader_state(uint8_t flashed, uint8_t failstartcounter, uint8_t usinggoldenimage) {
	char backup[INFOMEM_BLOCK_SIZE];
	register uint8_t* buffer;
	register uint16_t i;
	uint8_t*	 flash;
	
	bootloader_offline_state_t bootloader_state;
	
	bootloader_state.Flashed 					= flashed;
	bootloader_state.FailStartsCounter 			= failstartcounter;
	bootloader_state.UsingGoldenImage 			= usinggoldenimage;
	bootloader_state.Checksum 					= flashed + usinggoldenimage + __APPVER_MAJOR + __APPVER_MINOR;

	flash = (uint8_t*)INFOMEM_START;

	// backup INFOMEM block into RAM
	memcpy(backup, flash, INFOMEM_BLOCK_SIZE);

	// merge backup with new data
	buffer = (uint8_t*)backup;
	
	memcpy(buffer, &bootloader_state, BLD_OFFLINE_STATE_SIZE);

	// init flash access
    FCTL3 = FWKEY;
	// erase flash
    FCTL1 = FWKEY | ERASE;
	*flash = 0;
	// write flash
    FCTL1 = FWKEY | WRT;
    
	buffer = (uint8_t*)backup;
	
	for(i = 0; i < INFOMEM_BLOCK_SIZE; i++) {
		*flash = *buffer;
		buffer++; flash++;
	}
    FCTL1 = FWKEY;
    FCTL3 = FWKEY | LOCK;
}



#define BUFFER_SIZE 			64
#define FLASH_SEGMENT_SIZE		0x200
uint8_t _buffer[BUFFER_SIZE];


/**
 * functions loads next "BUFFER_SIZE" characters from the hex file with the update
 * only the bytes that were already processed are discarded from the buffer (detected via the bytesLeft variable)
 * this means that e.g. calling this function twice in a row would not lead to any change to the content of the buffer
 */ 
void _Program_Fetch(unsigned int* bytesLeft, char** p) {
	//copy the not yet read bytes to the beginning of the bufer
	memcpy(_buffer, *p, *bytesLeft);						
	
	//fill the rest of the buffer with new data
	*bytesLeft = Fat_ReadFile( &FAT_State, &UpdateFile, _buffer + *bytesLeft, BUFFER_SIZE - *bytesLeft );
	
	if( *bytesLeft == 0 )
		return;
	
	*p = _buffer;

	*bytesLeft = BUFFER_SIZE;
}

/**
 * Reads the application header data from the hex file.
 * Gets the application version.
 * 
 * TODO: add checksum check of the application header!
 */
bool _ReadHexVersion(uint8_t* major, uint8_t* minor, uint32_t* compileTime) {
	  	// buffer
		char*	p;
		unsigned int	fpReset;
		unsigned int	charsLeft;
		
		// file info
		unsigned long	filePos;

		// line info
	  	unsigned char	lineLen;
		unsigned int	flashAdr;
		unsigned char	recordType;

		charsLeft = 0;
		p = _buffer;
		filePos = UpdateFile.Size & 0x0000FFFF;	// countdown
		fpReset = 0;

		do {
			//
			// read a single line of intel hex file
			//
			
			// secure header
			if( charsLeft < 10 ) {
				_Program_Fetch( &charsLeft, &p );
				if( charsLeft == 0 )
					goto readfailed;
			}
			
			// read header
			p++;							// colon
			lineLen = hex2byte( &p );		// number of data bytes in record
			flashAdr = hex2word( &p );		
			recordType = hex2byte( &p );

			charsLeft -= 9; filePos -= 9;

			// secure data + checksum
			lineLen <<= 1;					// number of data characters to read
			if( charsLeft < (lineLen + 2) ) {
				_Program_Fetch( &charsLeft, &p );
				if( charsLeft == 0 )
					goto readfailed;
			}
			
			//
			// read data record
			//
			if( recordType == 00 ) {		// data record
				filePos -= lineLen; charsLeft -= lineLen;


				lineLen >>= 1;			// number of data bytes to read

				while( lineLen > 0 ){

					// find the compile time:
					// compile time is a a value of the type uint32_t == four bytes
					if (flashAdr == __HEX_COMPILETIME_ADDRESS_B4) {
						// read 1st byte
						uint32_t tmp = hex2byte(&p);
						tmp = tmp << 24;
						*compileTime |= tmp;
						
						return true;
					}
					else
					if (flashAdr == __HEX_COMPILETIME_ADDRESS_B3) {
						// read 1st byte
						uint32_t tmp = hex2byte(&p);
						tmp = tmp << 16;
						*compileTime |= tmp;
					}
					else
					if (flashAdr == __HEX_COMPILETIME_ADDRESS_B2) {
						// read 1st byte
						uint32_t tmp = hex2byte(&p);
						tmp = tmp << 8;
						*compileTime |= tmp;
					}
					else
					if (flashAdr == __HEX_COMPILETIME_ADDRESS_B1) {
						// read 1st byte
						uint32_t tmp = hex2byte(&p);
						*compileTime = tmp;
					}
					else
					if( flashAdr == __HEX_VERSION_MAJOR_ADDRESS) {	 //minor version number is stored after the major version number
						
						*major = hex2byte(&p);
						*minor = hex2byte(&p);
						//p += 4;
						
						lineLen--;
						flashAdr += 1; 
						//*major = update_major;
						//*minor = update_minor;

						//we have the data, the function can abort
						//return true;
					}
					else {
						//just skip this byte (hex coded == 2 characters)
						p += 2;
					}

					lineLen--;
					flashAdr += 1; 
				}
			}
			
			if( recordType == 01 ) {		// end record (has lineLen == 0)
				// done
				return false;
			}
			
			  //finish_line:
			  // it's possible that the next ':' wasn't read yet
			  // that's why the update of the buffer will be forced now:
			  _Program_Fetch( &charsLeft, &p );
			  
			  finish_line_loop:
			    if (charsLeft == 0)
			    	goto readfailed;
			    
				if( *p != ':' ) {	// skip untl the next ':'
					p++;
					filePos--;
					charsLeft--;
					
					goto finish_line_loop;
				}
			
		} while( filePos > 0 );
	    
	 readfailed:
	  return false;
	  
}

/**
 * Deletes the entire flashed application
 */
void Prog_Clear() {
	unsigned int adr;
	
	for( adr = 0xFC00; adr >= 0x4600 ; adr -= FLASH_SEGMENT_SIZE )
		_Prog_Flash_clr((unsigned int*)(adr));
}

/**
 * reprograms the node, reads the open hex file using the _Program_Fetch function.
 */
fp_t _Reprogram() {
  	// buffer
	char*	p;
	unsigned int	flashBuffer;
	unsigned int	fpReset;
	unsigned int	charsLeft;
	
	// file info
	unsigned long	filePos;

	// line info
  	unsigned char	lineLen;
	unsigned int	flashAdr;
	unsigned char	recordType;
	
	#if ENABLE_FLASH_CHECKSUM
		unsigned char	checksum;
		unsigned char b;
	#endif
	
	// state
	bool			ivt_cleared;
		
	ivt_cleared = 0;
	charsLeft = 0;
	p = _buffer;
	filePos = UpdateFile.Size & 0x0000FFFF;	// countdown
	fpReset = 0;

	// Mass erase application area
	Prog_Clear();
	
	do {
		//
		// read a single line of intel hex file
		//
		
		// secure header
		if( charsLeft < 10 ) {
			_Program_Fetch( &charsLeft, &p );
			if( charsLeft == 0 )
				goto readfailed;
		}
		
		// read header
		p++;							// colon
		lineLen = hex2byte( &p );		// number of data bytes in record
		flashAdr = hex2word( &p );		
		recordType = hex2byte( &p );
		// header done
		charsLeft -= 9; filePos -= 9;
		
		#if ENABLE_FLASH_CHECKSUM
			// calculate checksum of header
			checksum = lineLen + (flashAdr & 0x00FF) + ((flashAdr & 0xFF00) >> 8) + recordType;
		#endif
		
		// secure data + checksum
		lineLen <<= 1;					// number of data characters to read
		if( charsLeft < (lineLen + 2) ) {
			_Program_Fetch( &charsLeft, &p );
			if( charsLeft == 0 )
				goto readfailed;
		}
		
		//
		// read and flash data record
		//
		if( recordType == 00 ) {		// data record
			filePos -= lineLen; charsLeft -= lineLen;
			
			if (flashAdr >= 0xE582) {
				STATUS_LED_ON;
			}
			
			// test if the decoded address is valid (outside bootloader address space)
			if( flashAdr >= __INFO_START && flashAdr < __INFO_END ) {

				// ignore this line, this is the infomem area, should be preserved
				//p += lineLen;
				goto finish_line;
			}
			else if ( flashAdr >= ((unsigned int)(__bootloader_start))			//don't overwrite the bootloader (values set via the linker script of the bootloader)
					  && flashAdr <  ((unsigned int)(__bootloader_end)) )	{

				// wrong image
				goto badimage;
			}

			// are this the interrupt vectors? special handling is required: use everything from the hex file except the reset address --> the bootloader should be always executed first
			if (flashAdr >= 0xFE00 ) {
				if (!ivt_cleared) {
					// clear segment containing interrupt vector table
					_Prog_Flash_clr( (unsigned int*)flashAdr );
					ivt_cleared = true;
					#if BOOTLD_IS_SYSTEM_LOADER
						//and write the reset address to the ISR table, so that the bootloader will run after POR!
						_Prog_Flash_ww((unsigned int*)0xFFFE, (appHeader.resetAdr));
					#else
						// -> node will be unrecoverable from now on till finish
					#endif
				}
			}

			lineLen >>= 2;			// number of data words to write
			for( ; lineLen > 0; lineLen--, flashAdr += 2) {
				flashBuffer = hex2byte(&p);
				flashBuffer|= hex2byte(&p) << 8;

				#if ENABLE_FLASH_CHECKSUM
					checksum += (flashBuffer & 0x00FF) + ((flashBuffer & 0xFF00) >> 8);
				#endif
				
				#if BOOTLD_IS_SYSTEM_LOADER
					// Reset address already set to bootloader, don't modify
					if( flashAdr == 0xFFFE ) {
						fpReset = flashBuffer;
						flashBuffer = appHeader.resetAdr;
					} 
						
					_Prog_Flash_ww((unsigned int*)flashAdr, flashBuffer);
					STATUS_LED_OFF;
				#else
					_Prog_Flash_ww((unsigned int*)flashAdr, flashBuffer);				
				#endif
			}
		} else if( recordType == 03 ) {
			// start address
			p += 4;
			flashBuffer = hex2word( &p );
			#if ENABLE_FLASH_CHECKSUM
				checksum += (flashBuffer & 0x00FF) + ((flashBuffer & 0xFF00) >> 8);	
			#endif
			filePos -= 8; charsLeft -= 8;
		}
		
		#if ENABLE_FLASH_CHECKSUM
			checksum = ~checksum + 1;
			b = hex2byte( &p );
			filePos -= 2; charsLeft -= 2;
			if( checksum != b )
				goto readfailed;
		#endif
		
		if( recordType == 01 ) {		// end record (has lineLen == 0)
			// done :)
			#if BOOTLD_IS_SYSTEM_LOADER
				return (fp_t)fpReset;
			#else
				FCTL3 = 0xDEAD;
				return NULL;
			#endif
		}
		
		  finish_line:
		  // it's possible that the next ':' wasn't read yet
		  // that's why the update of the buffer will be forced now:
		  _Program_Fetch( &charsLeft, &p );
		  
		  finish_line_loop:
		    if (charsLeft == 0)
		    	goto readfailed;
		    
			if( *p != ':' ) {	// skip untl the next ':'
				p++;				// CR/LF
				filePos--;
				charsLeft--;
				
				goto finish_line_loop;
			}

	} while( filePos > 0 );
    
 badimage:
 readfailed:
  return NULL;
}


/**
 * MSB430MS helper function
 */
void _SHR_Shift8bit(unsigned char input) {
	unsigned char counter;
	
	for (counter = 0; counter < 8; counter++) {
		//pull down the SRCLK (shift clock)
		_SHR_SRCLK_LOW;
		
		//set the SER (data)
		if (input & BIT7) {
			_SHR_SER_HIGH;
		}
		else {
			_SHR_SER_LOW;
		}
			
		//move to the next bit
		input = input << 1;
		
		//pull high the SRCLK (shift clock) - the SER will be read and shifted by the SN74HC595
		_SHR_SRCLK_HIGH;	
	}
}

appHeader_t* _applicationHeader;


volatile uint16_t dcoDiff;       ///< Everytime dcoDiff==0 TIMER_B1 interrupt writes a new DCO diff measure into it.
volatile uint8_t  checkDCOState; ///< DCO will only be changed if two times low or high in a row to avoid peaks. Implemented with this variable.
#define DELTA 1100                   // Target DCO = DELTA(1100)*(4096Hz) , DELTA=550 => 4505600 Hz  */
#define UP 1                         // Temporary.
#define DOWN 2                       // Temporary.

bool DCO_set()
{
	uint16_t Compare, Oldcapture = 0, i, direction = 0;
	// ACLK is devided by 8 = 4096 Hz. RSEL=6 no division for MCLK and SSMCLK. XT2 is off.
	BCSCTL1 = 0xB7;
	// Init FLL to desired frequency using the 32762Hz crystal DCO frquenzy = 2,4576 MHz
	BCSCTL2 = 0x00;
	DCOCTL = 0xFF;
	for ( i = 0xFFFF; i > 0; i-- );   // Delay for XTAL to settle
	CCTL2 = CCIS0 + CM0 + CAP;        // Define CCR2, CAP, ACLK
	TACTL = TASSEL1 + TACLR + MC1;    // SMCLK, continous mode
	while ( 1 ) {
		CCTL2 &= ~( CCIFG + COV );      // clear IFG and possible overflow
		// 1. capture
		while ( !( CCTL2 & CCIFG ) );           // Wait until capture occured!
		CCTL2 &= ~CCIFG;                        // Capture occured, clear flag
		Oldcapture = CCR2;                      // Save current captured SMCLK
		// 2. - 5. capture, "software divider" so aclk stays
		// untouched and 2400 bps are working!
		while ( !( CCTL2 & CCIFG ) );         // Wait until capture occured!
		CCTL2 &= ~CCIFG;                      // Capture occured, clear flag
		Compare = CCR2;
		Compare = Compare - Oldcapture;         // SMCLK difference
		if ( DELTA == Compare ) break;          // if equal, leave "while(1)"
		else {
			if ( DELTA < Compare ) {           // DCO is too fast, slow it down
				if ( direction == UP ) break;    // stop if we were going up (jumped over exact delta)
				if ( !direction ) direction = DOWN; // if starting set direction
				DCOCTL--;
				if ( DCOCTL == 0xFF ) {
					if ( !(( BCSCTL1 & 0x07 ) == 0x00 ) ) WDTCTL = 0xDEAD;
					BCSCTL1--;            // Did DCO role under? -> Select next lower RSEL
				}
			} else {
				if ( direction == DOWN ) break;   // stop if we were going down (jumped over exact delta)
				if ( !direction ) direction = UP; // if starting set direction
				DCOCTL++;
				if ( DCOCTL == 0x00 ) {
					if (( BCSCTL1 & 0x07 ) == 0x07 ) WDTCTL = 0xDEAD;
					BCSCTL1++;            // Did DCO role over? // -> Select next higher RSEL
				}
			}
		}
	}
	CCTL2 = 0;      // Stop CCR2 function
	TACTL = 0;      // Stop Timer_A
	dcoDiff = 0;
	checkDCOState = 0;
	return true;
}

/**
 * Entry point for the bootloader
 */
int main(void){
	////////////////////////////////////// boot sequence
	WDTCTL = WDTPW | WDTHOLD;			// stop watchdog timer for the initialization

	////////////////////////////////////// setup hardware ports, setup the MSB430 and the MSB430MS
	
	/*
	 * After a POR, the initial MSP430 conditions are:
	 *  - The RST/NMI pin is configured in the reset mode.
	 *  - I/O pins are switched to input mode as described in the Digital I/O chapter.
	 *  - Other peripheral modules and registers are initialized as described in their
	 *    respective chapters in this manual.
	 *  - Status register (SR) is reset.
	 *  - The watchdog timer powers up active in watchdog mode.
	 *  - Program counter (PC) is loaded with address contained at reset vector
	 *    location (0FFFEh). CPU execution begins at that address.
	 * 
	 * POR == POWER-ON RESET
	 * PUC == POWER-UP CLEAR
	 * 
	 * AWARE: different power on conditions after POR and PUC
	 * 
	 * AWARE: watchdog timer generates PUC not POR
	 * 
	 * A POR is a device reset. A POR is only generated by the following three
	 * events:
	 *  - Powering up the device
	 *  - A low signal on the RST/NMI pin when configured in the reset mode
	 *  - An SVS low condition when PORON = 1.
	 * 
	 * A PUC is always generated when a POR is generated, but a POR is not
	 * generated by a PUC. The following events trigger a PUC:
	 *  - A POR signal
	 *  - Watchdog timer expiration when in watchdog mode only
	 *  - Watchdog timer security key violation
	 *  - A Flash memory security key violation
	 * 
	 * 
	 * Unused I/O pins should be configured as I/O function, output direction, and left
	 * unconnected on the PC board, to reduce power consumption.
	 * 
	 */

	P1SEL = 0x00;
	P1OUT = 0x00;
	P1DIR = 0x00;
	P2SEL = 0x00;
	P2OUT = 0x18;
	P2DIR = 0x1A;
	P3SEL = 0x00;
	P3OUT = 0x09;
	P3DIR = 0x21;
	P4SEL = 0x00;
	P4OUT = 0x00;
	P4DIR = 0x00;
	P5SEL = 0x0E;
	P5OUT = 0xF9;
	P5DIR = 0xFD;
	P6SEL = 0x07;
	P6OUT = 0x00;
	P6DIR = 0xC8;
	
	DCO_set();
	
	//WRITE_SP(&__stack);

	// init the MSB430MS board (stop the beeper, turn off the power supplies etc...)
	_SHR_RCLK_LOW;
	_SHR_Shift8bit(0x00);
	_SHR_Shift8bit(0x00);
	_SHR_Shift8bit(0x0F);
	_SHR_RCLK_HIGH;
	
	////////////////////////////////////// Important: clock configuration
	////////////////////////////////////// 7.372 MHz are supplied via XT2: the external quartz (100uA power consumption)
	/*BCSCTL1 = RSEL2;

	do {								// wait for the xtal to stabilize:
		uint8_t i;

		IFG1 &= ~OFIFG;					// clear oscillator fault flag
		
		for (i = 0xFF; i > 0; i--);		// time required for flag the to set
	}
	while ((IFG1 & OFIFG) != 0);		// repeat until the oscillator fault flag is cleared by the MSP430
	
	BCSCTL2 = SELM_2 + SELS;			// MCLK and SMCLK == XT2
	////////////////////////////////////// done.
	*/
	// setup the clock for flash memory operations
	// added by MH (Hutta): FN4 -> FN3
	FCTL2 = FWKEY | FSSEL_SMCLK | FN3;				// flash controller clock: 7372000 / 16 = 460kHz
	  
	////////////////////////////////////// setup the ACLK to be supplied by the XT: External 32kHz quartz is attached (32768Hz)
	// --> it's preset already after POR/PUC
	////////////////////////////////////// done.

	////////////////////////////////////// initialize pheripherials
	USART_Init();						// USART (defaults to UART 115200 bps)

	////////////////////////////////////// boot sequence complete.
	
	////////////////////////////////////////// 
	_DINT();								// make sure that interrupts are disabled
											// interrupts are not used in the bootloader
	
	////////////////////////////////////////////////////////////////////////////////////////////
	////////////////////////////////////////////////////////////////////////////////////////////
	/////// bootloader logic starts here.... ///////////////////////////////////////////////////
	////////////////////////////////////////////////////////////////////////////////////////////
	////////////////////////////////////////////////////////////////////////////////////////////
	
#if DEBUGOUTPUT
	// prints the version of the bootloader
	UARTTX("\r\n[BLD V:");
	LOG(__APPVER_MAJOR + 0x30);
	LOG(__APPVER_MINOR + 0x30);
#endif
	
	////////////////////////////////////////// FIND STATUS OF THE LAST BOOTLOADER RUN
	
	// bootloader state is stored at the beginning of the infomem area
	// the data doesn't have to be copied, direct access via pointer
	bootloader_offline_state_t* bootloader_state = (bootloader_offline_state_t*) INFOMEM_START;
		
	// test the checksum of the saved bootloader state
	uint8_t bld_checksum = bootloader_state->Flashed + bootloader_state->UsingGoldenImage + __APPVER_MAJOR + __APPVER_MINOR;

	if (bld_checksum != bootloader_state->Checksum) {
		//the state is not valid -> maybe new bootloader, maybe just installed, etc.
		//recover the valid state now:
		INFOMEM_write_bootloader_state(APP_NOT_FLASHED, 0, NOT_USING_GOLDENIMAGE);
	}

	////////////////////////////////////////// FIND THE FLASHED APPLICATION AND THE VERSION
	
	//the applicationHeader of the installed application could be dynamicaly found:
	//a search would start after the bootloader program area and the program would
	//search for a valid application header on a beginning of every
	//page: this is more flexible but costs program memory (not a lot)
	//this also cost development time. in current phase of the project
	//the location of the application header is well known.
	//it can be found in the linker script of the installed application.
	_applicationHeader = (appHeader_t*) 0x4600;

	//states whether the flashed application is valid (checksum of the applicationHeader is OK + the bootloader knows the last flashing was successful)
	uint8_t _ValidApplicationFlashed = 0;

	//check the checksum of the header of the installed application
	//TELLS ME: IS IT SAFE TO RUN THE APPLICATION, IS IT SAFE TO ASSUME THE APPLICATION IS THERE
	//IMPORTANT: only the header of the application is checked, it is possible, that the rest of the
	//content is damaged (e.g. because the previous flasing operation failed)
	uint8_t _checksum = _applicationHeader->type;
	_checksum += _applicationHeader->appVer[0]; 
	_checksum += _applicationHeader->appVer[1];
	
	if (_checksum == _applicationHeader->checksum) {
		// ok--> the checksum of the application header is fine.
		//check whether the application was completly flashed:
		if (bootloader_state->Flashed) {
			_ValidApplicationFlashed = 1;
		/*
			#if DEBUGOUTPUT
				UARTTX("flashed_app:");
					LOG(_applicationHeader->appVer[0] + 0x30);
					LOG(_applicationHeader->appVer[1] + 0x30);
					LOG('|');
			#endif		
		*/			
			if (bootloader_state->FailStartsCounter > 7) {
				_ValidApplicationFlashed = 0;
				
				if (bootloader_state->UsingGoldenImage) {
					// even the golden image fails...
					#if DEBUGOUTPUT
						UARTTX("golden image failed!");
					#endif
					goto update_error;
				}
				
				#if DEBUGOUTPUT
					UARTTX("switching to golden image!");
				#endif

				// force the golden image!
				INFOMEM_write_bootloader_state(bootloader_state->Flashed, 0, USING_GOLDENIMAGE);
			}
		}
		else {
			#if DEBUGOUTPUT
				UARTTX("prev flash failed!");
			#endif		
		}
	}
	else {
		#if DEBUGOUTPUT
			UARTTX("no flashed app!");
		#endif		
	}
	
	////////////////////////////////////////// ACCESS THE SD CARD
	{
		uint8_t counter;
		for (counter = 0; counter < 10; counter++) {			//TODO: after the cold start the init has to be executed twice: this is not caused by power on time of the sd card, this was tested already - some registers are changed during the first run - this needs more time to be investigated. after the deadline.
			SD_Card_Init();
			if (SD_Card_Init_Result == SD_INIT_SUCCESS)
				break;
		}
	}
	
	if (SD_Card_Init_Result != SD_INIT_SUCCESS) {
		#if DEBUGOUTPUT
			UARTTX("SD!");
		#endif			
		VisualMessage(2);
		goto update_error;
	}
	
	if (Fat_Init(&FAT_State) != FAT_ERROR_NONE) {
		#if DEBUGOUTPUT
			UARTTX("FAT!");
		#endif
		VisualMessage(3);
		goto update_error;
	}
	
	////////////////////////////////////////// FIND THE UPDATE AND THE VERSION
	
	// find the file
	
	Fat_GetRootDir(&FAT_State, &hRootDir);				// access the root directory
	
	uint8_t _updateAvailable = false;
	
	if (bootloader_state->UsingGoldenImage) {
		// using the golden image, try to find the golden image
		_updateAvailable = Fat_FindFileSimple(&FAT_State, &hRootDir, FILENAME_GOLDENIMAGE, 0xFF, &dirEntry, &UpdateFile);

		if (!_updateAvailable) {
			// the golden image was not found
			#if DEBUGOUTPUT
				UARTTX("no golden image!");
			#endif	
			goto update_error;
		}
		#if DEBUGOUTPUT
			UARTTX("using golden image!");
		#endif			
	}
	else {
		// try to find the update, this is the default behavior
		_updateAvailable = Fat_FindFileSimple(&FAT_State, &hRootDir, FILENAME_UPDATE, 0xFF, &dirEntry, &UpdateFile);
		if (!_updateAvailable) {
			// the update was not found, try to find the golden image
			#if DEBUGOUTPUT
				UARTTX("no update!");
			#endif				
			Fat_GetRootDir(&FAT_State, &hRootDir);				// access the root directory				
			_updateAvailable = Fat_FindFileSimple(&FAT_State, &hRootDir, FILENAME_GOLDENIMAGE, 0xFF, &dirEntry, &UpdateFile);
			if (!_updateAvailable) {
				// golden image was not found neither...
				#if DEBUGOUTPUT
					UARTTX("no golden image!");
					goto update_error;
				#endif	
			}
			else {
				#if DEBUGOUTPUT
					UARTTX("using golden image|");
				#endif						
			}
		}
		else {
			#if DEBUGOUTPUT
				UARTTX("using update image|");
			#endif					
		}
	}
	
	
	// the file was found, the handle of the file is stored unter UpdateFile
	
	////////////////////////////////////////// COMPARE VERSIONS
	uint8_t minor, major;
	uint32_t _compileTime;

	if(!_ReadHexVersion(&major, &minor, &_compileTime)) {		// failed to detect version of the hex file
		#if DEBUGOUTPUT
			UARTTX("update corrupted!");
		#endif
		goto update_error;
	}
/*
#if DEBUGOUTPUT
	UARTTX("hexfile:");
		LOG(major + 0x30);
		LOG(minor + 0x30);
		LOG('|');
#endif	
*/
	if (_ValidApplicationFlashed) {
		// the node contains a valid application
		if (_compileTime == (_applicationHeader->compileTime)
			/*major == _applicationHeader->appVer[0]
			&&
			minor == _applicationHeader->appVer[1]*/) 
		{	// versions are equal --> don't flash, start the application

			goto run_application;
		}
#if DEBUGOUTPUT
		else 
			if (_compileTime > (_applicationHeader->compileTime)
					/*major == _applicationHeader->appVer[0]
					&&
					minor == _applicationHeader->appVer[1]*/) 
				{	// versions are equal --> don't flash, start the application

					UARTTX("update is newer|");
				}
	#endif
 
	}
	
	////////////////////////////////////////// FLASH THE UPDATE
	// versions are not equal or no valid application was detected on the chip. flashing is required.

	// reset the file with the udpate
	//Fat_GetRootDir(&FAT_State, &hRootDir);				// access the root directory
	
	//Fat_FindFileSimple(&FAT_State, &hRootDir, FILENAME_UPDATE, 0xFF, &dirEntry, &UpdateFile);
	File_Reset(&UpdateFile);
	
	// indicate "flashing" state -> "flashing" state means, the application on the node is not usable now
	INFOMEM_write_bootloader_state(0, 0, bootloader_state->UsingGoldenImage);
	#if DEBUGOUTPUT
		UARTTX("flashing... ");
	#endif	
		
	if (_Reprogram()) {
		//indicate the "flashed" state -> "flashed" state means, the application is save to be started
		INFOMEM_write_bootloader_state(APP_FLASHED, 0, bootloader_state->UsingGoldenImage);
		// returned function pointer to the main function
		// the function pointer could be called
		//... I just do reset now, the versions will be equal and the app will just start.
		#if DEBUGOUTPUT
			UARTTX(" ok! restart.]\r\n");
		#endif	
			
		UART_WAIT_TXDONE();
		WDTCTL = 0xDEAD;
	}
	else {
		// returned null --> failed
		#if DEBUGOUTPUT
			UARTTX(" error");
		#endif	
		goto update_error;
	}
	
	////////////////////////////////////////// RUN THE APPLICATION

	run_application:
	
		//increase the RUN counter
		INFOMEM_write_bootloader_state(APP_FLASHED, bootloader_state->FailStartsCounter + 1, bootloader_state->UsingGoldenImage);
	
		#if DEBUGOUTPUT
			LOG(bootloader_state->FailStartsCounter + 0x30);
			UARTTX(". RUN]\r\n");
		#endif
		if( _applicationHeader ) {
			((fp_t)_applicationHeader->resetAdr)();		//this never returns ("main")
		}

		
	////////////////////////////////////////// ERROR HANDLING
		
	/*
	 * something went really wrong and there is no recovery path... shut down to save energy...
	 */
	stop:
#if DEBUGOUTPUT
	UARTTX("FATAL ERROR: turning off...");
#endif	
		VisualMessage(32);
		while (1) {
			LPM4;
		}
	
	update_error:
		if (_ValidApplicationFlashed)
			goto run_application;
		else
			goto stop;

	return 0;
}


