/**
 * Copyright (c) 2008, Freie Universitt Berlin, Tomasz Naumowicz
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:
 * 
 * Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
 * Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.
 * Neither the name of the Freie Universitt Berlin nor the names of its contributors may be used to endorse or promote products derived from this software without specific prior written permission.
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * 
 * Based on the ScatterWeb firmware from the Freie Universitaet Berlin under the BSD license.  
 */



/**
 * USART control module:
 * 
 * contains the code for the UART and for the SPI mode.
 */

#include "System.h"

enum usart_mode			USART_CurrentMode;

volatile uint8_t		_usart_locked;

/**
 * UART SUPPORT
 */
#define USART_WAIT_LOCK(mode)	( (USART_CurrentMode != mode ) && (_usart_locked) )

void _USART_Reconfigure(enum usart_mode mode) {
	//wait for all buffered data to be transmitted
	UART_WAIT_TXDONE();
	
	//set the configuration:
	switch (mode) {
#if DEBUGOUTPUT
		case USART_MODE_UART: 
				////////////////////////////////////// MSB430H Specific: Select UART function
				P5OUT |= 0x01;						// unselect SPI
				P3SEL |= 0xC0;						// select rs232

				// The required USART initialization/re-configuration process is:
				// 1) Set SWRST (BIS.B #SWRST,&UxCTL)
				U1CTL = SWRST;
				// 2) Initialize all USART registers with SWRST = 1 (including UxCTL)
				U1CTL |= CHAR;          			// 8-bit character
				U1TCTL |= SSEL1;					// UCLK = SCLK

				//set the speed according to the stored configuration, fixed speed 115200
				UBR01 = 0x27;
				UBR11 = 0x00;
				UMCTL1 = 0x10;
						
				// 3) Enable USART module via the MEx SFRs (URXEx and/or UTXEx)
				ME2 |= UTXE1 | URXE1;				// Enable USART1 TXD/RXD	

				// 4) Clear SWRST via software (BIC.B #SWRST,&UxCTL)
				U1CTL &= ~SWRST;

				// 5) Enable interrupts (optional) via the IEx SFRs (URXIEx and/or UTXIEx)
				//IE2 |= URXIE1; //interrupt enable for RX
				//IFG1 &= ~URXIE1;
				
				//IE2 |= UTXIE1; //interrupt enable for TX
				//IFG1 &= ~UTXIE1;
				
			break;
#endif
		case USART_MODE_SPI:
				////////////////////////////////////// MSB430H Specific: Select UART function
				P3SEL &= ~0xC0;						// unselect rs232
	
				// The required USART initialization/re-configuration process is:
				// 1) Set SWRST (BIS.B #SWRST,&UxCTL)
				U1CTL = SWRST;
				// 2) Initialize all USART registers with SWRST = 1 (including UxCTL)
				U1CTL |= CHAR | SYNC | MM;		// 8-bit SPI Master
				UTCTL1 |= CKPH | SSEL1 | SSEL0 | STC;	// SMCLK, 3-pin mode, clock idle low, data valid on rising edge, UCLK delayed
	
				//set the speed according to the stored configuration
				UBR01 = 0x02;
				UBR11 = 0x00;
				UMCTL1 = 0x00;
						
				// 3) Enable USART module via the MEx SFRs (URXEx and/or UTXEx)
				ME2 |= USPIE1;							// Enable USART1 SPI
	
				// 4) Clear SWRST via software (BIC.B #SWRST,&UxCTL)
				U1CTL &= ~SWRST;
	
				// 5) Enable/Disable interrupts (optional) via the IEx SFRs (URXIEx and/or UTXIEx)
				//IE2 &= ~URXIE1; //disable interrupt for RX
				//IFG1 &= ~URXIE1;
				
				//IE2 |= UTXIE1; //interrupt enable for TX
				//IFG1 &= ~UTXIE1;
				
			break;
		default:
			break;
	}
	
	//re-enable interrupts
	//_EINT();
}

bool USART_Lock(enum usart_mode mode) {
	if ((USART_CurrentMode != mode ) && (_usart_locked))
		return false;										// USART is locked already in another mode
	
	// lock can be set:
	_usart_locked++;
	
	if (USART_CurrentMode != mode) {
		_USART_Reconfigure(mode);
		USART_CurrentMode = mode;
	}

	return true;
}

bool USART_LockWait(enum usart_mode mode) {
	while (USART_WAIT_LOCK(mode))
		_NOP();
	
	return USART_Lock(mode);
}

bool USART_Unlock(enum usart_mode mode) {
	if (USART_CurrentMode != mode)
		return false;
	
	if (_usart_locked > 0) {
		_usart_locked--;
		
		if (_usart_locked == 0)
			if (USART_CurrentMode != USART_MODE_UART) {
				_USART_Reconfigure(USART_MODE_UART);
				USART_CurrentMode = USART_MODE_UART;
			}
		
		return true;
	}
	
	return false;
}




/**
 * Initis the USART, sets the UART mode as default with the speed 115200.
 * This function call enables interrupts.
 */
void USART_Init() {
	//_usart_defaultMode = USART_MODE_UART;
	USART_CurrentMode = USART_MODE_UART;
	_usart_locked = 0;
	_USART_Reconfigure(USART_CurrentMode);
}
#if DEBUGOUTPUT
void UARTTX(uint8_t* line) {
	if (USART_MODE_IS(USART_MODE_UART)) {
		while (*line != '\0') {
			UART_WAIT_TX();
			UART_TX = *line;
			line++;
		}
	}
}

void UARTPUTC(uint8_t c) {
	UART_WAIT_TX(); UART_TX = (c);
}
#endif

/**
 * SPI1 support
 */

#define	SPI_idle_SYMBOL		0xFF

uint8_t SPI_rx() {
  SPI_rxTX:
	if((IFG2 & UTXIFG1) == 0) goto SPI_rxTX;
	UART_TX = SPI_idle_SYMBOL;
  SPI_rx:
	if((IFG2 & URXIFG1) == 0) goto SPI_rx;
	return(UART_RX);
}

void SPI_tx(register const uint8_t c) {
  SPI_tx:
	if((IFG2 & UTXIFG1) == 0) goto SPI_tx;
	UART_TX = c;
  SPI_txRX:
	if((IFG2 & URXIFG1) == 0) goto SPI_txRX;
	UART_RX;
}

void SPI_read(void* pDestination, const uint16_t size, const bool incDest) {
	register uint8_t* p = (uint8_t*)pDestination;
	register uint16_t i;
	SPI_READTX:
	  if((IFG2 & UTXIFG1) == 0) goto SPI_READTX;
	  for(i = size; i > 0; i--) {
		UART_TX = SPI_idle_SYMBOL;
		SPI_READRX:
		if((IFG2 & URXIFG1) == 0) goto SPI_READRX;
		*p = UART_RX;
		if(incDest) p++;
	  }
}

void SPI_write(const void* pSource, const uint16_t size, const uint8_t startToken, const bool incSource) {
 	register unsigned char* p = (unsigned char*)pSource;
 	register uint16_t i;
 	SPI_tx(startToken);
	for( i = size; i > 0; i-- ) {
		SPI_tx(*p);
		if( incSource ) p++;
	}
}

void SPI_idle(register const uint16_t clocks) {
	register uint16_t i;
	for( i = 0; i < clocks; i++ ) {
		UART_RESET_RX();
		UART_TX = SPI_idle_SYMBOL;
		UART_WAIT_RX();
	}
}

uint8_t SPI_waitToken(const uint8_t token, const uint16_t timeout) {
	uint16_t	i;
	uint8_t	rx;
	for( i = 0; i < timeout; i++ ) {
		rx = SPI_rx();
		if( rx == token )
			return 1;
	}
	return 0;
}









