/**
 * Copyright (c) 2008, Freie Universitt Berlin, Tomasz Naumowicz
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:
 * 
 * Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
 * Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.
 * Neither the name of the Freie Universitt Berlin nor the names of its contributors may be used to endorse or promote products derived from this software without specific prior written permission.
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * 
 * Based on the ScatterWeb firmware from the Freie Universitaet Berlin under the BSD license.  
 */

#include "fat.internals.h"


uint8_t get_lowest_bit_index(register int i) {
	register uint8_t bit = 0;
	
	do {
		if( i & 0x01 )
			return bit;
		i >>= 1;
		bit++;
	} while(i != 0);

	return 0;
}

enum fat_init_result Fat_Init(fat_state_t* pFS) {
	uint8_t		ret;
	uint32_t		address;
	uint16_t		BytsPerSec;
	uint32_t	BpbSA;			///< first sector of VBR
	
	struct fat_partition	part;
	struct fat_partition*	pPart = &part;
	struct fat				fatTbl;
	struct fat*				pFAT = &fatTbl;
	struct fat_bpb*			pBPB = (struct fat_bpb*)&fatTbl;

	memset(pFS, 0, sizeof(fat_state_t));

	if( sizeof(struct fat_bpb) != 25 ) {
		// if this fails, structure packing does not work correctly
		//FAT_Init_Result = FAT_STRUCT_PACKING;
		return FAT_STRUCT_PACKING;
	}

	// test for BIOS signature
	#if !SD_CACHE
		Sd_set_blocklength(1); // 2
	#endif
	ret = Sd_read(pFAT, 510, 2); //last 2 bytes from the first block (block has 512 bytes), this is too tricky, reading only 2 bytes into a bigger struct. it gets the BytsPerSec propably
#if FATDEBUGGING
	printf("BIOS signature: %.2x\r\n", ((uint16_t*)pFAT)[0]);
#endif

	// BIOS signature (found in MBR and VBR)
	if( ((uint16_t*)pFAT)[0] != FAT_BIOS_SIGNATURE ) {//ah, and now it should be the FAT BIOS SIGNATURE... hmm, why use the pFAT as a buffer then???? 
//		FAT_Init_Result = FAT_ERROR_SPEC;
		return FAT_ERROR_SPEC;
	}
	
	// read vbr and partition table from first sector and
	// decide which one exists
	#if !SD_CACHE
		Sd_set_blocklength(7); // 128
	#endif
	// read VBR
	ret = Sd_read( pFAT, FAT_BPB_OFFSET, sizeof(struct fat) ); //now it makes sense, so the use of pFAT earlier was only to save memory...		
#if FATDEBUGGING
	printf("read VBR\r\n");
#endif
	if( ret == 0 ) {
//		FAT_Init_Result = FAT_ERROR_READ;
		return FAT_ERROR_READ;
	}
	// read 1. partition table entry
	ret = Sd_read( pPart, 0x01BE, sizeof(struct fat_partition) );	
#if FATDEBUGGING
	printf("read 1. partition table\r\n");
#endif

#if FATDEBUGGING
	printf("Partition Table Entry first two bytes (swapped): %.2x\r\n", ((uint16_t*)pPart)[0]);
#endif

	if( ret == 0 ) {
//		FAT_Init_Result = FAT_ERROR_READ;		
		return FAT_ERROR_READ;
	}

	//
	// decide if MBR or VBR
	//
	ret = 0;
	// MBR with FAT16 partition
	if( (pPart->type == 4) || (pPart->type == 6) )
		ret |= 0x01;
	// sanity check for valid FAT BPB
	if( (pFAT->bpb.SecPerTrk > 0) && (pFAT->bpb.BytsPerSec > 0) && (pFAT->bpb.SecPerClus <= 64) )
		ret |= 0x04;
	
	//
	// handle MBR
	//
	if( ret == 0x01 ) {	
#if FATDEBUGGING
	printf("handling MBR\r\n");
#endif
		// is likely to be MBR
		BpbSA = pPart->distance_to_first; //BpbSA contains the CLUSTER NUMBER!!
#if FATDEBUGGING
		printf("sector: %i\r\n", BpbSA);
#endif
		address = 0 + FAT_BPB_OFFSET + (BpbSA << 9); //0 this is a source of possible failures. we have to use the CURRENT CLUSTER, but at init it is 0 so it's fine.
#if FATDEBUGGING
		printf("address: %lu\r\n", address);
#endif
		// read VBR
		ret = Sd_read( pFAT, address, sizeof(struct fat) );
		if( ret == 0 ) {
//			FAT_Init_Result = FAT_ERROR_READ;			
			return FAT_ERROR_READ;
		}
	} else {
		BpbSA = 0;
	}

	//
	// handle VBR
	//
	
	// copy BPB values (note MSB430 has issues with word which are not word-aligned...)
	uint16_t u16tmp;
	uint32_t u32tmp;
		
	pFS->RsvdSecCnt = ((uint8_t*)&pFAT->bpb)[4];
	pFS->RsvdSecCnt = pFS->RsvdSecCnt  << 8;
	pFS->RsvdSecCnt += ((uint8_t*)&pFAT->bpb)[3];


	pFS->NumFATs = pBPB->NumFATs;			// usually 2
	pFS->RootEntCnt = pFAT->bpb.RootEntCnt;
	
	if( pFAT->bpb.TotSec16 != 0 )
		pFS->TotSec = pFAT->bpb.TotSec16;
	else {
		u32tmp = ((uint8_t*)&pFAT->bpb)[24];
		u32tmp = u32tmp  << 8;
		u32tmp += ((uint8_t*)&pFAT->bpb)[23];
		u32tmp = u32tmp  << 8;
		u32tmp += ((uint8_t*)&pFAT->bpb)[22];
		u32tmp = u32tmp  << 8;
		u32tmp += ((uint8_t*)&pFAT->bpb)[21];

		pFS->TotSec = u32tmp;
		//pFS->TotSec = pFAT->bpb.TotSec32; we can't do that because MSB430 has issues with word which are not word-aligned...
	}

#if FATDEBUGGING
	printf("TotSec = %lu %lx (%x, %x, %x, %x)\r\n", pFS->TotSec, pFS->TotSec, ((uint8_t*)&pFAT->bpb)[24], ((uint8_t*)&pFAT->bpb)[23], ((uint8_t*)&pFAT->bpb)[22], ((uint8_t*)&pFAT->bpb)[21]);
#endif
	
	u16tmp = ((uint8_t*)&pFAT->bpb)[12];
	u16tmp = u16tmp  << 8;
	u16tmp += ((uint8_t*)&pFAT->bpb)[11];
	if( u16tmp != 0 )
		pFS->FATSz = u16tmp;
	else {
		#if FAT32_SUPPORT
			memcpy(&pFS->FATSz, &pFAT->vbr.fat32.FATSz32, 4);
		#else
			/// TODO: FAT32
			//FAT_Init_Result = FAT_ERROR_TYPE;			
			return FAT_ERROR_TYPE;
		#endif
	}

#if FATDEBUGGING

	printf("\r\nBytsPerSec %i\r\n"
		   "SecPerClus %i\r\n"
		   "RsvdSecCnt %i\r\n"
		   "NumFats	   %i\r\n"
		   "RootEntCnt %i\r\n",
		pFAT->bpb.BytsPerSec,
		pFAT->bpb.SecPerClus,
		pFS->RsvdSecCnt,
		pFAT->bpb.NumFATs,
		pFAT->bpb.RootEntCnt
	);
#endif

	// precalculation
	BytsPerSec = pFAT->bpb.BytsPerSec;	
	pFS->BytsPerSec_bit = get_lowest_bit_index( BytsPerSec );
	pFS->SecPerClus_bit = get_lowest_bit_index( pFAT->bpb.SecPerClus );
	pFS->BytsPerClus_bit = pFS->BytsPerSec_bit + pFS->SecPerClus_bit;
	pFS->RootDirSz = ((pFS->RootEntCnt << 5) + (BytsPerSec - 1)) >> pFS->BytsPerSec_bit;
	pFS->FatSA = BpbSA + pFS->RsvdSecCnt;
	pFS->FatA = pFS->FatSA;
	pFS->FatA = pFS->FatA << pFS->BytsPerSec_bit;
	pFS->FirstClusterSA = pFS->FatSA + (pFS->NumFATs * pFS->FATSz) + pFS->RootDirSz;

#if FATDEBUGGING

	printf("\r\RootDirSz %i\r\n"
		   "FatSA %i\r\n"
		   "FatA %lx\r\n"
		   "FATSz %i\r\n"
		   "FirstClusterSA	   %i\r\n",
		pFS->RootDirSz,
		pFS->FatSA,
		pFS->FatA,
		pFS->FATSz,
		pFS->FirstClusterSA
	);
#endif
	
	#if FAT_WRITE
		pFS->LastFreeCA = 0;
	#endif
	
	pFS->FATType = _Fat_GetType(pFS, pFAT);
	#if !FAT12_SUPPORT
		if( pFS->FATType & FAT12 ) {
			//FAT_Init_Result = FAT_ERROR_TYPE;
			return FAT_ERROR_TYPE;
		}
	#endif
	#if !FAT32_SUPPORT
		if( pFS->FATType & FAT32 ) {
			//FAT_Init_Result = FAT_ERROR_TYPE;
			return FAT_ERROR_TYPE;
		}
	#endif

	// first sector of root directory
	#if FAT32_SUPPORT
		if( pFS->FATTYPE & FAT32 )
			pFS->RootDirCA = pFAT->vbr.fat32.RootClus;
		else
			pFS->RootDirCA = 1;
	#else
		pFS->RootDirCA = 1;
	#endif
	
	//FAT_Init_Result = FAT_ERROR_NONE;
	return FAT_ERROR_NONE;
}



void _Fat_getClusterFAT(const fat_state_t* pFS, fat_ca_t* pCA, uint32_t* pAddress) {
	//
	// calculate FAT entry for cluster
	//

	*pAddress = *pCA;
	#if FAT12_SUPPORT
		if( pFS->FATType == FAT12 )
			*pAddress += *pCA >> 1;
		else
			*pAddress <<= pFS->FATType;
	#else
		*pAddress <<= pFS->FATType;
	#endif
	*pAddress += pFS->FatA;	
}

void _Fat_getClusterPos(const fat_state_t* pFS, const fat_ca_t* pCA, uint32_t* pAddress) {
	uint32_t temp = (uint32_t) *pCA;
	temp= temp- FAT_CLUSTER_OFFSET;
	temp= temp<< pFS->SecPerClus_bit;
	*pAddress =  temp + pFS->FirstClusterSA; //bug, this line had an error!!!

	//*pAddress <<= pFS->BytsPerClus_bit; //this was a bug:
#if FATDEBUGGING
	printf("\r\n_Fat_getClusterPos: cluster: %u starts in sector: %lu ", *pCA,  *pAddress);
#endif
	*pAddress <<= pFS->BytsPerSec_bit;  
#if FATDEBUGGING
	printf("and has address %lu\r\n", *pAddress);
#endif

}

bool _fat_getNextCA(const fat_state_t* pFS, fat_ca_t cluster, fat_ca_t* pNextCluster) {
	uint32_t	address;	
	uint8_t 	bl;
	bool			ret;		// mult-purpose response buffer

	_Fat_getClusterFAT( pFS, &cluster, &address );
	//
	// read FAT
	//
	bl = FAT_ENTRY_WIDTH(pFS);
	ret = Sd_read( pNextCluster, address, 1 << bl );
	RETF(ret);

   #if FAT12_SUPPORT
	if( pFS->FATType == FAT12 ) {
		if( cluster & 0x0001 )
			*pNextCluster >>= 4;
		else
			*pNextCluster &= 0x0FFF;

		if( *pNextCluster > FAT_BAD12 )
			return false;
	}
   #endif

   #if FAT32_SUPPORT
   	if( pFS->FATType == FAT32 ) {
   		if( *pNextCluster > FAT_BAD32 )
	   		return false;
   	} else
   #endif
	if( *pNextCluster > FAT_BAD16 )
		return false;
	
	return true;
}
						  
/**
 * This function returns the FAT type 12/16/32. That a valid FAT table
 * was found and does no additional validation (although this would be a
 * good place, if needed).
 */
FAT_TYPE_T _Fat_GetType(const fat_state_t* pFS, struct fat* pFAT) {
	char*	name = pFAT->vbr.fat16.fat_name;

	#if FAT_STRICT_TYPE_CHECK
		uint32_t		DataSec;
		fat_ca_t	ClusterCount;
		//
		// Calculate FAT type according to Microsoft white paper
		//
		DataSec = pFS->TotSec;
		DataSec -= pFS->RsvdSecCnt
		           + (pFS->NumFATs * pFS->FATSz)
		           + pFS->RootDirSz;
		
		// number of clusters in the data region
		ClusterCount = DataSec >> pFS->SecPerClus_bit;
		if( ClusterCount < 4085 ) {
			// not all devices format coforming to specification
			if( name[3] == '6' )
				return FAT16;	// ill-sized FAT16
			return FAT12;
		} else if( ClusterCount < 65525 )
			return FAT16;
		else
			return FAT32;
	#else
		// usually works, but is not reliable
		if( name[3] == '1' ) {
			if( name[4] == '6' )
				return FAT16;
			else if( name[4] == '2' )
				return FAT12;
		}
		return FAT32;
	#endif
}


