/**
 * Copyright (c) 2008, Freie Universitt Berlin, Tomasz Naumowicz
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:
 * 
 * Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
 * Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.
 * Neither the name of the Freie Universitt Berlin nor the names of its contributors may be used to endorse or promote products derived from this software without specific prior written permission.
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * 
 * Based on the ScatterWeb firmware from the Freie Universitaet Berlin under the BSD license.  
 */


#include "fat.internals.h"

/**
 * @name Directory navigation
 */
void Fat_GetRootDir(const fat_state_t* pFS, file_handle_t* hRoot ) {
	hRoot->StartCA = pFS->RootDirCA;
	hRoot->Flags = FILE_HANDLE_FLAG_ROOTDIR;
	hRoot->Size = pFS->RootDirSz;
	hRoot->Size <<= pFS->BytsPerSec_bit;
	File_Reset(hRoot);	
}

bool Fat_GetNextDirEntry(
		const fat_state_t*	pFS, 
		file_handle_t*		hParent,
		uint32_t*				pAddress,
		struct fat_direntry*		resultDirEntry)
{
	uint8_t			ret;
	struct fat_direntry  dirEntry;
	uint8_t*			nameFlag;
	
	if( hParent->Flags & FILE_HANDLE_FLAG_EOF )
		return false;
	
	if( *pAddress == 0 ) {
		// first call
		*pAddress = _Fat_GetFilePos(pFS, hParent);	//this is ok
		*pAddress -= FAT_DIRENTRY_SIZE; //required for proper loop init
	}
	nameFlag = (uint8_t*)&dirEntry;
	
	while(1) {
		// Test EOF
		if( hParent->Flags & FILE_HANDLE_FLAG_EOF )
			break;

		if(pAddress)
			*pAddress += FAT_DIRENTRY_SIZE; //advance here

		// read
		#if !SD_CACHE
			Sd_SetBlockLength(5); // 32
		#endif
		ret = Sd_read( &dirEntry, *pAddress, FAT_DIRENTRY_SIZE ); //ok, so the first entry is being read
		RETF(ret);
	
		// advance
		File_Seek( pFS, hParent, ret, NULL ); //now, we skip to the next entry, the Address is being modified here!
	
		if( *nameFlag == FAT_NAME_FLAG_DELETED )
			continue;
		if( (dirEntry.Attributes & FAT_ATTR_LONG_NAME) == FAT_ATTR_LONG_NAME )
			continue;
		if( *nameFlag == FAT_NAME_FLAG_END )	
			break;

		// result
		if( resultDirEntry )
			memcpy( resultDirEntry, &dirEntry, sizeof(dirEntry) ); // and here the result is retuned, but the Address was already modified, and this is not good

		return true;
	}
	
	hParent->Flags |= FILE_HANDLE_FLAG_EOF;
	return false;
}

void Fat_DirEntryToHandle(file_handle_t* hDir, uint32_t* pAddress, struct fat_direntry* dirEntry) {
	hDir->DirEntryPos = *pAddress; 
	hDir->StartCA = dirEntry->FstClusLO;
	#if FAT32_SUPPORT
		memcpy((uint8_t*)&hDir->StartCA, dirEntry->FstClusHI, 2);
		//hDir->StartCA |= (dirEntry->FstClusHI << 16);
	#endif
	hDir->Size = dirEntry->FileSize;	
	hDir->Flags = FILE_HANDLE_FLAG_NONE;
	if( dirEntry->Attributes & FAT_ATTR_DIRECTORY )
		hDir->Flags |= FILE_HANDLE_FLAG_DIRECTORY;	
	File_Reset(hDir);
}



#if FAT_CHANGEDIR
	/**
	 * @param	hCurDir		Pointer to file handle structure which holds the current directory
	 * 						(initialize to Root directory before first call)
	 * @param	name		Pointer to character string of up to 11 chars with the name of
	 * 						the directory to which to change.
	 * 						@see Fat_FindFile for syntax of directory name
	 * 
	 * ChangeDir function mimicks the well know DOS shell CD command.
	 * The current directory pointer is modified every time, when the directory
	 * was successfully to change to has been found.
	 * 
	 * This function is a wrapper for Fat_FindFile adding dot and dotdot directory handling.
	 */
	bool Fat_ChangeDir(fat_state_t*	pFS, file_handle_t* hCurDir, char* name) {
		struct fat_direntry	dirEntry;
		
		File_Reset(hCurDir);
		if( name[0] == '.' ) {			
			if( name[1] == '.' ) {
				uint32_t	address;

				File_Seek( pFS, hCurDir, FAT_DIRENTRY_SIZE, NULL );				
				address = _Fat_GetFilePos(pFS, hCurDir);

				#if !SD_CACHE
					Sd_set_blocklength(5); // 32
				#endif
				Sd_read( &dirEntry, address, FAT_DIRENTRY_SIZE );
				Fat_DirEntryToHandle( hCurDir, &address, &dirEntry );
			}
			return true;
		}
		return Fat_FindFileSimple( pFS, hCurDir, name, FAT_ATTR_DIRECTORY, &dirEntry, hCurDir );
	}
#endif
///@}
