/**
 * Copyright (c) 2008, Freie Universitt Berlin, Tomasz Naumowicz
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:
 * 
 * Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
 * Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.
 * Neither the name of the Freie Universitt Berlin nor the names of its contributors may be used to endorse or promote products derived from this software without specific prior written permission.
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * 
 * Based on the ScatterWeb firmware from the Freie Universitaet Berlin under the BSD license.  
 */



#include "fat.internals.h"

/**
 * Reads a number of bytes from a file given by a handle.
 * Can read acorss sector boundaries.
 * Can read acorss cluster boundaries.
 * 
 * @param	hFile		pointer to handle to file to read from
 * @param	pBuffer		buffer to read into
 * @param	count		number of bytes to read
 * @return	UINT16	number of bytes read
 * 
 * @see Fat_FindFile_Simple
 */
uint16_t Fat_ReadFile(
	const fat_state_t*	pFS, 
	file_handle_t*		hFile,
	void*				pBuffer,
	uint16_t				count
) {
	uint16_t ret = 0;
	uint16_t readBytesCount = 0;
	uint32_t address;
	char*    p;				// pointer to current pos in receive buffer

	p = (char*)pBuffer;

	uint16_t bpc = 1 << pFS->BytsPerClus_bit; //bytes per cluster

	do {
		address = _Fat_GetFilePos( pFS, hFile ); //gets the absolute address of the data, based on the current position in the file

		if (hFile->ClusterOffset + count <= bpc) { //if the data to read is contained in the current...
			ret = Sd_read( p, address, count );    //... then just read this data and store it in the buffer
		}
		else
		{  //otherwise read only the part contained in the current cluster
			ret = Sd_read( p, address, bpc - hFile->ClusterOffset ); 
		}

		//now update the state of the file, this function udates the position in the file and current cluster
		File_Seek( pFS, hFile, ret, NULL );

		count = count - ret;

		p += ret; //update current position in the read buffer

		readBytesCount += ret;
	}
	while (count > 0);

	return readBytesCount;
}


#if FAT_WRITE

/**
 * Writes a number of bytes to a file given by a handle.
 * Can write across sector boundaires.
 * 
 * @param	hFile		pointer to handle to file to read from
 * @param	pBuffer		buffer to read into
 * @param	count		number of bytes to read
 * @return	UINT16	number of bytes read
 * 
 * @see Fat_FindFile_Simple
 */
uint16_t Fat_WriteFile(
	const fat_state_t*	pFS, 
	file_handle_t*		hFile,
	void*				pBuffer,
	uint16_t				count
) {

	//struct fat_direntry	DirEntry; //might be required if the size of the file has to be modified

	uint16_t ret = 0;
	uint16_t writtenBytesCount = 0;
	uint32_t address;
	char*    p;				// pointer to current position in the write buffer

	p = (char*)pBuffer;

	uint16_t bpc = 1 << pFS->BytsPerClus_bit; //bytes per cluster

	//there is always at least one cluster available for storing data. even if the file has size == 1 byte, it has to use
	//one cluster anyway. it is important to check at the beginning, if a reservation of new claster is required.
	//if yes, it should be reserved then.
	//the last step is the update of the filesize of the file on disk

	//afterwards a write across clusters should be realised.


	//allocate clusters and update the size of the file if it's required 
	if (hFile->Position + count > hFile->Size) {
		//the size of the file has to be increased. it's possible, that new clusters have to be allocated
#if FATDEBUGGING
		printf("Fat_WriteFile bef res: StartCA %i, CurrentCA %i, Position %lu\r\n", hFile->StartCA, hFile->CurrentCA, hFile->Position);
#endif

		Fat_FileResize( pFS, hFile, hFile->Size + count);
		
#if FATDEBUGGING
		printf("Fat_WriteFile after res: StartCA %i CurrentCA %i, Position %lu\r\n", hFile->StartCA, hFile->CurrentCA, hFile->Position);
#endif
	}

	//write across clusters (clusters are already reserved)
	//
	do {
		address = _Fat_GetFilePos( pFS, hFile ); //gets the absolute address for the data, based on the current position in the file

		if (hFile->ClusterOffset + count <= bpc) { //if the data to write will fit into the current cluster
			ret = Sd_write( address, p, count );    //... then just write the data (no write across clusters is required)
		}
		else
		{  //otherwise write only the part that fits inside of the cluster
			ret = Sd_write( address, p, bpc - hFile->ClusterOffset ); 
		}

		//now update the state of the file, this function udates the position in the file and current cluster
		File_Seek( pFS, hFile, ret, NULL );

		count = count - ret;

		p += ret; //update current position in the write buffer

		writtenBytesCount += ret;
	}
	while (count > 0);

	Sd_flush(); //make sure everything was stored on the card.

	return writtenBytesCount;
}

#endif

/**
 * @name Handle handling
 */
/*
void File_Close( file_handle_t* hFile ) {
	memset(hFile, 0, sizeof(file_handle_t));	// just like new
}
*/
void File_Reset( file_handle_t* hFile ) {
	hFile->CurrentCA = hFile->StartCA;
	hFile->Position = 0;
	hFile->ClusterOffset = 0;
	hFile->Flags &= ~FILE_HANDLE_FLAG_EOF;
}
	
#if FAT_FILE_HELPER
	
	void Fat_Date(fat_date_t* date, uint16_t year, uint8_t month, uint8_t day) {
		*date = (year - 1980) & 0x7F;
		*date <<= 4;
		*date += month & 0x0F;
		*date <<= 5;
		*date += day & 0x1F;	
	}
	void Fat_Time(fat_time_t* time, uint8_t hours, uint8_t minutes, uint8_t seconds) {
		*time = hours & 0x1F;
		*time <<= 6;
		*time += minutes & 0x3F;
		*time <<= 5;
		*time += (seconds >> 1) & 0x1F;
	}
	
	void Fat_CanonicalNameToFAT(char* fatName, register const char* fileName) {
		register int	j;
		register char*	f = fatName;
		memset(fatName, 0x20, 12);
		for( j = 0; j < 11; fileName++, f++, j++ ) {
			if( *fileName == 0 )
				return;
			else if( *fileName == '.' )
				f = fatName + 7;
			else if( *fileName > 0x20 )
				*f = *fileName;
		}
	}	
	
	#if FAT_FIND_SIMPLE
		bool Fat_FileExists(
			const fat_state_t* pFS,
			file_handle_t* hParent,
			const char* fileName
		) {
			char			fatName[11];
			struct fat_direntry	dirEntry;
			
			Fat_CanonicalNameToFAT(fatName, fileName);
			File_Reset(hParent);
			return Fat_FindFileSimple(pFS, hParent, fatName, 0xFF, &dirEntry, NULL);
		}	
	#endif
#endif

uint32_t _Fat_GetFilePos(const fat_state_t* pFS, const file_handle_t* hFile) {
	uint32_t address;

	if( FILE_HANDLE_INVALID(hFile) )
		return 0; 

	//
	// WORD sector adress calc
	//
	if( hFile->StartCA == 1 ) {
		// FAT16 Root Dir (comes before "first cluster")
		// for FAT32 RootDirSz is always 0
		address = pFS->FirstClusterSA - pFS->RootDirSz;		// absolute SA patch for fat16 root dir

		address <<= pFS->BytsPerSec_bit;
	} else {
		// cluster to relative sector address
		_Fat_getClusterPos(pFS, &hFile->CurrentCA, &address);
	}
	//
	// DWORD byte address
	// 
	address += hFile->ClusterOffset;	// add offset (+ allows for more than one cluster offset)

#if FATDEBUGGING
	printf("_Fat_GetFilePos: %lu\r\n", address);
#endif

	return address;
}

void File_Seek(
	const fat_state_t* pFS,
	file_handle_t* hFile,
	const uint32_t count,
	unsigned long *pAddress
) {
	uint32_t	offset;
	uint16_t	bpc = 1 << pFS->BytsPerClus_bit;
	
	//
	// Check for EOF
	//
	hFile->Position += count;
	if( hFile->Flags & FILE_HANDLE_FLAG_DIRECTORY ) {
		if( hFile->StartCA == 1 ) {
			// FAT12/16 root directory
			hFile->ClusterOffset += count;
			if( hFile->Position > hFile->Size )
				goto _FatIncFilePos_EOF;
			else {
				if(pAddress)
					*pAddress += count;
				return;
			}
		}
	} else {
		// Normal file
		if( hFile->Position > hFile->Size ) {
			goto _FatIncFilePos_EOF;
		}
	}
	
	//
	// Medium position
	//
	offset = count + hFile->ClusterOffset;
	if( offset >= bpc ) {
		fat_ca_t nextCA;
		do {	
			if( _fat_getNextCA( pFS, hFile->CurrentCA, &nextCA ) ) {
				hFile->CurrentCA = nextCA;
				offset -= bpc;
			} else {
				goto _FatIncFilePos_EOF;
			}
		} while( offset >= bpc );

		hFile->ClusterOffset = *((uint16_t*)&offset);
		if( pAddress )
			*pAddress = _Fat_GetFilePos(pFS, hFile);

	} else {
		hFile->ClusterOffset = *((uint16_t*)&offset);
		if( pAddress )
			*pAddress += count;
	}
	return;
	
 _FatIncFilePos_EOF:
	hFile->Flags |= FILE_HANDLE_FLAG_EOF;
}


///@}
