/**
 * Copyright (c) 2008, Freie Universitt Berlin, Tomasz Naumowicz
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:
 * 
 * Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
 * Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.
 * Neither the name of the Freie Universitt Berlin nor the names of its contributors may be used to endorse or promote products derived from this software without specific prior written permission.
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * 
 * Based on the ScatterWeb firmware from the Freie Universitaet Berlin under the BSD license.  
 */


#ifndef FAT_INTERNALS_H_
#define FAT_INTERNALS_H_

#include "System.h"

#ifdef FATDEBUGGING
#define RETF(x)		if( ! x )  {return false;} ;
#define RETT(x)		if( x )  {return false;} ;
#else
#define RETF(x)		if( ! x )  return false;
#define RETT(x)		if( x )  return false;
#endif

#define	MIN(x, y)	( x < y ) ? x : y 

#ifdef __TI_COMPILER_VERSION__
	#pragma pack(1)
#endif

/**
 * @brief MBR Partition entry
 */
struct fat_partition {
	uint8_t		state;
	uint8_t		firsthead;
	uint16_t	firstsector		:  6;
	uint16_t	firstcylinder	: 10;
	uint8_t		type;
	uint8_t		lasthead;
	uint16_t	lastsector		:  6;
	uint16_t	lastcylinder	: 10;
	uint32_t	distance_to_first;
	uint32_t	numsectors;
} // 16 bytes
#if defined(__MSP430GCC__)
	__attribute__((packed))
#endif
;

/**
 * @brief FAT BIOS Parameter Block
 * Common for all FAT types
 */
struct fat_bpb {
	uint16_t	BytsPerSec;//01
	uint8_t		SecPerClus;//2
	uint16_t	RsvdSecCnt;//34
	uint8_t		NumFATs;//5
	uint16_t	RootEntCnt;//67
	uint16_t	TotSec16;//89
	uint8_t		Media;//10.
	uint16_t	FATSz16;//11.12.
	uint16_t	SecPerTrk;//13.14
	uint16_t	NumHeads;//15.16
	uint32_t	HiddSec;//17.18.19.20
	uint32_t	TotSec32;//21.22.23.24
} // 25 bytes
#if defined(__MSP430GCC__)
	__attribute__((packed))
#endif
;

/**
 * @brief FAT BIOS Parameter Block Extension for FAT12 and FAT16
 */
struct fat16_bpbext {
	// ignore boot sector
	//  char		jump_code[3];
	//  char		oem_name[8];
	// BIOS parameter block
	// fat16 specific
	uint16_t	logical_drive_number;
	uint8_t		ext_signature;
	uint32_t	serial_number;
	char		volume_name[11];
	char		fat_name[8];
}
#if defined(__MSP430GCC__)
	__attribute__((packed))
#endif
;

/**
 * @brief FAT BIOS Parameter Block Extension for FAT32
 * @todo: for fat32 support
 */
struct fat32_bpbext {
	// ignore boot sector
	//  char		jump_code[3];
	//  char		oem_name[8];
	// BIOS parameter block
	// fat32 specific
	char	raw[46];
	char	fat_name[8];
}
#if defined(__MSP430GCC__)
	__attribute__((packed))
#endif
;

/**
 * @brief FAT Structure
 * FAT12, FAT16, FAT32
 */
struct fat {
	struct 		fat_bpb				bpb;
	union {
		struct 	fat16_bpbext		fat12;
		struct 	fat16_bpbext		fat16;
#if FAT32_SUPPORT
		struct 	fat32_bpbext		fat32;
#endif
	} vbr;
}
#if defined(__MSP430GCC__)
	__attribute__((packed))
#endif
;

#ifdef __TI_COMPILER_VERSION__
	#pragma pack()
#endif

#define FAT_BIOS_SIGNATURE		0xAA55		///< A multi-purpose magical number
#define FAT_EOC12				0x0FF8		///< FAT12 last cluster symbol
#define FAT_EOC16				0xFFF8		///< FAT16 last cluster symbol
#define FAT_EOC32				0x0FFFFFF8	///< FAT32 last cluster symbol
#define FAT_BAD12				0x0FF7
#define FAT_BAD16				0xFFF7
#define FAT_BAD32				0x0FFFFFF7
#define FAT_FREE				0
#define	FAT_BPB_OFFSET			11			///< offset from VBR start to bpb in bytes
#define	FAT_CLUSTER_OFFSET		2			///< reserved entries at beginning of FAT

	
/**
 * @name Private Interface
 */
/// Returns 0-based number of lowest bit set
uint8_t get_lowest_bit_index(int i);

/// Convert file handle to absolute address
uint32_t _Fat_GetFilePos(const fat_state_t* pFS, const file_handle_t* hFile);

/// Calculate FAT type (the specified way)
FAT_TYPE_T _Fat_GetType(const fat_state_t* pFS, struct fat* pFAT);

/// Calculate byte address of fat entry from cluster address
void _Fat_getClusterFAT(const fat_state_t* pFS, fat_ca_t* pCA, uint32_t* pAddress);
/// Calculate byte address from cluster address 
void _Fat_getClusterPos(const fat_state_t* pFS, const fat_ca_t* pCA, uint32_t* pAddress);
bool _fat_getNextCA(const fat_state_t* pFS, fat_ca_t cluster, fat_ca_t* pNextCluster);

/// Test a file handle for validity
#define FILE_HANDLE_INVALID(x)	( x ->StartCA == 0)

#if FAT_WRITE
	bool _Fat_GetChain(fat_state_t* pFS, fat_ca_t* pFirstCA, uint16_t count);
	bool _Fat_Update(fat_state_t* pFS, fat_ca_t cluster, fat_ca_t nextCluster);
	bool Fat_FileResize(fat_state_t* pFS, file_handle_t* hFile, unsigned long size);
#endif

///@}

#endif /*FAT_INTERNALS_H_*/
