/**
 * Copyright (c) 2008, Freie Universitt Berlin, Tomasz Naumowicz
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:
 * 
 * Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
 * Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.
 * Neither the name of the Freie Universitt Berlin nor the names of its contributors may be used to endorse or promote products derived from this software without specific prior written permission.
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * 
 * Based on the ScatterWeb firmware from the Freie Universitaet Berlin under the BSD license.  
 */


#include "fat.internals.h"

#if FAT_WRITE
/**
 * Write buffered block with FAT table data back to memory
 * (FAT table uses multiple copies)
 */
bool _Fat_WriteFATBlock(const fat_state_t* pFS, sd_cache_t* pBuf) {

#if FATDEBUGGING
	printf("[_Fat_WriteFATBlock]");
#endif


	uint16_t ret;
	uint32_t fatSize = pFS->FATSz;
	int i = 0;
	
	fatSize <<= pFS->BytsPerSec_bit;
	Sd_set_blocklength(9);
	do {
		ret = Sd_write_block( pBuf->address, pBuf->buffer);
		RETT(ret);
		i++;
		if( i >= pFS->NumFATs ) {
			pBuf->state &= ~SD_CACHE_DIRTY;
			//this is wrong: the address of the cache changed and...
			//well, actually - the data was stored there so it's equal == it's not dirty.. well, ok ;)
			SD_FREE_LOCK(pBuf);
			return true;
		}
		pBuf->address += fatSize;
	} while(1);
}
bool Fat_EraseCluster(fat_state_t* pFS, fat_ca_t cluster) {
	uint8_t	BlPerClus;
	uint16_t	ret;
	uint32_t	address;
	
	_Fat_getClusterPos( pFS, &cluster, &address );
	BlPerClus = 1 << (pFS->BytsPerClus_bit - 9);
	Sd_set_blocklength(9);
	do {
		ret = Sd_set_block( address, 0 );
		BlPerClus--;
	} while( BlPerClus );
	
	return( ret > 0 );
}

bool Fat_EraseChain(fat_state_t* pFS, fat_ca_t firstCA) {
	fat_ca_t	cluster = firstCA;
	bool		ret;
	
	do {
		ret = Fat_EraseCluster( pFS, cluster );
		RETF(ret);
	} while(_fat_getNextCA( pFS, cluster, &cluster ));
	
	return true;
}

/**
 * Create a chain of count clusters starting and return first of these.
 * @param[in/out]	pFirstCA	Pointer to buffer for number of first cluster of chain
 * @param[in]		count		Length of chain to create
 * @return			True if successful
 */
bool _Fat_CreateChain(fat_state_t* pFS, fat_ca_t* pFirstCA, uint16_t count) {
	//fat_ca_t		lastCA = pFS->TotSec << pFS->SecPerClus_bit; //as number of clusters, I guess we should shift right
	uint32_t u32tmp = pFS->TotSec;
	u32tmp = u32tmp >> pFS->SecPerClus_bit; //fixed

	fat_ca_t		lastCA =  (uint16_t) u32tmp; 

	fat_ca_t		prevCA = *pFirstCA;
	uint8_t			bl;
	uint32_t			blockAdr, tmpAdr;
	uint16_t			entriesPerBlock;
	uint16_t			i;
	fat_ca_t		thisCA = pFS->LastFreeCA;
	sd_cache_t*		pBuf;
	
	#if FAT12_SUPPORT
		if( pFS->FATType & FAT12 )
			return false;
	#endif

	if (count == 0)
		return true;

#if FATDEBUGGING
		printf("_Fat_CreateChain: prevCA %i pFirstCA %i\r\n", prevCA, *pFirstCA);
#endif
	bl = FAT_ENTRY_WIDTH(pFS);
	
	blockAdr = pFS->FatSA;
	blockAdr <<= pFS->BytsPerSec_bit;
	//blockAdr += pFS->LastFreeCA >> (9 - bl);
	entriesPerBlock = SD_WRITE_BLOCKLENGTH >> bl;
	thisCA = 0; //pFS->LastFreeCA; //lets search always from the start --> reason -> the loop init for(i=0; i<entriesPerBlock... is not correct, thisCA can't be ++ with i, they can easily get out of sync
	*pFirstCA = 0; //first new claimed cluster

	do {
		pBuf = Sd_cache_read_block(&blockAdr);
#if FATDEBUGGING
		printf("_Fat_CreateChain read from %lu and success %i\r\n", blockAdr, pBuf);
		printf("lastCA: %u, pFS->TotSec %lu,  pFS->SecPerClus_bit %i\r\n", lastCA, pFS->TotSec,  pFS->SecPerClus_bit);
#endif

		RETF(pBuf);

		for( i = 0; i < entriesPerBlock; i++, thisCA++ ) {
			//check if this is free --> soooo... not good, it just starts looking from where?
			//ok, good, from the LastFreeCA (find out who defines that, the fat_init sets it to 0)
			if( ((FAT16_SEC*)pBuf->buffer)[i] == 0 ) { //is free - this check was not enough - with FAT16 e.g. we have to check 2 i and i+1, hmm.. he does that because he converts to UINT16 --> but now it's not useable for file systems that are not FAT16
				// return first new claimed cluster
				if( *pFirstCA == 0 )
					*pFirstCA = thisCA;

#if FATDEBUGGING
		printf("_Fat_CreateChain after ifpfirstca0: prevCA %i pFirstCA %i, i %i. have to link from prevCA to pFirstCA.\r\n", prevCA, *pFirstCA, i);
#endif

				// link last cluster
				if( prevCA ) {
					/*
					if( prevCA < pFS->LastFreeCA ) { //the claimed is somewhere in the middle of the file allocation table
						SD_FREE_LOCK(pBuf);
						_Fat_Update(pFS, prevCA, thisCA);
						pBuf = Sd_cache_read_block(&blockAdr);
					} else { // I don't understand this part in the else-block... so when the prevCA is greater equal lastfreeCA..
						//ok, this is to optimize memory access, but still. this might fail -> when the addresses are too far away.
						prevCA -= pFS->LastFreeCA; // but why do I substract in here... hmm..
						((FAT16_SEC*)pBuf->buffer)[prevCA] = thisCA; //ok, this is a change in memory
						pBuf->state |= SD_CACHE_DIRTY;
					}
					*/

#if FATDEBUGGING
		printf("will start linking.\r\n");
#endif
					SD_FREE_LOCK(pBuf);
					_Fat_Update(pFS, prevCA, thisCA);

					//here the cache is usualy 'dirty'
					//now -- reading, if the cache is 'dirty', it has to be stored on the card first

					//it might be, that the updated cluster entry in the FAT happened SOMEWHERE right now!!!
					//and it wasn't duplicated... this is not good!!!, this should happen after every FAT_Update
					//this should be happening within the FAT UPDATE, because we don't know where it updates...

					//let's force it here:
					_Fat_getClusterFAT( pFS, &prevCA, &tmpAdr );//now we know where the update happened (we have the address)
					//hmm.. now the data has to be aligned...
					Sd_AlignAddress(&tmpAdr); //now it's aligned... uff... 
					pBuf = Sd_cache_read_block(&tmpAdr); //now we read the data into the buffer
					_Fat_WriteFATBlock(pFS, pBuf); //now the FAT entries will be duplicated

					//ok, and now we continue the regular processing
					pBuf = Sd_cache_read_block(&blockAdr);
				}

				count--;
				if( count ) {
					prevCA = thisCA;
				} else {
					// end of chain
					i <<= bl;
					bl <<= 1;
					memset(&pBuf->buffer[i], 0xFF, bl);
					//the cache is dirty now!
					pBuf->state |= SD_CACHE_DIRTY;
#if FATDEBUGGING
		printf("_Fat_CreateChain writing FF to i %i\r\n", i);
#endif
					_Fat_WriteFATBlock(pFS, pBuf);
					return true;
				}
			} //if free
		}//loop over clusters

		//hmm -> the chain was not ended closed -> duplicate the FAT in this block and proceed
		if (pBuf->state & SD_CACHE_DIRTY)
			_Fat_WriteFATBlock(pFS, pBuf);

		blockAdr += SD_WRITE_BLOCKLENGTH;
		pFS->LastFreeCA = thisCA;
	} while( pFS->LastFreeCA < lastCA );

	
	return false;
}

/**
 * Update FAT table entry cluster to point to nextCluster
 * \note Does not support 12 bit FAT
 * \todo FAT32 support
 */
bool _Fat_Update(fat_state_t* pFS, fat_ca_t cluster, fat_ca_t nextCluster) {
	uint32_t		address;
	uint16_t		ret;
	
	//						//cluster==cluster index/number, it gets translated to real position (eg. *2 with FAT16)
	_Fat_getClusterFAT( pFS, &cluster, &address );

	ret = Sd_write(address, &nextCluster, sizeof(fat_ca_t));
	//this is very good - we update the correct location - how to duplicate the entry now...

#if FATDEBUGGING
	printf("Linking cluster %i to %i\r\n", cluster, nextCluster);
	printf("Have to write %i to %lx\r\n", nextCluster, address);
	printf("Result: %i\r\n", ret);
#endif

	return( ret != 0 );
}

bool _Fat_FileAppendCluster(fat_state_t* pFS, file_handle_t* hFile, uint16_t count) {
	if( _Fat_CreateChain(pFS, &hFile->CurrentCA, count) ) {
		hFile->Flags &= ~FILE_HANDLE_FLAG_EOF;
		hFile->ClusterOffset = 0;
		return true;
	}
	return false;
}

bool Fat_FileResize(fat_state_t* pFS, file_handle_t* hFile, unsigned long size) {
	uint32_t		positionInTheFile; //has to be restored after the resize operation
	fat_ca_t		newNumClus;
	fat_ca_t		curNumClus;
	uint32_t		address;
	bool			ret;
		
	positionInTheFile = hFile->Position;

	File_Reset(hFile);

	//530 ==                    1000010010
	//17000 ==             100001001101000
	//36000 ==            1000110010100000
	//bytespersecbit            1000000000
	//secperclustbit       100000


	//intro: imagine filesize == 32
	//32 / 16kb == 2 sectors: the file uses two sectors
	//fielesize 31 -> 31/16 = 1, but the file still is using 2 sectors. an 1 has to be added.
	//but only not always, that's why the check: if clusters * 16kb == filesize then nothing should be done
	//an 1 should be added otherwise

	//number of clusters currently allocated
	curNumClus = (hFile->Size >> (pFS->BytsPerSec_bit + pFS->SecPerClus_bit));
	uint32_t tmp = curNumClus;
	if (tmp << (pFS->BytsPerSec_bit + pFS->SecPerClus_bit) != hFile->Size)
		curNumClus++;

	//number of clusters that should be allocated for the new size
	newNumClus = (size >> (pFS->BytsPerSec_bit + pFS->SecPerClus_bit));
	tmp = newNumClus;
	if (tmp << (pFS->BytsPerSec_bit + pFS->SecPerClus_bit) != size)
		newNumClus++;

		
	if (hFile->Size == 0)
		curNumClus = 0;

	if (size == 0)
		newNumClus = 0;

#if FATDEBUGGING
	printf("Fat_FileResize: cur %i new %i\r\n", curNumClus, newNumClus);
	printf("Fat_FileResize: first cluster: %i\r\n", hFile->StartCA);
#endif

	if( newNumClus > curNumClus ) {
		//add more clusters
		//address = hFile->Size - hFile->Position; //strange, since the position is 0 anyway, there was a reset of the file jusst few LOC before
		File_Seek( pFS, hFile, hFile->Size, NULL ); //seek to the end of the file, this updates the file handle and sets the CurrentCA with the index of the last cluster in the chain, very complicated way of doing that, but indees, saves LOC
		newNumClus -= curNumClus; //how many clusters to add
		
#if FATDEBUGGING
	printf("Fat_FileResize: pre AppendCluster\r\n");
#endif
	_Fat_FileAppendCluster(pFS, hFile, newNumClus); //append clusters
#if FATDEBUGGING
	printf("Fat_FileResize: post AppendCluster\r\n");
#endif
		//_Fat_FileAppendCluster(pFS, hFile, curNumClus); //append clusters
		//hFile->Position += 1; //the position is now in ??? it's not always in the next cluster, so what's the goal of this operation???
		
		//Fat_EraseChain( pFS, hFile->CurrentCA ); //??? 
	} if (newNumClus < curNumClus) {
		// TODO shorten cluster chain!
	}

	hFile->Size = size;
	File_Reset(hFile);
	File_Seek(pFS, hFile, positionInTheFile, NULL);

	//
	// Update dir entry
	//
	address = hFile->DirEntryPos + 28; //ajaj, absolute position, it would be nicer to read the directoryentry, udpate, and then store
									   //this is wrong address, the DirEntryPos has to be wrong
	ret = Sd_write(address, &size, sizeof(uint32_t));
#if FATDEBUGGING
	printf("Fat_FileResize: %lu, %i, %lu\r\n", size, ret, address);
#endif

	return ret != 0;
}

/**
 * Update contents an existing directory entry.
 */
bool Fat_UpdateDirEntry(
	fat_state_t*		pFS,
	uint32_t*				pAddress,
	struct fat_direntry*		pDirEntry
) {
	uint16_t	ret;
	
	//blockOffset = _Fat_AlignAddress(&blockAdr);
	ret = Sd_write(*pAddress, pDirEntry, sizeof(struct fat_direntry) );
	return( ret != 0 );
}

uint8_t counter2;

bool _Fat_CreateDirEntry(
	fat_state_t*	pFS,
	file_handle_t*	hParent,
	struct fat_direntry*	pDirEntry,
	uint32_t*			pAddress
) {
	struct fat_direntry*	pDE;
	uint16_t			ret;
	sd_cache_t*		pBuf;
	int 			i;


	File_Reset(hParent);
	do {
		*pAddress = _Fat_GetFilePos(pFS, hParent);
		
		pBuf = Sd_cache_read_block(pAddress);
		RETF(pBuf);
		pDE = (struct fat_direntry*)pBuf->buffer;
		
		for( i = 0; i < 16; i++, pDE++ )  {
			if( (pDE->Name[0] == (char)FAT_NAME_FLAG_DELETED) ||
				(pDE->Name[0] == (char)FAT_NAME_FLAG_END) )
			{
				memcpy( pDE, pDirEntry, sizeof(struct fat_direntry) );
				pBuf->state |= SD_CACHE_DIRTY;

				if( pDE->Name[0] == (char)FAT_NAME_FLAG_END ) {
					if( i == 15 ) {
						// worst case: new block needed
						SD_FREE_LOCK(pBuf);			
						File_Seek( pFS, hParent, SD_WRITE_BLOCKLENGTH, NULL );
						
						if( hParent->Flags & FILE_HANDLE_FLAG_EOF ) {
							if( hParent->StartCA == 1 ) {
								// root dir full
								return false;							
							} else {
								ret = _Fat_FileAppendCluster(pFS, hParent, 1);
								RETF(ret);
								*pAddress = _Fat_GetFilePos(pFS, hParent);
							}
						} else {
							// directory has more space (eg. clustersize > sectorsize)
							*pAddress += SD_WRITE_BLOCKLENGTH;
						}
						Sd_cache_flush();
						SD_GET_LOCK(pBuf);
						pBuf->address = *pAddress;
						
						// clear new directory block
						memset(pBuf->buffer, 0, SD_WRITE_BLOCKLENGTH);
						pDE = (struct fat_direntry*)pBuf->buffer;
					} else {
						pDE++;
					}
					*((uint8_t*)pDE) = FAT_NAME_FLAG_END;
					pBuf->state |= SD_CACHE_DIRTY;
				}
				goto _Fat_CreateDirEntry_Done;
			}
		}
		
		SD_FREE_LOCK(pBuf);
		File_Seek( pFS, hParent, SD_WRITE_BLOCKLENGTH, NULL );
		if( hParent->Flags & FILE_HANDLE_FLAG_EOF ) {
			ret = _Fat_FileAppendCluster(pFS, hParent, 1);
			RETF(ret);
		}
	} while(1);
	
 _Fat_CreateDirEntry_Done:
	SD_FREE_LOCK(pBuf);
	Sd_flush();
	return true;
}


uint8_t counter;

bool Fat_CreateFile(
	fat_state_t*		pFS,
	file_handle_t*		hParent,
	const char*			name,
	const uint8_t			attributes,
	const fat_date_t	date,
	const fat_time_t	time,
	const uint32_t		size,
	file_handle_t*		hFile,
	struct fat_direntry*		pDirEntry
) {
	struct fat_direntry	de;
	fat_ca_t		numC;
	uint8_t			ret;
	uint16_t			bps = (1 << pFS->BytsPerClus_bit) - 1;

	counter = 0;


	numC = size >> pFS->BytsPerClus_bit;
	if( size & bps )
		numC++;

#if FATDEBUGGING
	printf("Fat_CreateFile: numC %i\r\n", numC);
#endif

	File_Close(hFile);
	ret = _Fat_CreateChain(pFS, &hFile->StartCA, numC);
	RETF(ret);		// disk full? create chain should be undone in case of failure
	
	//
	// Create directory entry and handle
	//
	memset(((uint8_t*)&de)+12, 0, sizeof(struct fat_direntry) - 12);
	Fat_CanonicalNameToFAT(de.Name, name);
	memcpy(&de.CrtDate, &date, 2);
	memcpy(&de.CrtTime, &time, 2);
	memcpy(&de.WrtDate, &date, 2);
	memcpy(&de.WrtTime, &time, 2);
	memcpy(&de.LstAccDate, &date, 2);
	memcpy(&de.FstClusLO, &hFile->StartCA, 2);
	de.Attributes = attributes;
		
	if( attributes & FAT_ATTR_DIRECTORY ) {
		hFile->Flags = FILE_HANDLE_FLAG_DIRECTORY;
	} else {
		memcpy(&de.FileSize, &size, 4);
	}
	hFile->Size = size;
	File_Reset(hFile);
	

	if( pDirEntry )
		memcpy(pDirEntry, &de, sizeof(struct fat_direntry));
	
	ret = _Fat_CreateDirEntry(pFS, hParent, &de, &hFile->DirEntryPos);
	return ret;
}

#endif
