 /*
 * Copyright (c) 2008, Freie Universitt Berlin, Tomasz Naumowicz
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:
 * 
 * Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
 * Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.
 * Neither the name of the Freie Universitt Berlin nor the names of its contributors may be used to endorse or promote products derived from this software without specific prior written permission.
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */
 
#include "System.h"

fp_command_t _Commands_GetCommand(const char* name, uint8_t cmd_flags) {
	//find to string function
	extern command_t __commands_start;
	extern command_t __commands_end;
	
	command_t* current;
	
	for (current = &__commands_start; current <= &__commands_end; current++) {
		if (strncmp(current->commandName, name, 8) == 0) {
			if (current->flags & cmd_flags) {
				return current->handler;
			}
		}
	}
	
	return NULL;
}

void _Commands_PrintResponse(uint8_t executing_node, char* command_name, char* response) {
	if (response[0] != '\0')
		printf("[%u:%.8s] %s\r\n", executing_node, command_name, response);
	else
		printf("[%u:%.8s] OK\r\n", executing_node, command_name);
}

bool _Commands_SendResponse(uint8_t destination, char* command_name, char* response) {
	//char response_buffer[COMMAND_RESPONSE_LENGTH];
	//memcpy(response_buffer, command_name, 8);
	//memcpy(&response_buffer[8], response, COMMAND_RESPONSE_LENGTH - 8);
	
	network_payload_t payload;
	Network_InitSendArgs(&payload);
	payload.Buffers[0].Size = 8;
	payload.Buffers[0].Buffer = command_name;
	payload.Buffers[1].Size = COMMAND_RESPONSE_LENGTH - 8;
	if (payload.Buffers[1].Size > NETWORK_MAX_TOTAL_PAYLOAD_SIZE - 8)
		payload.Buffers[1].Size = NETWORK_MAX_TOTAL_PAYLOAD_SIZE - 8;
	
	payload.Buffers[1].Buffer = response;
	
	return Network_SendWithoutACK(destination, PID_REMOTECOMMAND_RESPONSE, &payload);
}


bool _Commands_ExecuteCommand(uint8_t requestedByNodeID, char* input) {
	//tokenize input
	//   <command><space><arg><space><arg><space><arg>\0

	// identify the command name and the arguments of the command
	char* command = NULL;
	char* arguments = NULL;
	
	command = strtok(input, " "); //input gets modified
	if (command == NULL)
		return false;
	
	arguments = strtok(NULL, "\0");

	// prepare the command_args_t struct for the command handler
	command_args_t commandArgs;
	
	commandArgs.executing_node = Configuration.NodeID;
	commandArgs.command_name = command;
	commandArgs.arguments = arguments;
	commandArgs.arguments_count = 0;
		if (arguments[0] != '\0') {
			commandArgs.arguments_count = 1;
			register uint8_t i;
			for (i = strlen(arguments); i > 0; i--) {
				if (arguments[i-1] == ' ')
					commandArgs.arguments_count++;
			}
		}
	
	//prepare the response buffer
	char response_buffer[COMMAND_RESPONSE_LENGTH]; response_buffer[0] = '\0';
	commandArgs.response = response_buffer;
	commandArgs.response_length = COMMAND_RESPONSE_LENGTH;
	
	// commandArgs are prepared for execution
	//LOG("[cs] t:%u c:'%-8.8s' a:%u '%s'", commandArgs.executing_node, commandArgs.command_name, commandArgs.arguments_count, commandArgs.arguments);
	
	// find the command handler
	fp_command_t cmd_handler = NULL;
	
	if (requestedByNodeID == Configuration.NodeID)
		cmd_handler = _Commands_GetCommand(commandArgs.command_name, COMMAND_SERIAL);
	else
		cmd_handler = _Commands_GetCommand(commandArgs.command_name, COMMAND_REMOTE);
	
	if (cmd_handler == NULL)
		return false;				// command handler not found
	
	// be aware of command's side effects, eg. change of the NodeID. The NodeID is used to deciede what to do with the command's response
	uint8_t _nodeID = Configuration.NodeID;
	cmd_handler(&commandArgs);
	
	if (requestedByNodeID == _nodeID)
		_Commands_PrintResponse(commandArgs.executing_node, commandArgs.command_name, commandArgs.response);
	else
		return _Commands_SendResponse(requestedByNodeID, commandArgs.command_name, commandArgs.response);

	return true;
}

bool Commands_ProcessSerialInput(char* input) {

	char* pchar = input;

	// check whether a NodeID was specified and identify the NodeID of the node selected for execution of the command:
	// if the input starts with @ then the command should be executed on the selected node
	// (allow to select the current node)
	uint8_t executing_node = Configuration.NodeID;
	if (*pchar == '@') {
		pchar++;						// skip the '@'
		executing_node = atoi(pchar);
		while (*pchar != ' ')			// skip the nodeid
			pchar++;
		while (*pchar == ' ')			// skip the white spaces after the id
			pchar++;
	}
	
	// the NodeID of the executing node was found, decide whether to process the input locally or to forward it to tne requested noe:
	if (executing_node == Configuration.NodeID)
		return _Commands_ExecuteCommand(Configuration.NodeID, pchar);
	else {
		network_payload_t payload;
		Network_InitSendArgs(&payload);
		payload.Buffers[0].Size = NETWORK_MAX_TOTAL_PAYLOAD_SIZE;
		payload.Buffers[0].Buffer = pchar;
		
		return Network_SendWithoutACK(executing_node, PID_REMOTECOMMAND_REQUEST, &payload);
	}
}

void _Commands_ProcessRadioRequest(uint8_t source, uint8_t* data) {
	// example: data == "id 3"
	// tries to execute the command
	
	char* data_copy = data; //holds the reference to the first character (the command name)
	bool executed = _Commands_ExecuteCommand(source, data);
	if (executed)
		return; //everything went fine, the response was sent.
	
	// execution has failed, send an error message
	_Commands_SendResponse(source, data_copy, "ERROR");
}

void _Commands_ProcessRadioResponse(uint8_t source, uint8_t* data) {
	_Commands_PrintResponse(source, &data[0], &data[8]);
}

void Commands_Init() {
	Network_RegisterHandler(PID_REMOTECOMMAND_REQUEST, _Commands_ProcessRadioRequest);
	Network_RegisterHandler(PID_REMOTECOMMAND_RESPONSE, _Commands_ProcessRadioResponse);
}

COMMAND(list, COMMAND_SERIAL, cmd_arg)
{
	extern command_t __commands_start;
	extern command_t __commands_end;
	
	command_t* current;
	
	printf("Commands:\r\n");

	for (current = &__commands_start; current < &__commands_end; current++) {
		printf(" %-8.8s [%p]", current->commandName, current->handler);
		if (current->flags & COMMAND_SERIAL)
			printf(" [local]");
		if (current->flags & COMMAND_REMOTE)
			printf(" [remote]");
		
		printf("\r\n");
	}
}

COMMAND(reset, COMMAND_SERIAL | COMMAND_REMOTE, cmd_args) {
	WDTCTL = 0x0000;
}

COMMAND(app, COMMAND_SERIAL | COMMAND_REMOTE, cmd_args) {
	appHeader_t* _applicationHeader;
	_applicationHeader = (appHeader_t*) 0x4600;

	snprintf(cmd_args->response, cmd_args->response_length,
			"'%-8.8s' V: %i.%i",
			_applicationHeader->appName,
			_applicationHeader->appVer[0],
			_applicationHeader->appVer[1]
	);
}
/*
COMMAND(api, COMMAND_SERIAL | COMMAND_REMOTE, cmd_args) {
	appHeader_t* _applicationHeader;
	_applicationHeader = (appHeader_t*) 0x4600;

	snprintf(cmd_args->response, cmd_args->response_length,
			"ScatterWeb.API %s",
			SCATTERWEB_API_VERSION
	);
}
*/
