/**
 * Copyright (c) 2008, Freie Universitt Berlin, Tomasz Naumowicz
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:
 * 
 * Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
 * Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.
 * Neither the name of the Freie Universitt Berlin nor the names of its contributors may be used to endorse or promote products derived from this software without specific prior written permission.
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include "System.h"

/**
 * Enables the ADC12 on the MSP430
 */
void MSB430H_ADC12_Enable() {
	ADC12CTL0 = ADC12ON;
	ADC12CTL1 = ADC12SSEL_ACLK | SHP;
	
	ADC12MCTL0 = INCH_A4;
	ADC12MCTL1 = INCH_A5 | EOS; 	// end of sequence
}

/**
 * Enables the ADC12 on the MSP430
 */
void MSB430H_ADC12_Disable() {
	ADC12CTL0 &= ~ADC12ON;
	ADC12CTL1 = ADC12SSEL_ACLK;
}

uint16_t MSB430H_ADC12_Read(enum adc_port port) {
	uint16_t result = 0;
	
	if (!(ADC12CTL0 & ADC12ON))
		return 0;
	
	switch (port) {
		case P64:
			ADC12CTL1 &= ~CSTARTADD_15;			//sets CSSTARTADD ot MEM0, INCHA4, P64

			ADC12CTL0 |= ADC12SC | ENC;
			ADC12CTL0 &= ~ADC12SC;

			while (ADC12CTL1 & ADC12BUSY) {
				_NOP();
			}
			
			ADC12CTL0 &= ~ENC;
			
			result = ADC12MEM0;
			
			break;
		case P65:
			ADC12CTL1 &= ~CSTARTADD_15;			//sets CSSTARTADD ot MEM0, INCHA4, P64
			ADC12CTL1 |= CSTARTADD_1;			//sets CSSTARTADD to MEM1, INCHA5, P65

			ADC12CTL0 |= ADC12SC | ENC;
			ADC12CTL0 &= ~ADC12SC;

			while (ADC12CTL1 & ADC12BUSY) {
				_NOP();
			}
			
			result = ADC12MEM1;
			break;
	}
	
	return result;
}

void MSB430H_InitPorts() {
		P1SEL = 0x00;
		P1OUT = 0x00;
		P1DIR = 0x00;
		P2SEL = 0x00;
		P2OUT = 0x18;
		P2DIR = 0x1A;
		P3SEL = 0x00;
		P3OUT = 0x09;
		P3DIR = 0x21;
		P4SEL = 0x00;
		P4OUT = 0x00;
		P4DIR = 0x00;
		P5SEL = 0x0E;
		P5OUT = 0xF9;
		P5DIR = 0xFD;
		P6SEL = 0x07;
		P6OUT = 0x00;
		P6DIR = 0xC8;
	}

#define DELTA 1100                   // Target DCO = DELTA(1100)*(4096Hz) , DELTA=550 => 4505600 Hz  */
#define UP 1                         // Temporary.
#define DOWN 2                       // Temporary.
#define DCOLOW  (1)                  // Symbolic name, DCO was last measure too low.
#define DCOHIGH (2)                  // Symbolic name, DCO was last measure too high.
#define DCODIFF (3)                  // Difference of dcoDiff to DELTA when DCO measure is taken as too high/low.

volatile uint16_t dcoCap;        ///< Contains always last TBCCR6 Value (Capture on ACLK)
volatile uint16_t dcoDiff;       ///< Everytime dcoDiff==0 TIMER_B1 interrupt writes a new DCO diff measure into it.
volatile uint8_t  checkDCOState; ///< DCO will only be changed if two times low or high in a row to avoid peaks. Implemented with this variable.

bool DCO_set()
{
	uint16_t Compare, Oldcapture = 0, i, direction = 0;
	// ACLK is devided by 8 = 4096 Hz. RSEL=6 no division for MCLK and SSMCLK. XT2 is off.
	BCSCTL1 = 0xB7;
	// Init FLL to desired frequency using the 32762Hz crystal DCO frquenzy = 2,4576 MHz
	BCSCTL2 = 0x00;
	DCOCTL = 0xFF;
	for ( i = 0xFFFF; i > 0; i-- );   // Delay for XTAL to settle
	CCTL2 = CCIS0 + CM0 + CAP;        // Define CCR2, CAP, ACLK
	TACTL = TASSEL1 + TACLR + MC1;    // SMCLK, continous mode
	while ( 1 ) {
		CCTL2 &= ~( CCIFG + COV );      // clear IFG and possible overflow
		// 1. capture
		while ( !( CCTL2 & CCIFG ) );           // Wait until capture occured!
		CCTL2 &= ~CCIFG;                        // Capture occured, clear flag
		Oldcapture = CCR2;                      // Save current captured SMCLK
		// 2. - 5. capture, "software divider" so aclk stays
		// untouched and 2400 bps are working!
		while ( !( CCTL2 & CCIFG ) );         // Wait until capture occured!
		CCTL2 &= ~CCIFG;                      // Capture occured, clear flag
		Compare = CCR2;
		Compare = Compare - Oldcapture;         // SMCLK difference
		if ( DELTA == Compare ) break;          // if equal, leave "while(1)"
		else {
			if ( DELTA < Compare ) {           // DCO is too fast, slow it down
				if ( direction == UP ) break;    // stop if we were going up (jumped over exact delta)
				if ( !direction ) direction = DOWN; // if starting set direction
				DCOCTL--;
				if ( DCOCTL == 0xFF ) {
					if ( !(( BCSCTL1 & 0x07 ) == 0x00 ) ) WDTCTL = 0xDEAD;
					BCSCTL1--;            // Did DCO role under? -> Select next lower RSEL
				}
			} else {
				if ( direction == DOWN ) break;   // stop if we were going down (jumped over exact delta)
				if ( !direction ) direction = UP; // if starting set direction
				DCOCTL++;
				if ( DCOCTL == 0x00 ) {
					if (( BCSCTL1 & 0x07 ) == 0x07 ) WDTCTL = 0xDEAD;
					BCSCTL1++;            // Did DCO role over? // -> Select next higher RSEL
				}
			}
		}
	}
	CCTL2 = 0;      // Stop CCR2 function
	TACTL = 0;      // Stop Timer_A
	dcoDiff = 0;
	checkDCOState = 0;
	return true;
}

// Subroutine to check and readjust the dco.
//
// This routine will run whenevery the Timer_B1 interrupt has left a new value in dcoDiff.
// The algorithm checks if the measured frequency is more than DCODIFF steps off the
// desired DELTA frequency. If that is the case two times in row the algorithm
// lowers/highers the dco frequency. Set bit DEBUG_DCO in Configuration.debugFlags to see
// readjustments of the dco.
//
// If stable, diff will always oscillate between two numbers, for example 1097 and 1098.
// At theses frequency (~4,5 Mhz) one step of DCO means 4 measured diffDCO steps,
// for example one DCO step up with the previous example will produce 1101/1102.
// To avoid constant changing of DCO frequency and proviode some hysteresis DCODIFF
// should be at least 3 (19200 rf and 115200 baud serial comm get only lost at diffs greater 50).
void DCO_check()
{
	if ( dcoDiff < ( DELTA - DCODIFF ) ) {      // frequency too low
		if ( checkDCOState != DCOLOW ) checkDCOState = DCOLOW;
		else {
#if DEBUGLOG_LEVEL >= 5
			LOG_HARDWARE( DEBUG_DCO, "DCO++", NULL );
#endif
			DCOCTL++;
			if ( DCOCTL == 0x00 ) {
				if (( BCSCTL1 & 0x07 ) == 0x07 ) {
#if DEBUGLOG_LEVEL >= 5
					LOG_HARDWARE( DEBUG_DCO, "Freq too high!", NULL );
#endif
					WDTCTL = 0xDEAD;
				}
				BCSCTL1++;      // Did DCO role over? // -> Select next higher RSEL
			}
#if DEBUGLOG_LEVEL >= 5
			LOG_HARDWARE( DEBUG_DCO, "DCO %b %b >%i", DCOCTL, BCSCTL1, dcoDiff );
#endif
		}
	} else {
		if ( dcoDiff > ( DELTA + DCODIFF ) ) {      // frequency too high
			if ( checkDCOState != DCOHIGH ) {
				checkDCOState = DCOHIGH;
			} else {
#if DEBUGLOG_LEVEL >= 5
				LOG_HARDWARE( DEBUG_DCO, "DCO--", NULL );
#endif
				DCOCTL--;
				if ( DCOCTL == 0xFF ) {
					if ( !(( BCSCTL1 & 0x07 ) == 0x00 ) ) {
#if DEBUGLOG_LEVEL >= 5
						LOG_HARDWARE( DEBUG_DCO, "Freq too low!", NULL );
#endif
						WDTCTL = 0xDEAD;
					}
					BCSCTL1--;      // Did DCO role under? -> Select next lower RSEL
				}
#if DEBUGLOG_LEVEL >= 5
				LOG_HARDWARE( DEBUG_DCO, "DCO %b %b >%i\r\n", DCOCTL, BCSCTL1, dcoDiff );
#endif
			}
		} else checkDCOState = 0; // freqency was ok
	}
	Timers_Add( DCO_check, 800, 0 );
}

// DCO Checker. Interrupt occurs every ~250us. Continously updates ::dcoDiff.
interrupt (TIMERB1_VECTOR) Timer_B(void)
{
	dcoDiff = TBCCR6 - dcoCap;
	dcoCap = TBCCR6;
	TBCCTL6 &= ~( CCIFG );
}

void DCO_start() {
// WICHTIG: DIESER TEIL WIRD NOCH IN DER TIMERS_INIT GEMACHT - ANFANG

  ////////////////////////////////// Init TIMERB to run continously from DCO at 4,505606 Hz and setup TB6 for DCO checking.
  TBCTL = TBSSEL1 + TBCLR;        // Clr timer, select SMCLK.
  TBCTL |= MC1;                   // Start Timer B in continuous mode
    //extern unsigned int dcoCap;
    //extern void DCO_check();
    //extern void Timer_B1();
    //SET_ISR(TIMERB1_VECTOR, Timer_B1);
    TBCCTL6 = CCIS_1 + CM_2 + CAP;    // Set TB6 to capture CCIB (which is ACLK on TB6) on one edge, clears CCIFG
    //////////////////////////////////// Await the first cap to init dcoCap.
    while(!(TBCCTL6 & CCIFG)) _NOP(); // Wait polling for first capture.
    dcoCap = TBCCR6;                  // Save first capture like in isr.
    TBCCTL6 &= ~(CCIFG);              // Clear ifg like in isr:
    TBCCTL6 |= CCIE;                  // Enable tb6 interrupt.
    Timers_Add(DCO_check, 800, 0);

// WICHTIG: DIESER TEIL WIRD NOCH IN DER TIMERS_INIT GEMACHT - ENDE
}
 
