/**
 * Copyright (c) 2008, Freie Universitt Berlin, Tomasz Naumowicz
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:
 * 
 * Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
 * Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.
 * Neither the name of the Freie Universitt Berlin nor the names of its contributors may be used to endorse or promote products derived from this software without specific prior written permission.
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include "System.h"
#include "Network.h"

struct radio_interface {
	uint8_t	NumChannels;
	uint8_t(*SelectChannel)(uint8_t channel);
	uint8_t(*SetTxPower)(uint8_t power);
};

uint8_t internalID;
uint8_t currentCCchannel;
uint8_t currentCCpower;

enum packettypes {
	ACK_PACKET,						///< Packet Type: Acknowledgement packet.
};


/* CC1020 */

#define CC1020_RSS 0x42
#define TX_CURRENT 0x87
#define RX_CURRENT 0x86
#define CC1020_NUM_CHANNELS		8
#define PDO            (P2IN & 0x01)                    // PDI (Data in) is on P21
#define PSEL_ON        do { P3OUT &=  ~0x01; } while(0) // PSEL is on P30 and low active
#define PSEL_OFF       do { P3OUT |=   0x01; } while(0)
#define PCLK_HIGH      do { P2OUT |=  0x08; }  while(0)
#define PCLK_LOW       do { P2OUT &= ~0x08; }  while(0)
#define PDI_HIGH       do { P2OUT |=  0x02; }  while(0) // PDO (Data out) is on P22
#define PDI_LOW        do { P2OUT &= ~0x02; }  while(0)
#define CAL_COMPLETE   (0x80)
#define PA_POWER       0x4F
#define CAL_TIMEOUT    0x7FFE
#define LOCK_TIMEOUT   0x7FFE
#define RESET_TIMEOUT  0x7FFE
#define LNA_POWER_ON   do { P2OUT &= ~0x10; }  while(0) // Enable power for LNA (P24, low-active)

enum cc1020_register {
	CC1020_MAIN          = 0x00,
	CC1020_INTERFACE     = 0x01,
	CC1020_RESET         = 0x02,
	CC1020_SEQUENCING    = 0x03,
	CC1020_FREQ_2A       = 0x04,
	CC1020_FREQ_1A       = 0x05,
	CC1020_FREQ_0A       = 0x06,
	CC1020_CLOCK_A       = 0x07,
	CC1020_FREQ_2B       = 0x08,
	CC1020_FREQ_1B       = 0x09,
	CC1020_FREQ_0B       = 0x0A,
	CC1020_CLOCK_B       = 0x0B,
	CC1020_VCO           = 0x0C,
	CC1020_MODEM         = 0x0D,
	CC1020_DEVIATION     = 0x0E,
	CC1020_AFC_CONTROL   = 0x0F,
	CC1020_FILTER        = 0x10,
	CC1020_VGA1          = 0x11,
	CC1020_VGA2          = 0x12,
	CC1020_VGA3          = 0x13,
	CC1020_VGA4          = 0x14,
	CC1020_LOCK          = 0x15,
	CC1020_FRONTEND      = 0x16,
	CC1020_ANALOG        = 0x17,
	CC1020_BUFF_SWING    = 0x18,
	CC1020_BUFF_CURRENT  = 0x19,
	CC1020_PLL_BW        = 0x1A,
	CC1020_CALIBRATE     = 0x1B,
	CC1020_PA_POWER      = 0x1C,
	CC1020_MATCH         = 0x1D,
	CC1020_PHASE_COMP    = 0x1E,
	CC1020_GAIN_COMP     = 0x1F,
	CC1020_POWERDOWN     = 0x20,
	CC1020_TEST1         = 0x21,
	CC1020_TEST2         = 0x22,
	CC1020_TEST3         = 0x23,
	CC1020_TEST4         = 0x24,
	CC1020_TEST5         = 0x25,
	CC1020_TEST6         = 0x26,
	CC1020_TEST7         = 0x27,
	CC1020_STATUS        = 0x40,
	CC1020_RESET_DONE    = 0x41,
	CC1020_AFC           = 0x43,
	CC1020_GAUSS_FILTER  = 0x44,
	CC1020_STATUS1       = 0x45,
	CC1020_STATUS2       = 0x46,
	CC1020_STATUS3       = 0x47,
	CC1020_STATUS4       = 0x48,
	CC1020_STATUS5       = 0x49,
	CC1020_STATUS6       = 0x4A,
	CC1020_STATUS7       = 0x4B
};

enum CC1020_lock_status {
	LOCK_NOK       = 0x00,
	LOCK_OK        = 0x01,
	LOCK_RECAL_OK  = 0x02,
	LOCK_CONTINOUS = 0x10
};

struct cc1020_config {
  uint8_t	registerValues[41];
  uint8_t	ackTimeout;
};

struct cc1020_config currentCC1020Config;

const struct cc1020_config DefaultConfig_19200 = {
	{
		0x01, // 0x00, MAIN
		0x0F, // 0x01, INTERFACE
		0xFF, // 0x02, RESET
		0x8F, // 0x03, SEQUENCING
		//869.525 bei 50kHz
		0x3A, // 0x04, FREQ_2A
		0x32, // 0x05, FREQ_1A
		0x97, // 0x06, FREQ_0A // 19200
		0x38, // 0x07, CLOCK_A // 19200
		0x3A, // 0x08, FREQ_2B
		0x37, // 0x09, FREQ_1B
		0xEB, // 0x0A, FREQ_0B // 19200
		0x38, // 0x0B, CLOCK_B // 19200
		0x44, // 0x0C, VCO     44
		0x51, // 0x0D, MODEM   Manchester
		0x2B, // 0x0E, DEVIATION // FSK
		0x4C, // 0x0F, AFC_CONTROL Ruetten 0xCC
		0x25, // 0x10, FILTER Bandwith 100 kHz
		0x61, // 0x11, VGA1
		0x55, // 0x12, VGA2
		0x2D, // 0x13, VGA3
		0x37, //0x29,  // 0x14, VGA4 ADJUSTED CS to 23!
		0x40, // 0x15, LOCK is Carrier SENSE
		0x76, // 0x16, FRONTEND
		0x87, // 0x17, ANALOG, RX=86/TX=87
		0x10, // 0x18, BUFF_SWING
		0x25, // 0x19, BUFF_CURRENT
		0xAE, // 0x1A, PLL_BW
		0x34, // 0x1B, CALIBRATE
		0x0F, // 0x1C, PA_POWER
		0xF0, // 0x1D, MATCH
		0x00, // 0x1E, PHASE_COMP
		0x00, // 0x1F, GAIN_COMP
		0x00, // 0x20, POWERDOWN
		0x4d, // 0x21,
		0x10, // 0x22,
		0x06, // 0x23,
		0x00, // 0x24,
		0x40, // 0x25,
		0x00, // 0x26,
		0x00, // 0x27, 
	},
	16                 
};

const struct cc1020_config DefaultConfig_115200 = {
	{
		0x01, // 0x00, MAIN
		0x0F, // 0x01, INTERFACE
		0xFF, // 0x02, RESET
		0x8F, // 0x03, SEQUENCING
		// 869.525 bei 200kHz
		0x3A, // 0x04, FREQ_2A
		0x32, // 0x05, FREQ_1A
		0x97, // 0x06, FREQ_0A // 19200
		0x29, // 0x07, CLOCK_A // 19200
		0x3A, // 0x08, FREQ_2B
		0x37, // 0x09, FREQ_1B
		0xEB, // 0x0A, FREQ_0B // 19200
		0x29, // 0x0B, CLOCK_B // 19200
		0x44, // 0x0C, VCO 44
		0x51, // 0x0D, MODEM Manchester
		0x58, // 0x0E, DEVIATION // FSK
		0x4C, // 0x0F, AFC_CONTROL Ruetten 0xCC
		0x80, // 0x10, FILTER Bandwith 500 kHz
		0x61, // 0x11, VGA1
		0x55, // 0x12, VGA2
		0x30, // 0x13, VGA3
		0x35, // 0x14, VGA4
		0x20, // 0x15, LOCK is Carrier SENSE
		0x76, // 0x16, FRONTEND
		0x87, // 0x17, ANALOG, RX=86/TX=87
		0x10, // 0x18, BUFF_SWING
		0x25, // 0x19, BUFF_CURRENT
		0xAE, // 0x1A, PLL_BW
		0x34, // 0x1B, CALIBRATE
		0x0F, // 0x1C, PA_POWER
		0xF0, // 0x1D, MATCH
		0x00, // 0x1E, PHASE_COMP
		0x00, // 0x1F, GAIN_COMP
		0x00, // 0x20, POWERDOWN
		0x4d, // 0x21,
		0x10, // 0x22,
		0x06, // 0x23,
		0x00, // 0x24,
		0x40, // 0x25,
		0x00, // 0x26,
		0x00  // 0x27,
	},
	4         
};

void CC1020_writeRegister(enum cc1020_register addr, uint8_t adata) {
  char  BitCounter;
  unsigned char Data ;
  PSEL_OFF;
  Data = (uint8_t)addr << 1;
  PSEL_ON;
  // Send address bits 
  for (BitCounter=0;BitCounter<7;BitCounter++) {
    nop();         // Delay50ns();
  	PCLK_LOW;
    nop();         // Delay50ns();
    if(Data & 0x80)  PDI_HIGH;
    else PDI_LOW;
    Data=Data<<1;
    PCLK_HIGH;
  }
  // Send read/write bit 
  // Ignore bit in data, always use 1 
  nop();           // Delay50ns();
  PCLK_LOW;
  PDI_HIGH;
  nop();           // Delay50ns();
  PCLK_HIGH;
  nop();           // Delay50ns();
  PCLK_LOW;
  Data=adata ;    
  // Send data bits 
  for (BitCounter=0;BitCounter<8;BitCounter++) {
    nop();         // Delay50ns();
  	PCLK_LOW;
    nop();         // Delay50ns();
    if(Data & 0x80) PDI_HIGH;
    else PDI_LOW;
    Data=Data<<1;
    PCLK_HIGH;
  }
  nop();           // Delay50ns();
  PCLK_LOW;
  nop();           // Delay50ns();
  PSEL_OFF;
}

void CC1020_reset() {
  // Reset CC1020
  CC1020_writeRegister(CC1020_MAIN, 0x0F&~0x01);
  // Bring CC1020 out of reset
  CC1020_writeRegister(CC1020_MAIN, 0x1F);
}

bool CC1020_delay( unsigned int ms )
{
	uint16_t tar = TAR;
	ms <<= 2;     // Wait with ACLK, take ms * 4 to get roughly ms (1024 hz)
	while ( TAR - tar < ms ) _NOP();
	return true;
}

void CC1020_wakeupRX(char RXANALOG) {
  // Turn on xtal oscillator core
  CC1020_writeRegister(CC1020_MAIN,0x1B);
  // Setup bias current adjustment
  CC1020_writeRegister(CC1020_ANALOG,RXANALOG);
  // Insert wait routine here, must wait for xtal oscillator to stabilise, 
  // typically takes 2-5ms.
  CC1020_delay(3); // DelayMs(5); // NOT NEEDED?
  // Turn on bias generator
  CC1020_writeRegister(CC1020_MAIN,0x19);
  CC1020_delay(1) ; // NOT NEEDED?
  // Turn on frequency synthesiser
  CC1020_writeRegister(CC1020_MAIN,0x11);
}

void CC1020_wakeupTX(char TXANALOG) {
  // Turn on xtal oscillator core
  CC1020_writeRegister(CC1020_MAIN,0xDB);
  // Setup bias current adjustment
  CC1020_writeRegister(CC1020_ANALOG,TXANALOG);
  // Insert wait routine here, must wait for xtal oscillator to stabilise, 
  // typically takes 2-5ms.
  CC1020_delay(3); // DelayMs(5); // NOT NEEDED?
  // Turn on bias generator
  CC1020_writeRegister(CC1020_MAIN,0xD9);
  CC1020_delay(1); // NOT NEEDED?
  // Turn on frequency synthesiser
  CC1020_delay(1);
  CC1020_writeRegister(CC1020_MAIN,0xD1);
}

void CC1020_writeSettings() {
  uint8_t i;  
  for(i=0;i<0x27;i++)
	  CC1020_writeRegister(i, currentCC1020Config.registerValues[i]);
}

uint8_t CC1020_readRegister(uint8_t addr) {
  char BitCounter;
  unsigned char Data = 0;
  PSEL_OFF; 
  Data=addr << 1;
  PSEL_ON;
  nop();           // Delay50ns();
  // Send address bits 
  for (BitCounter=0;BitCounter<7;BitCounter++) {
    nop();           // Delay50ns();
  	PCLK_LOW;
    nop();           // Delay50ns();
    if(Data & 0x80)  PDI_HIGH;
    else PDI_LOW;
    Data=Data<<1;
    PCLK_HIGH;
  }
  // Send read/write bit 
  // Ignore bit in data, always use 0 
  nop();             // Delay50ns();
  PCLK_LOW;
  PDI_LOW;
  nop();             // Delay50ns();
  PCLK_HIGH;
  nop();             // Delay50ns();
  PCLK_LOW;
  // Receive data bits       
  for (BitCounter=0;BitCounter<8;BitCounter++) {
    nop();           // Delay50ns();
  	PCLK_HIGH;
    nop();           // Delay50ns();
    Data=Data<<1;
    if (PDO) Data++; // Data = Data | PDO ;
    nop();           // Delay50ns();
    PCLK_LOW;
  }
  nop();             // Delay50ns();
  PSEL_OFF; 
  return Data;
}

uint8_t _CC1020_waitlock() {
	int t;
	uint8_t status;
	
	for( t = LOCK_TIMEOUT; t > 0; t-- ) {
		status = CC1020_readRegister(CC1020_STATUS);
		if( status & LOCK_CONTINOUS )
			return LOCK_OK;
	}
	return LOCK_NOK;
}

char CC1020_calibrate() {
  unsigned char tmp;
  enum CC1020_lock_status lock_status;
  // Turn off PA to avoid spurs during calibration in TX mode
  CC1020_writeRegister(CC1020_PA_POWER,0x00);
  // Start calibration
  CC1020_writeRegister(CC1020_CALIBRATE,0xB5);
  CC1020_delay(3); 
  while((tmp =(CC1020_readRegister(CC1020_STATUS)&CAL_COMPLETE)) == 0);
  CC1020_delay(2); 
  // Monitor lock
  lock_status = _CC1020_waitlock();
  // Restore PA_POWER
  CC1020_writeRegister(CC1020_PA_POWER, currentCC1020Config.registerValues[CC1020_PA_POWER]);
  // Return state of LOCK_CONTINUOUS bit
  return lock_status;
}

enum CC1020_lock_status CC1020_setupRX(char RXANALOG) {
  enum CC1020_lock_status lock_status;
  
  // Switch into RX, switch to freq. reg A
  CC1020_writeRegister(CC1020_MAIN,0x11);
  // Setup bias current adjustment
  CC1020_writeRegister(CC1020_ANALOG,RXANALOG);
  CC1020_delay(1); // Wait for 1 msec  
  
  // Monitor LOCK
  lock_status = _CC1020_waitlock();
  if( lock_status != LOCK_OK ) {
	  // (PLL out of LOCK)
	  // If recalibration ok
    if(CC1020_calibrate())
    	lock_status = LOCK_RECAL_OK; // Indicate PLL in LOCK
	// Else (recalibration failed)
  }
  
  // Switch RX part of CC1020 on
  CC1020_writeRegister(CC1020_MAIN,0x01);
  //WriteToCC1020Register(CC1020_INTERFACE,0x12);
  CC1020_writeRegister(CC1020_INTERFACE,0x02);
  
  // Return LOCK status to application
  return lock_status;
}

uint8_t CC1020_selectChannel(uint8_t channel) {
	static const struct {
		uint8_t	tx3;
		uint8_t tx2;
		uint8_t tx1;
		uint8_t rx3;
		uint8_t rx2;
		uint8_t rx1;
	} freq_settings[] = {
		{0x3A, 0x32, 0x97, 0x3A, 0x37, 0xEB}, // ch1
		{0x36, 0xCE, 0x89, 0x36, 0xD3, 0xDD}, // ch2
		{0x37, 0x7C, 0x25, 0x37, 0x81, 0x79}, // ch3
		{0x38, 0x29, 0xC1, 0x38, 0x2F, 0x17}, // ch4
		{0x38, 0xD7, 0x5D, 0x38, 0xDC, 0xB3}, // ch5
		{0x39, 0x84, 0xF9, 0x39, 0x8A, 0x4F}, // ch6
		{0x36, 0x20, 0xEB, 0x36, 0x26, 0x41}, // ch7
		{0x3A, 0xE0, 0x33, 0x3A, 0xE5, 0x89}  // ch8
	};
	
	if( (channel == 0) || (channel > CC1020_NUM_CHANNELS) )
		return 0;

	channel--;
	currentCC1020Config.registerValues[CC1020_FREQ_2A] = freq_settings[channel].tx3;
	currentCC1020Config.registerValues[CC1020_FREQ_1A] = freq_settings[channel].tx2;
	currentCC1020Config.registerValues[CC1020_FREQ_0A] = freq_settings[channel].tx1;
	currentCC1020Config.registerValues[CC1020_FREQ_2B] = freq_settings[channel].rx3;
	currentCC1020Config.registerValues[CC1020_FREQ_2B] = freq_settings[channel].rx3;
	currentCC1020Config.registerValues[CC1020_FREQ_1B] = freq_settings[channel].rx2;
	currentCC1020Config.registerValues[CC1020_FREQ_0B] = freq_settings[channel].rx1;
	channel++;

	CC1020_reset();
	CC1020_writeSettings();
	CC1020_wakeupRX(RX_CURRENT);
	CC1020_calibrate();
	CC1020_wakeupTX(TX_CURRENT);
	CC1020_calibrate();
	CC1020_wakeupRX(RX_CURRENT);
	CC1020_setupRX(RX_CURRENT) ;
	LNA_POWER_ON;
    
    return channel;
}

void CC1020_setupPD() {
  // Put CC1020 into power-down
  CC1020_writeRegister(CC1020_MAIN,0x1F);
  // Turn off PA to minimise current draw
  CC1020_writeRegister(CC1020_PA_POWER,0x00);
}

uint8_t CC1020_setTxPower(uint8_t power) {
	currentCC1020Config.registerValues[CC1020_PA_POWER] = power;
	return power;
}

struct radio_interface Radio = {
	CC1020_NUM_CHANNELS,
	*CC1020_selectChannel,
	*CC1020_setTxPower
};

void CC1020_loadSettings(const struct cc1020_config* config) {
	memcpy(&currentCC1020Config, config, sizeof(struct cc1020_config));
	currentCC1020Config.registerValues[CC1020_PA_POWER] = currentCCpower;
}

void CC1020_init() {
	CC1020_setupPD();
	CC1020_loadSettings(&DefaultConfig_19200);
	CC1020_selectChannel(currentCCchannel);
}

uint16_t tb1;
uint16_t tb2;

/* ------------------------------------------------ */

typedef uint8_t netaddr_t;

enum netaddr_predef {
	NETADDR_UNDEFINED	= 0x00,		///< reserved for undefined (not set) or to address "this" node if a node is addressed through a wire connection
	NETADDR_MIN		 	= 0x01,		///< smallest possible node address
	NETADDR_MAX		 	= 0xFE,		///< largest possible node address
	NETADDR_BROADCAST	= 0xFF		///< broadcast address
};

enum netheader_flags {
	NET_TLFLAG_NOCRC	= 0x01,	///< do no error checking
	NET_TLFLAG_ACKREQ	= 0x02	///< request ack for this packet	
};

enum netaddr_compareresult {
	ADRCOMP_ERROR      = 0,		///< address is undefined
	ADRCOMP_THIS	   = 0x01,	///< this node
	ADRCOMP_BROADCAST  = 0x02,	///< local broadcast
	ADRCOMP_OTHER      = 0x80	///< other node in this network
};

#if defined(__CCE__)
	#pragma pack(1)
#elif defined(__MSP430GCC__)
	__attribute__((packed))
#endif
typedef struct netheader {
	uint8_t		size;			///< size of packet	
	netaddr_t	to;				///< delivery addres
	uint8_t		flags;			///< flags (see netheader_flags)
	netaddr_t	from;			///< source address
	uint8_t		_num;			///< packet number (set automatically before sending)
	uint8_t		type;			///< protocol id
} netheader_t;
#if defined(__CCE__)
	#pragma pack()
#endif

struct netpacket_handler_args {
	const netheader_t (*const			netheader);		///< pointer to received packet's network layer header
	const void (*const					payload);		///< pointer to received packet's payload
	const uint16_t						payload_size;	///< received packet's payload size
	const uint8_t						rssi;			///< rssi value for this packet's transmission
	const enum netaddr_compareresult	cpaResult;		///< result of to-address comparison
};

typedef bool(*fp_net_handler)(struct netpacket_handler_args* args);

struct netpacket_callback_args {
	const netheader_t (*const			netheader);		///< pointer to received packet's network layer header
	uint16_t							timesRepeated;	///< number of retries (if ack flag set)
	bool								success;		///< transmission succeeded
};

typedef void(*fp_net_cb)(struct netpacket_callback_args* args);

#if NETWORK_PACKET_PAYLOAD_COUNT < 2
#warning "NETPACKET_SEND_BUFFER_COUNT too small (min 2)"
#endif

typedef struct netpacket_send_args {
	/// network header
	netheader_t							netheader;
	/// optional payload blocks that are sent in consecutive order
	struct netpacket_send_buffer		payload[NETWORK_PACKET_PAYLOAD_COUNT];
	/// optional pointer to callback function that is called after a packet has been sent or sending failed
	//fp_net_cb							fpcallback;
	network_send_callback_t				fpcallback;
} netpacket_send_args_t;

enum netaddr_compareresult Net_compareAddress(
	const netaddr_t (*const adr),
	const bool directConnection
);

typedef bool(*fpRadio_selectChannel)(uint8_t channel);

#define NETSYN_DEBUG(...)

// 256 Ringbuffer macro implemenation.
#define RB(name)   \
    volatile struct {     \
      uint16_t guard1;      \
      uint8_t buffer[256];  \
      uint8_t head, tail;   \
      uint16_t guard2;      \
    } name

// Gets the first byte of the buffer and remove it.
#define RB_GET(b, x) do { x = (b).buffer[(b).head]; (b).head++; } while(0)

// Get current length.
#define RB_LENGTH(b) ((uint8_t)( (b).tail - (b).head ))

// Appends another bytes to the buffer.
#define RB_PUT(b, x) do { (b).buffer[(b).tail] = x; (b).tail++; } while(0)


// Initializes the ringbuffer.
#define RB_INIT(b) do { (b).head = 0; (b).tail = 0; (b).guard1 = 0xABCD; (b).guard2 = 0xABCD; } while(0)

#define PREAMBLE_SIZE	9		///< preamble and sync word sent before each packet
#define TAILsize		2		///< tail bytes sent after packet + crc
// TX
RB(txPackets);

// Interrupt byte radio ringbuffer.
volatile uint8_t numPackets; // Current number of packets in radio tx buffer.
volatile uint8_t txNum; // Current tx Number.
volatile uint16_t txBufPos; // Current position in txBuf.
unsigned char txLockStatus; // Saves last value from SetupCC1020TX.
volatile uint8_t txBuf[140]; // Buffer for bulding packets with preamble.
volatile uint8_t ackBuf[PREAMBLE_SIZE + sizeof(netheader_t) + 2]; //   ""          ""      ACKS
volatile uint16_t txBufLength; // How much to send in TxBuffer


static netheader_t sentPacket;
static struct netpacket_callback_args txCallbackArgs = {&sentPacket,0,0};
/// Callback what to do after a packet has been sent.
static network_send_callback_t txCallback = NULL; 

volatile uint16_t txBackoff; // Var for storing backoff in states TX_WAITING, TX_BACKOFF.
volatile uint16_t ackTimeout; // Timeout how long to wait for an ack in RADIO_STROKE ticks, depends on baudrate and should come with the config.
volatile uint8_t txMode;
volatile bool radioEnabled = false;
uint16_t v0, v1, v2;
uint16_t tim1, tim2, tim3, tim4;

#define TIMESTART()   do { tim1=TAR; tim2=TBR; }while(0)
#define TIMESTOP()    do { tim3=TAR; tim4=TBR; print("TIM %i %i\n", tim3-tim1, tim4-tim2); }while(0)
#define MAX_REPEATED  (5)         // How often a packet get repeated.
#define RADIO_STROKE  (9011)      // ~ 2 ms @ 4505600 Hz.
// RX
__attribute__((packed)) typedef struct {
	uint8_t state; ///< state of this buffer (FF=Empty 00=Rec 01=FULL)
	uint8_t reserved; ///< space to align following data to 16 bit
	union {
		uint8_t raw[140]; ///< raw data as received
		struct {
			netheader_t nwh;
			uint8_t payload[140 - sizeof(netheader_t)];
		} decoded; ///< decoded packet data
	} buffer;
} rxbuffer_t; ///< receive buffer
rxbuffer_t* rxBuf; ///< currently selected receive buffer
rxbuffer_t rxBuf1; // Buffer for received packets.
rxbuffer_t rxBuf2; // Buffer for received packets.
uint8_t rxBufPos; // Position in this buffer.
volatile time_t rxTime; // When Synch byte was discovered.
volatile uint8_t rssi; // Received Signal Strength Indicator read from chipcon during receive.

#define DISABLE_RX_IRQ() do { IE1 &= ~(URXIE0); } while(0)
#define ENABLE_RX_IRQ() do { IFG1 &= ~URXIFG0; IE1 |= URXIE0; } while(0)
#define DIRQSTART()  
#define DIRQEND()    

#define CC_LOCK (P2IN&0x04)
#define LNA_POWER_ON  do { P2OUT &= ~0x10; } while(0) // Enable power for LNA (P24, low-active)
#define LNA_POWER_OFF do { P2OUT |=  0x10; } while(0) // Disable power for LNA (P24, low-active)
#define ADR2INT(x)  (*((uint16_t*)& x))

#define MAX_RXNUM_TABLE     (20)    // Length of rxNumTable.
struct rxNumTableEntry {
	netaddr_t id;
	uint8_t num;
};

static struct rxNumTableEntry rxNumTable[MAX_RXNUM_TABLE];
static const struct rxNumTableEntry* rxNumTableLast = &rxNumTable[MAX_RXNUM_TABLE];
uint8_t rxNumTablePos = 0; // Next pos to enter a new value into ::rxNumTable.

union {
	struct {
		uint8_t b2;
		uint8_t b1;
		uint8_t b4;
		uint8_t b3;
	};
	struct {
		uint16_t i1;
		uint16_t i2;
	};
} shiftbuf;

//TODO: PANICASSERT( WDS_RADIOPANIC, GET_ISR(TIMERB0_VECTOR)!=TX_STOPPED && GET_ISR(TIMERB0_VECTOR)!=TX_DONE );
#define RESTART_TX_TIMER()  do {    \
    TBCCR0 = TBR + RADIO_STROKE;  \
    TBCCTL0 &= ~CCIFG;            \
    TBCCTL0 |= CCIE;              \
    } while(0)

//TODO: PANICASSERT( WDS_RADIOPANIC, GET_ISR(TIMERB0_VECTOR)!=TX_WAITING );
#define STOP_TX_TIMER()  do { \
    TBCCTL0 &= ~(CCIFG+CCIE); \
  }while(0)

void _Net_logSent(char symbol, netheader_t* nh)
{
#if (DEBUGLOG_LEVEL & DEBUG_RADIO)
	if( DEBUGLOG_LEVEL_TESTHWRANGE( DEBUG_RADIO, MEDIUM, HIGH ) ) {
		printf("#[net S%c] N:%u F:%#.4x T:%#.4x P:%#.2x Fl:%#.2x L:%u\r\n",
				symbol, nh->_num, nh->from, nh->to,
				nh->type, nh->flags, nh->size - sizeof(netheader_t)
		);
	}
#endif
}

void _Net_logReceived(char symbol, netheader_t* rxHeader, uint16_t checkR,
		uint16_t checkC)
{
#if (DEBUGLOG_LEVEL & DEBUG_RADIO)
	// packet debug output
	if( DEBUGLOG_LEVEL_TESTHWRANGE(DEBUG_RADIO, MEDIUM, HIGH) ) {
		// show faulty packets only on high level
		if( DEBUGLOG_LEVEL_TEST(HIGH) && (symbol == '-') )
		return;
		printf("#[net R%c] N:%u F:%#.4x T:%#.4x P:%#.2x Fl:%#.2x L:%u R:%u Cr:%#x Cc:%#x\r\n",
				symbol, rxHeader->_num,
				rxHeader->from, rxHeader->to, rxHeader->type, rxHeader->flags,
				rxHeader->size - sizeof(netheader_t), rssi, checkR, checkC);
	}
#endif
}

/* ------------ FROM HERE CC1020 MAC LAYER (move) -------------- */

void _Net_logMacDump(char symbol, uint8_t* data, uint16_t length)
{
#if (DEBUGLOG_LEVEL & DEBUG_RADIO) && ((DEBUGLOG_LEVEL & DEBUGLOG_LEVEL_MASK) >= VERBOSE)
	// verbose logging -> dump complete packet
	if( DEBUGLOG_LEVEL_TESTHW(DEBUG_RADIO, VERBOSE) && length) {
		printf("#(%cX) ", symbol);
		// hex dump might be too long for Comm_print buffer
		String_writeHex(putchar, data, length);
		printf("\r\n");
	}
#endif
}

void TX_STOPPED() {
	//TODO: System_panic(WDS_RADIOPANIC, "TX_STOPPED"); // ADD
}

/*
 *   brief CRC generating lookup table.
 *   This mysterious table is just the CRC of each possible byte.  It can be
 *   computed using the standard bit-at-a-time methods.  The polynomial can
 *   be seen in entry 128, 0x8408.  This corresponds to x^0 + x^5 + x^12.
 *   Add the implicit x^16, and you have the standard CRC-CCITT.
 */
static uint16_t const ccitt_crc16_table[256] = { 0x0000, 0x1189, 0x2312, 0x329b, 0x4624,
		0x57ad, 0x6536, 0x74bf, 0x8c48, 0x9dc1, 0xaf5a, 0xbed3, 0xca6c, 0xdbe5,
		0xe97e, 0xf8f7, 0x1081, 0x0108, 0x3393, 0x221a, 0x56a5, 0x472c, 0x75b7,
		0x643e, 0x9cc9, 0x8d40, 0xbfdb, 0xae52, 0xdaed, 0xcb64, 0xf9ff, 0xe876,
		0x2102, 0x308b, 0x0210, 0x1399, 0x6726, 0x76af, 0x4434, 0x55bd, 0xad4a,
		0xbcc3, 0x8e58, 0x9fd1, 0xeb6e, 0xfae7, 0xc87c, 0xd9f5, 0x3183, 0x200a,
		0x1291, 0x0318, 0x77a7, 0x662e, 0x54b5, 0x453c, 0xbdcb, 0xac42, 0x9ed9,
		0x8f50, 0xfbef, 0xea66, 0xd8fd, 0xc974, 0x4204, 0x538d, 0x6116, 0x709f,
		0x0420, 0x15a9, 0x2732, 0x36bb, 0xce4c, 0xdfc5, 0xed5e, 0xfcd7, 0x8868,
		0x99e1, 0xab7a, 0xbaf3, 0x5285, 0x430c, 0x7197, 0x601e, 0x14a1, 0x0528,
		0x37b3, 0x263a, 0xdecd, 0xcf44, 0xfddf, 0xec56, 0x98e9, 0x8960, 0xbbfb,
		0xaa72, 0x6306, 0x728f, 0x4014, 0x519d, 0x2522, 0x34ab, 0x0630, 0x17b9,
		0xef4e, 0xfec7, 0xcc5c, 0xddd5, 0xa96a, 0xb8e3, 0x8a78, 0x9bf1, 0x7387,
		0x620e, 0x5095, 0x411c, 0x35a3, 0x242a, 0x16b1, 0x0738, 0xffcf, 0xee46,
		0xdcdd, 0xcd54, 0xb9eb, 0xa862, 0x9af9, 0x8b70, 0x8408, 0x9581, 0xa71a,
		0xb693, 0xc22c, 0xd3a5, 0xe13e, 0xf0b7, 0x0840, 0x19c9, 0x2b52, 0x3adb,
		0x4e64, 0x5fed, 0x6d76, 0x7cff, 0x9489, 0x8500, 0xb79b, 0xa612, 0xd2ad,
		0xc324, 0xf1bf, 0xe036, 0x18c1, 0x0948, 0x3bd3, 0x2a5a, 0x5ee5, 0x4f6c,
		0x7df7, 0x6c7e, 0xa50a, 0xb483, 0x8618, 0x9791, 0xe32e, 0xf2a7, 0xc03c,
		0xd1b5, 0x2942, 0x38cb, 0x0a50, 0x1bd9, 0x6f66, 0x7eef, 0x4c74, 0x5dfd,
		0xb58b, 0xa402, 0x9699, 0x8710, 0xf3af, 0xe226, 0xd0bd, 0xc134, 0x39c3,
		0x284a, 0x1ad1, 0x0b58, 0x7fe7, 0x6e6e, 0x5cf5, 0x4d7c, 0xc60c, 0xd785,
		0xe51e, 0xf497, 0x8028, 0x91a1, 0xa33a, 0xb2b3, 0x4a44, 0x5bcd, 0x6956,
		0x78df, 0x0c60, 0x1de9, 0x2f72, 0x3efb, 0xd68d, 0xc704, 0xf59f, 0xe416,
		0x90a9, 0x8120, 0xb3bb, 0xa232, 0x5ac5, 0x4b4c, 0x79d7, 0x685e, 0x1ce1,
		0x0d68, 0x3ff3, 0x2e7a, 0xe70e, 0xf687, 0xc41c, 0xd595, 0xa12a, 0xb0a3,
		0x8238, 0x93b1, 0x6b46, 0x7acf, 0x4854, 0x59dd, 0x2d62, 0x3ceb, 0x0e70,
		0x1ff9, 0xf78f, 0xe606, 0xd49d, 0xc514, 0xb1ab, 0xa022, 0x92b9, 0x8330,
		0x7bc7, 0x6a4e, 0x58d5, 0x495c, 0x3de3, 0x2c6a, 0x1ef1, 0x0f78 };

uint16_t System_ccittFcs(uint16_t fcs, uint8_t c)
{
	return (((fcs) >> 8) ^ ccitt_crc16_table[((fcs) ^ (c)) & 0xff]);
}

/**
* Initial FCS value, use this as starting value for fcs. in ::crc_calc.
*/
#define INIT_FCS  0xffff

#define TXMODE_BACKOFF  (0x01)

typedef enum
{
	TX_DONE_HANDLER = 0,
	TX_WAITING_HANDLER,
	TX_WAITACK_HANDLER,
	TX_BACKOFF_HANDLER,
	TX_STOPPED_HANDLER,
	TX_IDLE_HANDLER
} TimerBState;

static TimerBState currentTimerBState;
typedef void (*TimerBHandler)();
static TimerBHandler timerBHandler[6];

interrupt (TIMERB0_VECTOR) Timer_B0(void)
{
	timerBHandler[currentTimerBState]();
}


typedef enum
{
	radiotxdma_handler = 0,
	acktxdma_handler 
} DacDmaState;

static DacDmaState currentDacDmaState;
typedef void (*DacDmaHandler)();
static DacDmaHandler dacDmaHandler[2];

interrupt (DACDMA_VECTOR) DacDma(void)
{
	dacDmaHandler[currentDacDmaState]();
}


typedef enum
{
	rxreceivedHandler = 0,
	rxsynched_leftHandler,
	rxsynched0Handler,
	rxsynched_right0Handler,
	rxsynched_left0Handler,
	rxsearchingHandler,
	rxsynched_rightHandler
} Uart0RXState;

static Uart0RXState currentUart0RXState;
typedef void (*Uart0RXHandler)();
static Uart0RXHandler uart0RXHandler[7];

interrupt (USART0RX_VECTOR) Uart0RX(void)
{
	uart0RXHandler[currentUart0RXState]();
}



void TX_DONE() {
	// nothing happens here, just to mark this state
	
	/*
	This state is selected to Timer B0 after a transmission has been done.
	It's currently unclear why. Is it just to silently catch a timer interrupt
	or is it an error?
	*/
	// System_panic(WDS_RADIOPANIC, "TX_DONE");
}

// Currently waiting for an ACK.
// case TX_WAITACK: repeat packet if needed, go to idle if ack received.
void TX_WAITACK() {
	ackTimeout--;
	if(ackTimeout==0) {
		if(txCallbackArgs.timesRepeated > MAX_REPEATED) { // Unreachable destination
			//SET_ISR(TIMERB0_VECTOR, TX_DONE);
			currentTimerBState = TX_DONE_HANDLER;
			txBuf[0] = 0x00; // mark for no packet in txbuf
			txCallbackArgs.success = false;
			System_runtime_flag |= FLAG_RADIO_TX;
		} else {
			txCallbackArgs.timesRepeated++;
			//SET_ISR(TIMERB0_VECTOR, TX_WAITING);
			currentTimerBState = TX_WAITING_HANDLER;
			RESTART_TX_TIMER();
		}
	} else {
		RESTART_TX_TIMER();
	}
}

void radiotxdma() {
	netheader_t* txHeader = (netheader_t*)&txBuf[PREAMBLE_SIZE];

	DMA0CTL &= ~(DMAIFG + DMAIE);
	eint(); // eint part lasts ~7ms
	/////////////////// Reconfigure Port& for receive mode.
	P3SEL &= ~0x04; // Ensure Tx line is off
	P3SEL |= 0x0A; // select rx line and clk
	/////////////////// Switch transceiver to Rx
	CC1020_wakeupRX(RX_CURRENT);
	CC1020_setupRX(RX_CURRENT);
	LNA_POWER_ON;
	dint();

#if (DEBUGLOG_LEVEL & DEBUG_RADIO) && ((DEBUGLOG_LEVEL & DEBUGLOG_LEVEL_MASK) >= HIGH)
	if( DEBUGLOG_LEVEL_TESTHWRANGE(DEBUG_RADIO, HIGH, HIGH) ) {
		// notify about sending from queue
		printf("#[net S+] N:%u R:%u\r\n", txHeader->_num, txCallbackArgs.timesRepeated);
	}
#endif

	if( (txHeader->to != NETADDR_BROADCAST) &&
			((txHeader->flags & NET_TLFLAG_ACKREQ) == NET_TLFLAG_ACKREQ) )
	{
		// Not-Broadcasts need ack
		ackTimeout = 16; // 12; // ATTENTION!!
		//SET_ISR(TIMERB0_VECTOR, TX_WAITACK);
		currentTimerBState = TX_WAITACK_HANDLER;
		RESTART_TX_TIMER();
	} else {
		// Broadcasts dont need an ack, signal "sent"
		//SET_ISR(TIMERB0_VECTOR, TX_DONE); // set net state
		currentTimerBState = TX_DONE_HANDLER;
		txBuf[0] = 0x00; // mark for no packet in txbuf 
		txCallbackArgs.success = true;
		System_runtime_flag |= FLAG_RADIO_TX;
	}
	// isSending = 0; // no more sending
	ENABLE_RX_IRQ();
}

// Receiving side: Low level
//
// Preamble is 0xAA, Sync is 0xD391,
// Packet format:  AAAAAAAAAA D391 00 Length[1] To[2] From[2] Type[1] Num[1] 
//  
// 0  AA D3 91 00 | FF 00 |
// 1  55 69 C8 80 | 7F 80 |
// 2  AA B4 E4 40 | 3F C0 |
// 3  55 5A 72 20 | 1F E0 |
// 4  AA AD 39 10 | 0F F0 |
// 5  55 56 9C 88 | 07 F8 |
// 6  AA AB 4E 44 | 03 FC |
// 7  55 55 A7 22 | 01 FE |
// 8  AA AA D3 91 | 00 FF |
//
void acktxdma() {
	DMA0CTL &= ~(DMAIFG + DMAIE);
	eint(); // eint part lasts ~7ms
	/////////////////// Reconfigure Port& for receive mode.
	P3SEL &= ~0x04; // Ensure Tx line is off
	P3SEL |= 0x0A; // select rx line and clk
	/////////////////// Switch transceiver to Rx
	CC1020_wakeupRX(RX_CURRENT);
	CC1020_setupRX(RX_CURRENT);
	LNA_POWER_ON;
	dint();

	// ack has been sent completely

	// isSending = 0; // no more sending
	//SET_ISR(DACDMA_VECTOR, radiotxdma);
	currentDacDmaState = radiotxdma_handler;
	ENABLE_RX_IRQ();
}

enum CC1020_lock_status CC1020_setupTX(char TXANALOG) {
  enum CC1020_lock_status lock_status;
  // Setup bias current adjustment
  CC1020_writeRegister(CC1020_ANALOG,TXANALOG);
  // Switch into TX, switch to freq. reg B
  CC1020_writeRegister(CC1020_MAIN,0xC1);
  CC1020_delay(1); // Wait for 1 msec
  // Monitor LOCK, lasts 420 - 510 cycles @ 4505600 = 93 us - 113 us
  lock_status = _CC1020_waitlock();
  if( lock_status != LOCK_OK ) {
	  // (PLL out of LOCK)
	  // If recalibration ok
    if(CC1020_calibrate())
    	lock_status = LOCK_RECAL_OK; // Indicate PLL in LOCK
	// Else (recalibration failed)
  }
  // Restore PA_POWER
  CC1020_writeRegister(CC1020_PA_POWER, currentCC1020Config.registerValues[CC1020_PA_POWER]);
  // Turn OFF DCLK squelch in TX
  CC1020_writeRegister(CC1020_INTERFACE, 0x01);
  // Return LOCK status to application
  return (lock_status);
}

void _Net_sendACK(netheader_t* rxHeader)
{
	uint16_t pos;
	netheader_t* ackHeader;

	// Sending an ACK, first thing is to check if DMA controller
	// is already in use, if it is the TX state machine i currently sending
	// a packet. Should not happen when a recent packet was received, but
	// for safety:
	if (DMA0CTL & DMAEN)
		return;
	// Next thing is to set the ISR to acktxdma, this will block 
	// the tx machine from starting another dma transfer.
	//SET_ISR(DACDMA_VECTOR, acktxdma);
	currentDacDmaState = acktxdma_handler;
	// Start with 6 bytes preamble + 2 bytes sync word.
	((uint16_t*)ackBuf)[0] = 0xAAAA;
	((uint16_t*)ackBuf)[1] = 0xAAAA;
	((uint16_t*)ackBuf)[2] = 0xAAAA;
	((uint16_t*)ackBuf)[3] = 0x91D3;
	txBuf[8] = 0x00;
	pos = PREAMBLE_SIZE;

	// Queue header
	ackHeader = (netheader_t*)(ackBuf + pos);
	ackHeader->size = sizeof(netheader_t);
	ackHeader->flags = NET_TLFLAG_NOCRC;
	ackHeader->to = rxHeader->from;
	ackHeader->from = internalID;
	ackHeader->type = ACK_PACKET;
	ackHeader->_num = rxHeader->_num;
	pos += sizeof(netheader_t);
	// Add 2 Bytes FF as puffer.
	ackBuf[pos] = 0xFA;
	pos++;
	ackBuf[pos] = 0xFA;
	pos++;

#if (DEBUGLOG_LEVEL & DEBUG_RADIO)
	// low level debug
	_Net_logMacDump('T', (uint8_t*)&ackBuf[PREAMBLE_SIZE], pos - PREAMBLE_SIZE - 2);
	_Net_logSent('A', (netheader_t*)&ackBuf[PREAMBLE_SIZE]);
#endif

	// isSending = 1;
	DISABLE_RX_IRQ();
	_NOP();
	eint();
	// Switch transceiver to Tx
	LNA_POWER_OFF;
	CC1020_wakeupTX(TX_CURRENT);
	txLockStatus = CC1020_setupTX(TX_CURRENT) ;
	// Configure ports for Tx, Assumed UART is in SPI mode for chipcon receive
	P3SEL &= ~0x02; // Ensure Rx line is off
	P3SEL |= 0x0C; // select Tx line and clk
	U0CTL |= SWRST; // NEW    TODO Uart reinit?
	IFG1 &= ~UTXIFG0; // Reset IFG.
	DMACTL0 = DMA0TSEL_4; // Configure DMA Channel 0 for UART0 TXIFG.
	DMACTL1 = 0x0000; // No DMAONFETCH, ROUNDROBIN, ENNMI.
	// Single transfer Mode, dstadr unchanged, srcadr incremented, byte access
	// Important to use DMALEVEL when using USART TX interrupts so first edge 
	// doesn't get lost (hangs every 50. - 100. time)!
	DMA0CTL = DMADT_0 + DMADSTINCR_0 + DMASRCINCR_3 + DMASBDB + DMALEVEL + DMAIE;
	DMA0SA = (uint16_t) ackBuf;
	DMA0DA = (uint16_t) &TXBUF0;
	DMA0SZ = pos;
	// TODO ????
	// dint(); _NOP();
	U0CTL &= ~SWRST; // NEW
	DMA0CTL |= DMAEN;
}

volatile char rxoffset;

void rxreceived();

void rxsynched_checkdone() {
	uint8_t size;
	if( rxBufPos < sizeof(netheader_t) ) {
		size = sizeof(netheader_t);
	} else {
		size = rxBuf->buffer.decoded.nwh.size;
	}
	if( rxBufPos >= size )
		// SET_ISR(USART0RX_VECTOR, rxreceived);
		currentUart0RXState = rxreceivedHandler;
}

// receiving: offset with left-shift
void rxsynched_left() {
	shiftbuf.b3 = shiftbuf.b4;
	shiftbuf.b4 = RXBUF0;
	shiftbuf.i1 = shiftbuf.i2 << rxoffset;
	rxBuf->buffer.raw[rxBufPos++] = shiftbuf.b1;
	NETSYN_DEBUG("%.2x ", shiftbuf.b1);
	rxsynched_checkdone();
}

// this is to ignore the last sync-word byte
void rxsynched_left0() {
	shiftbuf.b3 = shiftbuf.b4;
	shiftbuf.b4 = RXBUF0;
	shiftbuf.i1 = shiftbuf.i2 << rxoffset;
	//SET_ISR(USART0RX_VECTOR, rxsynched_left);
	currentUart0RXState = rxsynched_leftHandler;
}

void rxsynched0();
void rxsynched_right0();

void rxsearching() {
	shiftbuf.b1 = shiftbuf.b2;
	shiftbuf.b2 = shiftbuf.b3;
	shiftbuf.b3 = shiftbuf.b4;
	shiftbuf.b4 = RXBUF0;
		
	if(shiftbuf.i1==0xAAD3 && shiftbuf.i2==0x9100) {		// AA D3 91 00 | FF 00 |
		// SET_ISR(USART0RX_VECTOR, rxsynched0);
		currentUart0RXState = rxsynched0Handler;
		//rxBuf->state = 00;
		//rxBuf->buffer.raw[rxBufPos++] = shiftbuf.b4;
		rxoffset = 0;
	}
	else if(shiftbuf.i1==0x5569 && shiftbuf.i2==0xC880) {	// 55 69 C8 80 | 7F 80 |
		rxoffset = -1;
	}
	else if(shiftbuf.i1==0xAAB4 && shiftbuf.i2==0xE440) { // 2  AA B4 E4 40 | 3F C0 |
		rxoffset = -2;
	}
	else if(shiftbuf.i1==0x555A && shiftbuf.i2==0x7220) { // 3  55 5A 72 20 | 1F E0 |
		rxoffset = -3;
	}
	else if(shiftbuf.i1==0xAAAD && shiftbuf.i2==0x3910) { // 4  AA AD 39 10 | 0F F0 |
		rxoffset = -4;
	}
	else if(shiftbuf.i1==0x5556 && shiftbuf.i2==0x9C88) { // 5  55 56 9C 88 | 07 F8 |
		rxoffset = 3;
	}
	else if(shiftbuf.i1==0xAAAB && shiftbuf.i2==0x4E44) { // 6  AA AB 4E 44 | 03 FC |
		rxoffset = 2;
	}
	else if(shiftbuf.i1==0x5555 && shiftbuf.i2==0xA722) { // 7  55 55 A7 22 | 01 FE |
		rxoffset = 1;
	} else {
		return;
	}
	NETSYN_DEBUG("s %i:", rxoffset);
	
	if( rxoffset > 0 ) {
		//SET_ISR(USART0RX_VECTOR, rxsynched_right0);
		currentUart0RXState = rxsynched_right0Handler;
	} else if( rxoffset < 0 ) {
		rxoffset = -rxoffset;
		//SET_ISR(USART0RX_VECTOR, rxsynched_left0);
		currentUart0RXState = rxsynched_left0Handler;
	}
	time_t time = rxTime;
	RTC_GetTime(&time); //TODO: RTC_GetTime(&rxTime); 
	rxTime = time;
	rssi = CC1020_readRegister(CC1020_RSS);
}

void rxreceived() {
	uint16_t checkR = 0; // received checksum
	uint16_t checkC = 0; // calculated checksum
	bool invoke_handler; // invoke system packet handler or not
	netheader_t* rxHeader // network layer packet header
	= &rxBuf->buffer.decoded.nwh;
	netheader_t* txHeader;
	int i;

	NETSYN_DEBUG("rx\r\n");
	
	// low level debug
	_Net_logMacDump('R', rxBuf->buffer.raw, rxHeader->size);

	// test checksum
	if( (rxHeader->flags & NET_TLFLAG_NOCRC) == 0 ) {
		uint8_t* x;

		// read received crc
		checkR = rxBuf->buffer.raw[rxHeader->size - 2] << 8;
		checkR |= rxBuf->buffer.raw[rxHeader->size - 1];

		// checksum done, "remove" from packet
		rxHeader->size -= 2;

		// calculate Checksum
		for(i = 0, checkC = INIT_FCS, x = rxBuf->buffer.raw; i < rxHeader->size; i++, x++)
		checkC = System_ccittFcs(checkC, *x);
	}

	// packet debug output
#if (DEBUGLOG_LEVEL & DEBUG_RADIO)
	if( DEBUGLOG_LEVEL_TESTHWRANGE(DEBUG_RADIO, MEDIUM, HIGH) ) {
		char c;
		if( rxHeader->type == ACK_PACKET )
		c = 'A';
		else if( checkR == checkC )
		c = '+';
		else
		c = '-';
		_Net_logReceived( c, rxHeader, checkR, checkC );
	}
#endif

	if( checkR == checkC ) {

		/* ---- network layer packet handler  ---- */
		switch(rxHeader->type) {
			case ACK_PACKET:
			// Waiting for this ack?!? This part perhaps not irq safe ?! TODO
			txHeader = (netheader_t*)&txBuf[PREAMBLE_SIZE];
			if( rxHeader->from == txHeader->to ) {
				//if( (GET_ISR(TIMERB0_VECTOR) == TX_WAITACK) &&
				if( (currentTimerBState == TX_WAITACK_HANDLER) &&
						(rxHeader->_num == sentPacket._num) )
				{
					STOP_TX_TIMER();
					// SET_ISR(TIMERB0_VECTOR, TX_DONE);
					currentTimerBState = TX_DONE_HANDLER;
					txBuf[0] = 0x00; // mark for no packet in txbuf
					txCallbackArgs.success = true;
					System_runtime_flag |= FLAG_RADIO_TX;
				} else {
#if (DEBUGLOG_LEVEL & DEBUG_RADIO)
					LOG_HARDWARE(DEBUG_RADIO, "#(RX) ACK in wrong state", NULL);
#endif
				}
			}
			// do not pass ack packets to the application
			invoke_handler = false;
			break;

			default:
			// send ack if requested
			if( (rxHeader->to == internalID) &&
					((rxHeader->flags & NET_TLFLAG_ACKREQ) == NET_TLFLAG_ACKREQ)
			)
			{
				_Net_sendACK(rxHeader);
			}

			invoke_handler = true;
			break;
		}

	} else {
		// checksum mismatch
		invoke_handler = false;
	}

	/* ---- clean up buffers  ---- */
	dint(); _NOP();

	if( invoke_handler ) {
		// this buffer is ready
		rxBuf->state = 0x01;
		// need to switch to other buffer to keep the current packet for the handler
		// TODO check and warn if buffer is not already empty
		rxBuf = (rxBuf == &rxBuf1) ? &rxBuf2 : &rxBuf1;
		rxBufPos = 0;
		// pass packet to application
		System_runtime_flag |= FLAG_RADIO_RX;
	} else {
		rxBufPos = 0;
	}

	/* ---- set ready to receive  ---- */
	// SET_ISR(USART0RX_VECTOR, rxsearching);
	currentUart0RXState = rxsearchingHandler;
	IE1 |= URXIE0; // Enable interrupt.
	IFG1 &= ~URXIFG0;
	DIRQEND();
}

// receiving: no offset
void rxsynched0() {
	rxBuf->buffer.raw[rxBufPos++] = RXBUF0;
	NETSYN_DEBUG("%.2x ", rxBuf->buffer.raw[rxBufPos]);
	rxsynched_checkdone();
}

// receiving: offset with right-shift
void rxsynched_right() {
	shiftbuf.b3 = shiftbuf.b4;
	shiftbuf.b4 = RXBUF0;
	shiftbuf.i1 = shiftbuf.i2 >> rxoffset;
	rxBuf->buffer.raw[rxBufPos++] = shiftbuf.b2;
	NETSYN_DEBUG("%.2x ", shiftbuf.b2);
	rxsynched_checkdone();
}

// this is to ignore the last sync-word byte
void rxsynched_right0() {
	shiftbuf.b3 = shiftbuf.b4;
	shiftbuf.b4 = RXBUF0;
	shiftbuf.i1 = shiftbuf.i2 >> rxoffset;
	// SET_ISR(USART0RX_VECTOR, rxsynched_right);
	currentUart0RXState = rxsynched_rightHandler;
}

void TX_BACKOFF();

// Channel busy, waiting for free channel.
// case TX_WAITING: look for free chanel (rxState = RX_SEARCHING) and go to TX_BACKOFF
void TX_WAITING() {
	//if(GET_ISR(USART0RX_VECTOR)==rxsearching && (CC_LOCK)) {
	if(currentUart0RXState==rxsearchingHandler && (CC_LOCK)) {
		txBackoff = (((TBR + TAR) & 0x001E) + 4);
		// SET_ISR(TIMERB0_VECTOR, TX_BACKOFF);
		currentTimerBState = TX_BACKOFF_HANDLER;
	}
	RESTART_TX_TIMER();
}

// Channel clear, waiting for random backoff time elapsed.
// case TX_BACKOFF: wait until backoff time elapsed, then start sending
void TX_BACKOFF() {
	// If we are currently sending an ack (no matter how this
	// could happen?!)  go back to TX_WAITING.
	//if(GET_ISR(DACDMA_VECTOR)==acktxdma) {
	if(currentDacDmaState==acktxdma_handler) {
		// SET_ISR(TIMERB0_VECTOR, TX_WAITING);
		currentTimerBState = TX_WAITING_HANDLER;
		RESTART_TX_TIMER();
	}
	//else if((GET_ISR(USART0RX_VECTOR)!=rxsearching || (!(CC_LOCK))) && (txMode & TXMODE_BACKOFF)) {
	else if((currentUart0RXState!=rxsearchingHandler || (!(CC_LOCK))) && (txMode & TXMODE_BACKOFF)) {
		// SET_ISR(TIMERB0_VECTOR, TX_WAITING);
		currentTimerBState = TX_WAITING_HANDLER;
		RESTART_TX_TIMER();
	}
	else if(txBackoff==0) { // Start sending.
		// REENTRANT!, case needs longer than 7ms
		STOP_TX_TIMER();
		DISABLE_RX_IRQ();
		_NOP();
		eint();
		v0 = TAR;
		// isSending = 1;
		// Switch transceiver to Tx
		LNA_POWER_OFF;
		CC1020_wakeupTX(TX_CURRENT);
		txLockStatus = CC1020_setupTX(TX_CURRENT);

		// low level debug
		_Net_logMacDump('T', (uint8_t*)&txBuf[PREAMBLE_SIZE], txBufLength - PREAMBLE_SIZE - 2);

		// Configure ports for Tx, Assumed UART is in SPI mode for chipcon receive
		P3SEL &= ~0x02; // Ensure Rx line is off
		P3SEL |= 0x0C; // select Tx line and clk
		U0CTL |= SWRST; // NEW
		IFG1 &= ~UTXIFG0; // Reset IFG.
		DMACTL0 = DMA0TSEL_4; // Configure DMA Channel 0 for UART0 TXIFG.
		DMACTL1 = 0x0000; // No DMAONFETCH, ROUNDROBIN, ENNMI.
		// Single transfer Mode, dstadr unchanged, srcadr incremented, byte access
		// Important to use DMALEVEL when using USART TX interrupts so first edge 
		// doesn't get lost (hangs every 50. - 100. time)!
		DMA0CTL = DMADT_0 + DMADSTINCR_0 + DMASRCINCR_3 + DMASBDB + DMALEVEL + DMAIE;
		DMA0SA = (uint16_t) txBuf;
		DMA0DA = (uint16_t) &TXBUF0;
		DMA0SZ = txBufLength;
		U0CTL &= ~SWRST; // NEW, needed??
		DMA0CTL |= DMAEN; // Start tx, when finished ends in radiotxdma
	}
	// Backoff time still running, decrease counter and restart timer.
	else {
		txBackoff--;
		RESTART_TX_TIMER();
	}
}

#define TXMODE_LP       (0x02)

void TX_IDLE() {
	// If IDLE and something is in buf prepare packet in txBuf and
	// go to txState=TX_WAITING to wait for a free channel.
	if(numPackets > 0) {
		int pos; // position in txbuf
		uint8_t x;
		uint16_t ix;
		uint16_t length;
		uint16_t checksum = INIT_FCS;

		/*
		 Prepare txBuf with packet to send:
		 - 8 byte preamble and sync word
		 - 8 byte header (in ringbuf txPackets)
		 - x byte data   ( "    "       "     )
		 - 2 byte checksm( "    "       "     )
		 dont forget to take 2 bytes callback off ringbuf
		 */

		// Start with 6 bytes preamble + 3 bytes sync word
		((uint16_t*)txBuf)[0] = 0xAAAA;
		((uint16_t*)txBuf)[1] = 0xAAAA;
		((uint16_t*)txBuf)[2] = 0xAAAA;
		((uint16_t*)txBuf)[3] = 0x91D3;
		// third byte saves masking operations in rxsearching
		txBuf[8] = 0x00;
		pos = PREAMBLE_SIZE;

		// take preamble and header from queue, calc checksum
		length = PREAMBLE_SIZE + sizeof(netheader_t);
		for(; pos < length; pos++ ) {
			RB_GET(txPackets, x);
			txBuf[pos] = x;
			checksum = System_ccittFcs(checksum, x);
		}

		// Copy header
		memcpy((void*)&sentPacket, (void*)&txBuf[PREAMBLE_SIZE], sizeof(netheader_t));

		// fetch payload
		length = PREAMBLE_SIZE + sentPacket.size;
		for(; pos < length; pos++ ) {
			RB_GET(txPackets, x);
			txBuf[pos] = x;
			checksum = System_ccittFcs(checksum, x);
		}

		// append CRC and modify (!) size    
		if( (sentPacket.flags & NET_TLFLAG_NOCRC) == 0 ) {
			txBuf[pos++] = checksum >> 8;
			txBuf[pos++] = checksum;
			txBuf[PREAMBLE_SIZE] += 2;
		}

		// Add 2 Bytes FF as puffer.
		txBuf[pos] = 0xFA; pos++;
		txBuf[pos] = 0xFA; pos++;
		txBufLength = pos;

		RB_GET(txPackets, x); ix = (x << 8);
		RB_GET(txPackets, x); ix |= x;
		txCallback = (network_send_callback_t)ix;

		numPackets--;
		txCallbackArgs.timesRepeated = 0;
		// go into waiting
		if(txMode & TXMODE_BACKOFF) {
			// SET_ISR(TIMERB0_VECTOR, TX_WAITING);
			currentTimerBState = TX_WAITING_HANDLER;
		} else {
			// txMode==TXMODE_NOBACKOFF
			// SET_ISR(TIMERB0_VECTOR, TX_BACKOFF);
			currentTimerBState = TX_BACKOFF_HANDLER;
			txBackoff = 1; // No backoff
			// Dont clear if here so state TX_BACKOFF will run
			// directly after this interrupt
		}
		// Go directly to TX_WAITING or TX_BACKOFF
		TBCCTL0 &= ~(CCIE + CCIFG);
		TBCCTL0 |= CCIE + CCIFG;
	} else {
		// If no packet has to be sent stop here and do nothing.
		// TimerTx interrupt will be reenabled if a packet is queued by
		// radio_queue_packet.
		STOP_TX_TIMER();
		// SET_ISR(TIMERB0_VECTOR, TX_STOPPED);
		currentTimerBState = TX_STOPPED_HANDLER;
		//RadioOFF
		if(txMode & TXMODE_LP) {
			CC1020_setupPD();
			LNA_POWER_OFF;
		}
	}
}


bool Net_init()
{
	CC1020_init();
	return true;
}

// name UART0 Switching functions
void _Net_resetU0()
{
	//Reset SEL for P3[1-3] (CC DIO, DIO, DCLK) and P3[4-5] (Camera Rx+Tx)
	P3SEL &= ~0x3E;
	IFG1 &= ~(UTXIE0 + URXIE0);
	ME1 &= ~(UTXE0 + URXE0); // Disable Uart0 Tx + Rx
	UCTL0 = SWRST; // U0 into reset state.
}

/*
 *   brief Initializes UART0 for receiving from Chipcon
 *
 *   Data (for transmit+receive) must be stable at the rising edge of the clock.
 *   This means:
 *     Transmit : Set bit on falling edge of clock.
 *     Receive  : Read bit on rising edge of clock.
 *
 *   CKPH: Normal UCLK clocking scheme.
 *   CKPL: 1 The inactive level is high; data is output with the falling edge of
 *         UCLK; input data is latched with the rising edge of UCLK.
 */
void _Net_initU0()
{
	_Net_resetU0();
	UCTL0 |= CHAR + SYNC; // 8-bit character, SPI, Slave mode
	UTCTL0 = CKPL + STC; // CKPH works also, but not CKPH+CKPL or none of them!!
	URCTL0 = 0x00;
	UBR00 = 0x00; // No baudrate divider 
	UBR10 = 0x00; // settings for a spi
	UMCTL0 = 0x00; // slave.
	ME1 |= USPIE0; // Enable USART0 TXD/RXD, disabling does not yield any powersavings
	P3SEL |= 0x0A; // Select rx line and clk
	UCTL0 &= ~SWRST; // Last thing is clear reset bit
	IE1 |= URXIE0; // Enable interrupt.
}

bool Net_on()
{
	if (radioEnabled)
		return true;
	LNA_POWER_ON;
	// Transmitting side vars
	txMode = TXMODE_BACKOFF;
	numPackets = 0;
	txBackoff = 1; //Init
	// Receiving side vars
	rxBuf1.state = 0xFF;
	rxBuf2.state = 0xFF;
	rxBuf = &rxBuf1;
	rxBufPos = 0;
	// Interrupt routings
	//SET_ISR(USART0RX_VECTOR, rxsearching);
	currentUart0RXState = rxsearchingHandler;
	STOP_TX_TIMER();
	// SET_ISR(TIMERB0_VECTOR, TX_STOPPED);
	currentTimerBState = TX_STOPPED_HANDLER;
	// SET_ISR(DACDMA_VECTOR, radiotxdma);
	currentDacDmaState = radiotxdma_handler;
	// Ring buffers  
	RB_INIT(txPackets);
	// Uart
	_Net_initU0();
	CC1020_setupRX(RX_CURRENT);
	if (txMode & TXMODE_LP) {
		CC1020_setupPD();
		LNA_POWER_OFF;
	}
	radioEnabled = true;
	return true;
}

static bool _Net_HistoryFind(netheader_t* nwh) {
	register struct rxNumTableEntry* e;
	
	for(e = rxNumTable; e < rxNumTableLast; e++) {
		if( (e->id == nwh->from) && (e->num == nwh->_num) )
			return true;
	}
	return false;
}

static bool _Net_HistoryAdd(netheader_t* nwh) {
	register struct rxNumTableEntry* e;
	
	for(e = rxNumTable; e < rxNumTableLast; e++) {
		if(e->id == nwh->from) {
			// update existing entry
			e->num = nwh->_num;
			return true;
		}
	}
	// If no entry found from rxPacket.from create a new one..
	e = &rxNumTable[rxNumTablePos];
	rxNumTablePos++;
	if (rxNumTablePos == MAX_RXNUM_TABLE)
		rxNumTablePos = 0;
	e->id = nwh->from;
	e->num = nwh->_num;
	return false;
}

typedef struct{
	uint8_t packetType;
	network_packet_handler_t packetHandler;
} netpacket_rxHandler;

#define RXHandlerCount 8
netpacket_rxHandler RXHandler[RXHandlerCount];

enum netaddr_compareresult
Net_compareAddress(const netaddr_t (*const adr), const bool directConnection)
{
	if (*adr == internalID)
		return ADRCOMP_THIS;

	if (directConnection) {
		if( (*adr == 0) )
			return ADRCOMP_THIS;
	} else if( *adr == 0 )
		return ADRCOMP_ERROR;
	if( *adr == NETADDR_BROADCAST )
		return ADRCOMP_BROADCAST;
	else
		return ADRCOMP_OTHER;
}

bool sniffer_mode;

static void _Net_packetHandler(rxbuffer_t* fullBuf)
{
	netheader_t* nwh = &fullBuf->buffer.decoded.nwh;
	/*
	 second part of network layer radio handler. this part is invoked by the
	 "scheduler".
	 */

	// Packet already received?
	if( _Net_HistoryFind(nwh) ) {
		// We already received that, so do nothing.
		return;
	}

	// First time received: enter current num into history.
	_Net_HistoryAdd(nwh);
	
	enum netaddr_compareresult cpaResult;
	cpaResult = Net_compareAddress((netaddr_t*)&nwh->to, false);
	if (cpaResult == ADRCOMP_ERROR ) {
		return;
	}
	
	if(cpaResult == ADRCOMP_THIS || cpaResult == ADRCOMP_BROADCAST) {
		uint8_t rxType = nwh->type;
		uint8_t i;
		for(i=0; i < RXHandlerCount; i++) {
			if(RXHandler[i].packetType == rxType) {
				RXHandler[i].packetHandler(nwh->from, fullBuf->buffer.decoded.payload);
				return;
			}
		}
		return;
	}
	
	if(cpaResult == ADRCOMP_OTHER && sniffer_mode == true) {
		uint8_t rxType = nwh->type;

		uint8_t i;
		for(i=0; i < RXHandlerCount; i++) {
			if(RXHandler[i].packetType == rxType) {
				RXHandler[i].packetHandler(nwh->from, fullBuf->buffer.decoded.payload);
				return;
			}
		}	
	}
}

bool Net_rxHandler()
{
	//////////////////// Look into both buffers for something to handle
	do {
		if (rxBuf1.state == 0x01) {
			_Net_packetHandler(&rxBuf1);
			rxBuf1.state = 0xFF; // Set buffer empty again.
		} else if (rxBuf2.state == 0x01) {
			_Net_packetHandler(&rxBuf2);
			rxBuf2.state = 0xFF;
		} else {
			// TODO: CLEAR into Systems
			CLEAR(System_runtime_flag, FLAG_RADIO_RX);
			return true;
		}
	} while (1);
}

bool Net_txHandler()
{
	if(txCallback != NULL) {
		txCallback(txCallbackArgs.success);
	}
	// SET_ISR(TIMERB0_VECTOR, TX_IDLE);
	currentTimerBState = TX_IDLE_HANDLER;
	RESTART_TX_TIMER();
	return true;
}

bool Net_send(netpacket_send_args_t* args)
{
	int i;
	uint8_t x;
	uint16_t size;
	if (!radioEnabled) return false;

	size = sizeof(netheader_t);
	for( i = 0; i < NETWORK_PACKET_PAYLOAD_COUNT; i++ ) {
		size += args->payload[i].Size;
	}
	if ((254 - RB_LENGTH(txPackets)) > size) {
		// we have space in buf, so queue it
		args->netheader.size = size;

		
		// Header
		args->netheader._num = ++txNum;
		for (i = 0; i < sizeof(netheader_t); i++) {
			RB_PUT(txPackets, ((uint8_t*)&args->netheader)[i]);
		}

		// Payload
		for( x = 0; x < NETWORK_PACKET_PAYLOAD_COUNT; x++ ) {
			for (i = 0; i < args->payload[x].Size; i++) {
				RB_PUT(txPackets, ((uint8_t*)args->payload[x].Buffer)[i]);
			}
		}
		
		// Queue function callback 
		x = ((uint16_t)args->fpcallback) >> 8;
		RB_PUT(txPackets, x);
		// upper 8 bits of callback
		x = ((uint16_t)args->fpcallback) & 0x00FF;
		RB_PUT(txPackets, x);
		// lower 8 bits of callback

		numPackets++;

		// Debug output
		_Net_logSent('Q', &args->netheader);

		// If first packet in buf immediatly start Interrupt for sending
		// if(GET_ISR(TIMERB0_VECTOR)==TX_STOPPED) {
		if(currentTimerBState == TX_STOPPED_HANDLER) {
			//RadioON
			if(txMode & TXMODE_LP) {
				//uint16_t t1, t2, t3, t4;
				//t1 = TAR;
				//t3 = TBR;
				CC1020_wakeupRX(RX_CURRENT);
				CC1020_calibrate();
				CC1020_wakeupTX(TX_CURRENT);
				CC1020_calibrate();
				CC1020_wakeupRX(RX_CURRENT);
				CC1020_setupRX(RX_CURRENT);
				LNA_POWER_ON;
				//t2 = TAR;
				//t4 = TBR;
			}
			dint();
			_NOP();
			TBCCTL0 &= ~(CCIE + CCIFG);
			//SET_ISR(TIMERB0_VECTOR, TX_IDLE);
			currentTimerBState = TX_IDLE_HANDLER;
			TBCCTL0 |= CCIE + CCIFG;
		}
		eint();
		return true;
	} else {
		// Not enough space in ringbuf!
		// not queued source->destination P:protocol F:flags N:number L:length
		_Net_logSent('N', &args->netheader);

		/// TODO  2005-12-7 ONLY QUICK HACK TO REBOOT DEVICE WHEN TX STATE MACHINE IS HANGING IN TX_WAITING:
		/// TODO  2007-08-16 WHY?
		/*
		if(GET_ISR(TIMERB0_VECTOR) == TX_WAITING) {
			char* msg = "#(MACTX) DEADLOCK TX_WAITING";
			LOG_HARDWARE(DEBUG_RADIO, msg, NULL);
			System_panic(WDS_RADIOPANIC, msg);
		}
		*/
		return false;
	}
}

void Network_Init(uint8_t nodeID) {
	internalID = nodeID;
	currentCCpower = PA_POWER;
	currentCCchannel = 1;
	sniffer_mode = false;
	
	CC1020_reset();
	CC1020_setupPD();

	timerBHandler[0] = TX_DONE;
	timerBHandler[1] = TX_WAITING;
	timerBHandler[2] = TX_WAITACK;
	timerBHandler[3] = TX_BACKOFF;
	timerBHandler[4] = TX_STOPPED;
	timerBHandler[5] = TX_IDLE;
	
	dacDmaHandler[0] = radiotxdma;
	dacDmaHandler[1] = acktxdma;
	
	uart0RXHandler[0] = rxreceived;
	uart0RXHandler[1] = rxsynched_left;
	uart0RXHandler[2] = rxsynched0;
	uart0RXHandler[3] = rxsynched_right0;
	uart0RXHandler[4] = rxsynched_left0;
	uart0RXHandler[5] = rxsearching;
	uart0RXHandler[6] = rxsynched_right;
	
	Net_init();
	Net_on( TXMODE_BACKOFF );
}

void Network_InitSendArgs(network_payload_t* payload) {
	uint8_t i;
	for(i = 0; i < NETWORK_PACKET_PAYLOAD_COUNT; i++) {
		payload->Buffers[i].Size = 0;
		payload->Buffers[i].Buffer = NULL;
		
	}
}

bool Network_SendWithoutACK(uint8_t destination, uint8_t packetType, network_payload_t* payload) {
	netpacket_send_args_t internalPacket;
	
	internalPacket.netheader.to = destination;
	internalPacket.netheader.from = internalID;
	//if(internalPacket.netheader.to == 0) internalPacket.netheader.to = 255;
	//if(internalPacket.netheader.to == 255) internalPacket.netheader.to = 0;
	internalPacket.netheader.type = packetType;
	internalPacket.netheader.flags = 0;
	internalPacket.fpcallback = NULL;
		
	uint8_t i;
	for(i = 0; i < NETWORK_PACKET_PAYLOAD_COUNT; i++) {
		internalPacket.payload[i].Size = payload->Buffers[i].Size;
		internalPacket.payload[i].Buffer = payload->Buffers[i].Buffer;
	}
	return Net_send(&internalPacket);
}

#define NET_TLFLAG_ACKREQ 0x02

void Network_SendWithACK(uint8_t destination, uint8_t packetType, network_payload_t* payload, network_send_callback_t callback) {
	
	netpacket_send_args_t internalPacket;
	internalPacket.netheader.to = destination;
	internalPacket.netheader.from = internalID;
	//if(internalPacket.netheader.to == 0) internalPacket.netheader.to = 255;
	//if(internalPacket.netheader.to == 255) internalPacket.netheader.to = 0;
	internalPacket.netheader.type = packetType;
	internalPacket.netheader.flags = NET_TLFLAG_ACKREQ;
	internalPacket.fpcallback = callback;
		
	uint8_t i;
	for(i = 0; i < NETWORK_PACKET_PAYLOAD_COUNT; i++) {
		internalPacket.payload[i].Size = payload->Buffers[i].Size;
		internalPacket.payload[i].Buffer = payload->Buffers[i].Buffer;
	}
	
	Net_send(&internalPacket);
}

void Network_RegisterHandler(uint8_t packetType, network_packet_handler_t packetHandler) {
	uint8_t i;
	for(i=0; i < RXHandlerCount; i++) {
		if(RXHandler[i].packetType == 0){
			RXHandler[i].packetType = packetType;
			RXHandler[i].packetHandler = packetHandler;
			return;
		}
	}

}

void Network_SetFrequencyChannel(uint8_t channel) {
	uint8_t t_currentCCchannel = CC1020_selectChannel(channel);
	if(t_currentCCchannel != 0) currentCCchannel = t_currentCCchannel;
}

uint8_t Network_GetFrequencyChannel() {
	return currentCCchannel;
}

void Network_SetTransmitPower(uint8_t txpower) {
	//currentCCpower = txpower;
	//CC1020_setTxPower(txpower);
}

void Network_SetNodeID(uint8_t nodeID) {
	internalID = nodeID;
}

void Network_SnifferMode(bool enabled) {
	sniffer_mode = enabled;
}

/*
enum netheader_flags {
	NET_TLFLAG_NOCRC	= 0x01,	///< do no error checking
	NET_TLFLAG_ACKREQ	= 0x02	///< request ack for this packet	
};
*/
