#include "System.h"

void Programming_Infomem_Read(uint16_t offset, void* buffer, uint8_t size) {
	uint8_t* address = (uint8_t*)INFOMEM_START + offset;
	memcpy(buffer, address, size);
}

bool Programming_Infomem_Write(uint16_t offset, uint8_t count, ...) {
	char backup[INFOMEM_BLOCK_SIZE];
	
	register uint8_t* buffer;
	register uint16_t i;
	uint8_t*	 flash;

	if( offset > ( 2 * INFOMEM_BLOCK_SIZE) )
		return false;

	flash = (uint8_t*)INFOMEM_START;

	if (offset >= INFOMEM_BLOCK_SIZE) {
		flash = (uint8_t*) (INFOMEM_START + INFOMEM_BLOCK_SIZE);
		offset -= INFOMEM_BLOCK_SIZE;
	}

	dint();

	// backup into RAM
	memcpy(backup, flash, INFOMEM_BLOCK_SIZE);

	// merge backup with new data
	va_list argp;
	va_start(argp, count);

	buffer = (uint8_t*)backup + offset;
	for( i = 0; i < count; i++ ) {
		register uint16_t	size;
		register uint8_t* data;

		data = va_arg(argp, uint8_t*);
		size = va_arg(argp, uint16_t);
		memcpy(buffer, data, size);
		buffer += size;
	}
	va_end(argp);

	// setup the clock for flash memory operations
	FCTL2 = FWKEY | FSSEL_SMCLK | FN3;				//Added by Hutta // flash controller clock: 7372000 / 16 = 460kHz
													// TODO: check whether the power consumption increases: maybe it should be stopped after the update of the infomem

	// init flash access
    FCTL3 = FWKEY;

	// erase flash
    FCTL1 = FWKEY | ERASE;

    *flash = 0;

	// write flash
    FCTL1 = FWKEY | WRT;

    buffer = (uint8_t*)backup;
	
    for( i = 0; i < INFOMEM_BLOCK_SIZE; i++ ) {
		*flash = *buffer;
		buffer++; flash++;
	}
    
    FCTL1 = FWKEY;
    FCTL3 = FWKEY | LOCK;

	eint();
	return true;
}


extern int __bootloader_start;


void Programming_ProcessBootloaderInfo() {
	// bootloader state is stored at the beginning of the infomem area
	// the data doesn't have to be copied, direct access via pointer
	bootloader_offline_state_t* bootloader_state = (bootloader_offline_state_t*) INFOMEM_START;

	// find the bootloader header, it is at the fixed position defined in the linker script
	appHeader_t* _bootloaderHeader = (appHeader_t*) 0x2500;	//TODO: it should be enough to use the define from the linker script, not available yet due to forbidden x-references
	
	// test the checksum of the bootloader header
	uint8_t bld_header_checksum = _bootloaderHeader->type + _bootloaderHeader->appVer[0] + _bootloaderHeader->appVer[1];
	
	if (_bootloaderHeader->checksum != bld_header_checksum) {
		LOG("Bootloader not detected.");
		return;
	}
		
	// bootloader was detected.
	
	LOG("Bootloader V%i.%i found.", _bootloaderHeader->appVer[0],  _bootloaderHeader->appVer[1]);
		
	// test the checksum of the saved bootloader state
	uint8_t bld_checksum = bootloader_state->Flashed + bootloader_state->UsingGoldenImage + _bootloaderHeader->appVer[1] + _bootloaderHeader->appVer[0];

	if (bld_checksum != bootloader_state->Checksum) {
		LOG("ERROR: Bootloader state not found.");
		return;
	}


	// both preconditions are met: bootloader was detected and its state was found

	if (bootloader_state->FailStartsCounter > 0) {
		bootloader_offline_state_t new_bootloader_state;
		
		new_bootloader_state.Flashed 					= bootloader_state->Flashed;
		new_bootloader_state.FailStartsCounter 			= bootloader_state->FailStartsCounter - 1;
		new_bootloader_state.UsingGoldenImage 			= bootloader_state->UsingGoldenImage;
		new_bootloader_state.Checksum 					= new_bootloader_state.Flashed + new_bootloader_state.FailStartsCounter + _bootloaderHeader->appVer[0] + _bootloaderHeader->appVer[1];

		Programming_Infomem_Write(0, 1, &new_bootloader_state, BLD_OFFLINE_STATE_SIZE);


		LOG("Reported successful run to the bootloader.\r\n");
	}

}


void Programming_PrintApplicationHeader() {
	////////////////////////////////////////// print welcome screen via log output
	printf("\r\n\r\n");
	
	appHeader_t* _applicationHeader;
	_applicationHeader = (appHeader_t*) 0x4600;

	LOG("Booting app: '%s' V: %i.%i for %s with %s",
		_applicationHeader->appName,
		_applicationHeader->appVer[0],
		_applicationHeader->appVer[1],
		_applicationHeader->boardName,
		_applicationHeader->extBoardName
		);
	/*
	LOG("Entry Point at %p",
		_applicationHeader->resetAdr
		);
	*/
	
	char rtc_buff[RTC_TIME_STRING_LENGTH];
	
	rawtime_t rawtime;
	time_t time;
	rawtime.miliseconds = 0;
	rawtime.seconds = _applicationHeader->compileTime;			
	RTC_ConvertToTime(&rawtime, &time);
	RTC_ToString(&time, rtc_buff);
	
	LOG("Compiled: %s UTC",
		rtc_buff
		);
}
