/**
 * Copyright (c) 2008, Freie Universitt Berlin, Tomasz Naumowicz
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:
 * 
 * Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
 * Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.
 * Neither the name of the Freie Universitt Berlin nor the names of its contributors may be used to endorse or promote products derived from this software without specific prior written permission.
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#ifndef PROGRAMMING_H_
#define PROGRAMMING_H_



#define INFOMEM_START		0x1000
#define INFOMEM_BLOCK_SIZE	128

//#define INFOMEM_CONFIG		0		///< start of config data
/// start of panic message
//#define INFOMEM_PANIC		(INFOMEM_CONFIG + sizeof(config_t) + 1)
//#define INFOMEM_PANIC_SIZE	0x0040u		///< size of panic message field
#define INFOMEM_USER_SPACE		128		///< start of user defined data

#define BLD_OFFLINE_STATE_SIZE					4

__attribute__((packed)) typedef struct {
  uint8_t Flashed;
  uint8_t FailStartsCounter;
  uint8_t UsingGoldenImage;
  uint8_t Checksum;
} bootloader_offline_state_t;


/**
 * @brief	Read bytes from infomemory
 * @param[out]		buffer		Pointer to buffer for read data
 * @param[in]		offset		Offset in infomemory (0-254)
 * @param[in]		size		Number of bytes to read
 */
void Programming_Infomem_Read(uint16_t offset, void* buffer, uint8_t size);

/**
 * @brief	Write bytes to infomemory
 * @param[in]		offset		Offset in infomemory (0-254)
 * @param[in]		count		Number of items following
 *								each item is a pair pointer, length
 *
 * Example: Infomem_write( 0, 2, &a,3, &b,1 );
 *
 * \note: The MSP430 has two consecutive blocks of infomemory. Each is 128 Bytes large.
 *        The offset is the relative address starting at the beginning of the first block.
 *        You can write an arbitrary number of bytes at any offse, but this 
 *        function cannot write across the two blocks of infomemory.
 */
bool Programming_Infomem_Write(uint16_t offset, uint8_t count, ...);

#ifndef __COMPILE_TIME
	#define __COMPILE_TIME	0
	#warning "Define __COMPILE_TIME through compiler to current UTC time!"
#endif

// added by Hutta 
#define __APP_NAME	"UNKNOWN "
#define __APPVER_MINOR	0
#define __APPVER_MAJOR 0
// added by Hutta */

#ifndef __APP_NAME
	#define __APP_NAME	"UNKNOWN "
	#error "Define __APP_NAME in the makefile"
#endif

#ifndef __APPVER_MINOR
	#define __APPVER_MINOR	0
	#error "Define __APPVER_MINOR in the makefile"
#endif

#ifndef __APPVER_MAJOR
	#define __APPVER_MINOR	0
	#error "Define __APPVER_MINOR in the makefile"
#endif

#ifndef __PLATFORM_BOARD
	#define __PLATFORM_BOARD		"MSB430H "
#endif

#ifndef __PLATFORM_EXTBOARD
	#define __PLATFORM_EXTBOARD		"MSB430MS"
#endif

#define APPHEADER(name, type, hwBoard, hwExtBoard, appName, appVerH, appVerL, fpReset)	\
	const appHeader_t name = {															\
		type ,																			\
		hwBoard , hwExtBoard ,															\
		appName ,{ appVerH , appVerL },													\
		(unsigned int) fpReset,															\
		__COMPILE_TIME ,																\
		( type + appVerH + appVerL) % 0xFF					\
	}

typedef struct {					///																SIZE in bytes
	uint8_t		type;				/// type of the header											1
	uint8_t		boardName[9];		/// name of supported hardware platform (8 chars + \0)			9
	uint8_t		extBoardName[9];	/// name of supported extension board (8 chars + \0)			9
	uint8_t		appName[9];			/// name of application (8 chars + \0)							9
	uint8_t		appVer[2];			/// major and minor version of application						2
	uint16_t	resetAdr;			/// address of application reset function						2
	uint32_t	compileTime;		/// compile time in seconds since 1970-01-01 00:00:00 UTC		4
	uint8_t		checksum;			/// checksum of the header 										1
} appHeader_t; 						///																37 bytes

/**
 * @name Public interface
 * @{
 */

/**
 * Software header. 
 * Theses variables are defined by the linker script, they are needed in the software header
 *
 * @since         1.0
 */

extern const appHeader_t appHeader;

#define APP(type, appName, appVerH, appVerL)												\
	extern void __attribute__((__noreturn__))   _reset_vector__(void);   																\
	__attribute__((section(".app_header")))													\
	APPHEADER( appHeader, type, __PLATFORM_BOARD, __PLATFORM_EXTBOARD, appName, appVerH, appVerL, &_reset_vector__  )


#define APPH_OS						0x80
#define APPH_LOADER					0x40
#define APPH_FLASHABLE				0x20
#define APPH_VERSION				0x00

#define APPH_FLAG_MASK				0xF8
#define APPH_VERSION_MASK			0x07

#define	APPH_STATIC_APP 			APPH_VERSION
#define	APPH_FLASHABLE_APP			(APPH_VERSION | APPH_FLASHABLE)

void Programming_PrintApplicationHeader();

void Programming_ProcessBootloaderInfo();

/* 
bool Prog_TestMetadata( appHeader_t* md );
bool Prog_EnableBootloader();


///@}
#endif
*/

#endif /*PROGRAMMING_H_*/
