/**
 * Copyright (c) 2008, Freie Universitt Berlin, Tomasz Naumowicz
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:
 * 
 * Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
 * Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.
 * Neither the name of the Freie Universitt Berlin nor the names of its contributors may be used to endorse or promote products derived from this software without specific prior written permission.
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include "System.h"

/**
 * Real Time Clock implementation
 * RTC.c, RTC.h
 * Tomasz Naumowicz
 *
 * Status: 
 * feature complete.
 * 
 * Last update: 15.04.2008
 */

/**
 * Real Time Clock implementation
 * 
 * the resolution of the RTC can be configured using two defines:
 * 	RTC_RESOLUTION_125_MILISECONDS
 *  RTC_RESOLUTION_8_MILISECONDS
 * 		Uses slow floating point operations. Every sampling of the time takes 500us (41 times slower than the 125ms resolution)
 * 
 *  set one or non of this defines to 1 in your AppConfig.h
 *  when they are not defined at all, default configuration will be used: not miliseconds support
 * 
 * ----------------------------------------------------------------------------------
 * The RTC is supported by the Timers.c, the HW TimerA is used to drive the RTC.
 * This is the energy efficient version: the RTC is updated on every overflow of the TimerA and on request of the time
 */ 


/**
 * _SystemTime
 * NEVER DIRECTLY ACCESS THIS VARIABLE. USE THE API LISTED IN TIMERS.H
 * This variable is used internally by the RTC and consists of a 32 bit second counter and a
 * 16 bit millisecond counter.
 *
 * _SystemTime gets incremented by the Timer_A1 Interrupt in Timers.c every 8 seconds.
 * IT DOESN'T REPRESENT THE _CURRENT_ REAL TIME
 * ==> additional operations are required to get the current time ==> USE THE API
 * For that reason this var isn't declared public in the header file.
 */
volatile rawtime_t _SystemTime;

// Days per Month
#define JAN   (31)
#define FEB   (28)
#define MAR   (31)
#define APR   (30)
#define MAY   (31)
#define JUN   (30)
#define JUL   (31)
#define AUG   (31)
#define SEP   (30)
#define OCT   (31)
#define NOV   (30)
#define DEC   (31)

static uint8_t days_per_month[12] = {JAN, FEB, MAR, APR, MAY, JUN, JUL, AUG, SEP, OCT, NOV, DEC};

/**
 * Returns the current system time as rawtime_t
 * You have to pass a pointer to a rawtime_t struct
 * this struct will be filled with current time.
 * 
 * version without ms takes 10us to process and return
 * version with ms takes 12us to process and return
 * version with most accurate ms count (using floating point operation) 500us == 41 times slower than with 125ms resolution
 */
rawtime_t* RTC_GetRawTime(rawtime_t* time) {
	uint16_t _TAR_Copy;
	
	// critial section starts here
		//		the _SystemTime and current TAR have to be added
		//		in the meantime the TimerA could overflow and the SAR adds 8 seconds
		//		to the _SystemTime and the TAR resets --> the addition will give false results if this happenes
		//		==> Interrupts HAVE to be disabled during this operation.
		
		_DINT();											// disable interrupts
		
		time->seconds = _SystemTime.seconds;				// copy the _SystemTime
		time->miliseconds = _SystemTime.miliseconds;
		
		do {												// copy the current value of TAR
			_TAR_Copy = TAR;
		}
		while (_TAR_Copy != TAR);							// majority-voting: at least two equal readings
		
		_EINT();											// enable interrupts
	// end of the critical section
	
	// general idea:
	// time->miliseconds += (_TAR_Copy mod TICKS_PER_SECOND) % MILISECONDS_PER_TICK
	// we would use 8 ticks per milisecond and not 8.192 per milisecond.
	// we would not be accurate close to end of a second because we would count quicker
	// 		in the 976ms we would assume it is already 1000 ms and...
	// 		in the (976,1000)ms we have results > 1000 ...
	// soulution: we reduce the accuracy to 1/8th of a second == 125 ms and will round to the lowest
	// 125 ms == 1024 ticks
	// 1024 dec == 100 0000 0000 bin == 0x0400 hex
	//
	// (_TAR_Copy & 0x1FFF) == ticks count mod 8192
	// ((_TAR_Copy & 0x1FFF) >> 10) == (ticks count mod 8192) / 1024
	// this tells me how many 1024 ticks we had in this second
	// ((_TAR_Copy & 0x1FFF) >> 10) * 125 == ((ticks count mod 8192) / 1024) * 125
	// this tells me the number of ms with the resoultion of 125 ms

	// version with 125 ms accuracy
	#if RTC_RESOLUTION_125_MILISECONDS
		time->miliseconds += ((_TAR_Copy & 0x1FFF) >> 10) * 125;
	#endif
		
	// version with highest accuracy (8.192 ms), slow
	#if RTC_RESOLUTION_8_MILISECONDS
		time->miliseconds += ((_TAR_Copy & 0x1FFF) / TICKS_PER_MILISECOND);
	#endif

	// version with highest accuracy and speed but not in real ms anymore
	#if RTC_RESOLUTION_IN_TICKS
		time->miliseconds += (_TAR_Copy & 0x1FFF);
	#endif
		
	// check for overflows only when miliseconds are used.
#if RTC_RESOLUTION_125_MILISECONDS || RTC_RESOLUTION_8_MILISECONDS
	while (time->miliseconds >= 1000) {
		time->miliseconds -= 1000;
		time->seconds++;
	}
#endif

#if RTC_RESOLUTION_IN_TICKS
	while (time->miliseconds >= 0xFFFF) {
		time->miliseconds -= 0xFFFF;
		time->seconds++;
	}
#endif

	time->seconds += _TAR_Copy >> 13;					// quicker
	
	return time;
}

/**
 * 8us
 */
void RTC_SetRawTime(rawtime_t* time) {
	uint16_t _TAR_Copy;	
	// critial section starts here
		//		the _SystemTime has to be modified
		//		in the meantime the TimerA could overflow and the SAR adds 8 seconds
		//		to the _SystemTime
		//		==> Interrupts HAVE to be disabled during this operation.
		//
		// the _SystemTime doesn't store the CURRENT time, the current time is a result of the addition of the TAR
		// the TAR stores up to 8 seconds:
		// use the current value of TAR and store the difference as _SystemTime
		// this is safe for _SystemTime.seconds but not safe for _SystemTime.miliseconds:
		//   the number of miliseconds is used for computations during GetRawTime
		//   miliseconds are converted to seconds if an overflow of miliseconds is detected
		//   this doesn't happend for _SystemTime.seconds
		//   since we work with unsigned numbers:
		//     small number of miliseconds - TAR(miliseconds) < 0 ==> overflow with a value around 65000
		//     this could be solved by logics but is not required.
		//     number of miliseconds will be ignored
		
		_DINT();											// disable interrupts
		
		do {												// copy the current value of TAR
			_TAR_Copy = TAR;
		}
		while (_TAR_Copy != TAR);							// majority-voting: at least two equal readings
		
		_SystemTime.seconds = time->seconds - (_TAR_Copy >> 13);	// copies the number of seconds stored in TAR

		_SystemTime.miliseconds = 0;

		_EINT();											// enable interrupts
	// end of the critical section

		//LOG_SYSTEM("RTC %lu", time->seconds);
}


long dysize(y)
{
	if((y%4) == 0)
		return(366);
	return(365);
}

/**
 * Converts rawtime (seconds elapsed since 00:00 hours, Jan 1, 1970 UTC) to time_ structure containing a calendar date and time broken down into its components
 * 500-625 us
 */ 
void RTC_ConvertToTime(rawtime_t* from, time_t* to) {
		int temp1, temp2;
		long secsInLastDay, day;
		
		long totalSeconds = (from->seconds);

		secsInLastDay = totalSeconds % 86400;
		day = totalSeconds / 86400;
		/*
		if (secsInLastDay<0) {
			secsInLastDay += 86400;
			day -= 1;
		}
		*/
		to->second = secsInLastDay % 60;
		temp2 = secsInLastDay / 60;
		to->minute = temp2 % 60;
		temp2 = temp2 / 60;
		to->hour= temp2;

		if (day>=0) {
			for(temp2=70; day >= dysize(temp2); temp2++)
				day -= dysize(temp2);
		}
		else {
			for (temp2=70; day<0; temp2--)
				day += dysize(temp2-1);
		}

		to->year = temp2 - 70;
		temp1 = day;
		
		if (dysize(temp2)==366)
			days_per_month[1] = 29;
		
		for(temp2=0; temp1 >= days_per_month[temp2]; temp2++)
			temp1 -= days_per_month[temp2];
		
		days_per_month[1] = 28;
		
		to->day = temp1+1;
		to->month = temp2 + 1;

		to->ms = from->miliseconds;
}


void RTC_ConverToRawTime(time_t* from, rawtime_t* to) {
	  uint32_t seconds;
	  uint16_t feb;
	  uint16_t i;
	  uint32_t mult;
	  
	  //84400 == 24 * 60 * 60 == seconds a day
	  seconds = (86400*365) * (from->year); 	//add years to seconds

	  seconds += (86400*((from->year+1)/4));    //add leap year days

	  feb = FEB;                           		//choose feb 28 or 29
	  if(((from->year+2) % 4) == 0) feb++;
	  
	  switch(from->month) {                  	// add month and day
	    case  1:
	    	mult = 0;
	    	break;
	    case  2:
	    	mult = JAN;
	    	break;
	    default:
	    	mult = JAN + feb;
	    	for( i = 2; i < (from->month - 1); i++ )
	    		mult += days_per_month[i];
	  }
	  mult += from->day - 1;
	  seconds += 86400 * mult;
	  // add hour minute second
	  seconds += ((((uint32_t)from->hour * 60) + from->minute) * 60) + from->second;
	  // write dest
	  to->seconds = seconds;
	  to->miliseconds = from->ms;
}


void RTC_SetTime(time_t* time) {
	rawtime_t rawtime;
	RTC_ConverToRawTime(time, &rawtime);
	RTC_SetRawTime(&rawtime);
}

/**
 * Input in the format "hh-mm-ss dd-mm-yyyy" or in the format "hh-mm-ss.mss dd-mm-yyyy"
 */
void RTC_FromString(const char* input, time_t* time) {
	char* pointer = (char*) input;
	
	time->hour = strtoul(pointer, &pointer, 10);
	pointer++;
	time->minute = strtoul(pointer, &pointer, 10);
	pointer++;
	time->second = strtoul(pointer, &pointer, 10);
	
	if(*pointer == '.') {							// detected format with miliseconds "hh-mm-ss.mss dd-mm-yyyy"
		pointer++;
		time->ms = strtoul(pointer, &pointer, 10);
	}
	pointer++;
	
	time->day= strtoul(pointer, &pointer, 10);
	pointer++;
	time->month = strtoul(pointer, &pointer, 10);
	pointer++;
	
	uint16_t year = strtoul(pointer, NULL, 10);
	//we store only the years after 1970 and the variable is a byte
	if (year < 1970)
		year = 1970;
	
	year -= 1970;
	
	if (year > 255)						// 0xFF ist the maximum
		year = 255;

	time->year = year;
	
	time->ms = 0;
}

time_t* RTC_GetTime(time_t* time) {
	rawtime_t rawtime;

	RTC_GetRawTime(&rawtime);
	RTC_ConvertToTime(&rawtime, time);

	return time;
}

/**
 * Converts the time_t to a string.
 * You have to provice a buffer for the resulting string, the size is defined in RTC_TIME_STRING_LENGTH
 * This function requires 3.5ms
 */
void RTC_ToString(time_t* time, char* buffer) {
#if RTC_RESOLUTION_125_MILISECONDS || RTC_RESOLUTION_8_MILISECONDS
	snprintf(
			buffer, RTC_TIME_STRING_LENGTH,
			"%.2u:%.2u:%.2u.%.3u %.2u-%.2u-%.4u",
			time->hour,
			time->minute,
			time->second,
			time->ms,
			time->day,
			time->month,
			time->year + 1970
			);
#else
#if  RTC_RESOLUTION_IN_TICKS
	snprintf(
			buffer, RTC_TIME_STRING_LENGTH,
			"%.2u:%.2u:%.2u.%.4u %.2u-%.2u-%.4u",
			time->hour,
			time->minute,
			time->second,
			time->ms,
			time->day,
			time->month,
			time->year + 1970
			);
#else
	snprintf(
			buffer, RTC_TIME_STRING_LENGTH,
			"%.2u:%.2u:%.2u %.2u-%.2u-%.4u",
			time->hour,
			time->minute,
			time->second,
			time->day,
			time->month,
			time->year + 1970
			);
#endif
#endif
}

/**
 * This function tells whether the node has a valid time
 * everything pre 2008 will be interpeted as a non vaild time.
 */
bool RTC_ValidTime(rawtime_t* time) {
	//00:00:00 01.01.2008 = 1199145600 seconds
	return (time->seconds > 1199145600);
}

// RTC starts with 946684800 seconds == 01.01.2000 00:00:00
void RTC_Init() {
	_SystemTime.seconds = RTC_START_OF_EPOCH;
	_SystemTime.miliseconds = 0;
}


COMMAND(time, COMMAND_SERIAL | COMMAND_REMOTE, cmd_args)
{
	time_t time;

	// check whether the time change was requested
	if (cmd_args->arguments_count > 0) {
		RTC_FromString(cmd_args->arguments, &time);
		RTC_SetTime(&time);
	} else {
		RTC_GetTime(&time);
	}
	
	char buffer[RTC_TIME_STRING_LENGTH];
	RTC_ToString(&time, buffer);
	snprintf(cmd_args->response, cmd_args->response_length, "%s", buffer);
}

COMMAND(uptime, COMMAND_SERIAL | COMMAND_REMOTE, cmd_args) {
	rawtime_t rt; time_t bt;
	char buffer[RTC_TIME_STRING_LENGTH];
	
	rt.miliseconds = 0;
	rt.seconds = (*TICKS() >> 13);  //10000000000000
	
	RTC_ConvertToTime(&rt, &bt);
	RTC_ToString(&bt, buffer);			// the string contains also the date, which is not required
	buffer[8] = '\0';					// get ride of the date
	
	snprintf(
			cmd_args->response,
			cmd_args->response_length,
			"%.2ud %s",
			bt.day - 1,
			buffer
		);
}

