/**
 * Copyright (c) 2008, Freie Universitt Berlin, Tomasz Naumowicz
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:
 * 
 * Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
 * Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.
 * Neither the name of the Freie Universitt Berlin nor the names of its contributors may be used to endorse or promote products derived from this software without specific prior written permission.
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#ifndef RTC_H_
#define RTC_H_

#ifndef RTC_RESOLUTION_125_MILISECONDS
	#define RTC_RESOLUTION_125_MILISECONDS	0
#endif

#ifndef RTC_RESOLUTION_8_MILISECONDS
	#define RTC_RESOLUTION_8_MILISECONDS	0
#endif

#ifndef RTC_RESOLUTION_IN_TICKS
	#define RTC_RESOLUTION_IN_TICKS			0
#endif

#if RTC_RESOLUTION_8_MILISECONDS
	#warning "Uses slow floating point operations. Every sampling of the time takes 500us (41 times slower than the 125ms resolution)"
#endif

#if (RTC_RESOLUTION_125_MILISECONDS && RTC_RESOLUTION_8_MILISECONDS)   
	#error "Selected configuration is not valid!"
#endif

#if (RTC_RESOLUTION_125_MILISECONDS && RTC_RESOLUTION_IN_TICKS)   
	#error "Selected configuration is not valid!"
#endif

#if (RTC_RESOLUTION_8_MILISECONDS && RTC_RESOLUTION_IN_TICKS)   
	#error "Selected configuration is not valid!"
#endif

#if RTC_RESOLUTION_125_MILISECONDS || RTC_RESOLUTION_8_MILISECONDS
	#define RTC_TIME_STRING_LENGTH				24	//hh-mm-ss.mss dd-mm-yyyy\n
#else
#if RTC_RESOLUTION_IN_TICKS
	#define RTC_TIME_STRING_LENGTH				25	//hh-mm-ss.tick dd-mm-yyyy\n
#else
	#define RTC_TIME_STRING_LENGTH				20	//hh-mm-ss dd-mm-yyyy\n
#endif
#endif

/**
 * Time Type (uint32 secs, uint16 ms)
 * It is almost universally expected to be an integral value representing
 * the number of seconds and miliseconds elapsed since 00:00 hours, Jan 1, 1970 UTC.
 * miliseconds_as_ticks is miliseconds expressed as ticks: 8192 ticks == second, 1 tick == 8.192 ms
 */
typedef struct {
  uint32_t seconds;
  uint16_t miliseconds;
} rawtime_t;

/**
 * Time structure
 * Structure containing a calendar date and time broken down into its components.
 *
 * tm_ms		0-999 
 * tm_sec		0-59
 * tm_min		0-59 
 * tm_hour		0-23
 * tm_mday		1-31
 * tm_month		1-12
 * tm_year		years since 1970 
 */
typedef struct {
  uint8_t year;
  uint8_t month;
  uint8_t day;
  uint8_t hour;
  uint8_t minute;
  uint8_t second;
  uint16_t ms;
} time_t;


void RTC_Init();

void RTC_SetTime(time_t* time);

void RTC_SetRawTime(rawtime_t* time);

rawtime_t* RTC_GetRawTime(rawtime_t* time);

time_t* RTC_GetTime(time_t* time);

/**
 * This function tells whether the node has a valid time
 * everything pre 2008 will be interpeted as a non vaild time.
 */
bool RTC_ValidTime(rawtime_t* time);

/**
 * Converts rawtime_t (seconds+miliseconds elapsed since 00:00 hours, Jan 1, 1970 UTC)
 * to time_ structure containing a calendar date and time broken down into its components.
 */ 
void RTC_ConvertToTime(rawtime_t* from, time_t* to);

void RTC_ConverToRawTime(time_t* from, rawtime_t* to);

void RTC_ToString(time_t* time, char* buffer);

void RTC_FromString(const char* input, time_t* time);

#define RTC_START_OF_EPOCH 			946684800  // RTC starts with 946684800 seconds == 01.01.2000 00:00:00

#endif /*RTC_H_*/
