/*
Copyright 2006, Freie Universitaet Berlin. All rights reserved.

These sources were developed at the Freie Universitt Berlin, Computer
Systems and Telematics group.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are
met:

- Redistributions of source code must retain the above copyright
notice, this list of conditions and the following disclaimer.

- Redistributions in binary form must reproduce the above copyright
notice, this list of conditions and the following disclaimer in the
documentation and/or other materials provided with the distribution.
 
- Neither the name of Freie Universitaet Berlin (FUB) nor the names of its
contributors may be used to endorse or promote products derived from
this software without specific prior written permission.

This software is provided by FUB and the contributors on an "as is"
basis, without any representations or warranties of any kind, express
or implied including, but not limited to, representations or
warranties of non-infringement, merchantability or fitness for a
particular purpose. In no event shall FUB or contributors be liable
for any direct, indirect, incidental, special, exemplary, or
consequential damages (including, but not limited to, procurement of
substitute goods or services; loss of use, data, or profits; or
business interruption) however caused and on any theory of liability,
whether in contract, strict liability, or tort (including negligence
or otherwise) arising in any way out of the use of this software, even
if advised of the possibility of such damage.

This implementation was developed by the CST group at the FUB.

For documentation and questions please use the web site
http://scatterweb.mi.fu-berlin.de and the mailinglist
scatterweb@lists.spline.inf.fu-berlin.de (subscription via the Website).
Berlin, 2008
*/

#include "System.h"

typedef struct {
  uint16_t x;
  uint16_t y;
  uint16_t z;
} TD_t;

/**
 *    gs_1   gs_2   type    unit
 *      0     0     1.5      g
 *      1     0     2.0      g
 *      0     1     4.0      g
 *      1     1     6.0      g
 */
#define SET_3D_1_5G do { P5OUT &= ~0x30; } while(0)
#define SET_3D_2_0G do { P5OUT &= ~0x30; P5OUT |= 0x10; } while(0)
#define SET_3D_4_0G do { P5OUT &= ~0x30; P5OUT |= 0x20; } while(0)
#define SET_3D_6_0G do { P5OUT &= ~0x30; P5OUT |= 0x30; } while(0)

#define SHT11_CLK_OFF  do{ P3OUT &= ~0x20; } while(0)
#define SHT11_CLK_ON   do{ P3OUT |=  0x20; } while(0)

#define SHT11_DATA_OFF do{ P3OUT &= ~0x10; } while(0)
#define SHT11_DATA_ON  do{ P3OUT |=  0x10; } while(0)
#define SHT11_DATA     (P3IN & 0x10)

#define SHT11_OUTPUT   P3DIR |= 0x10;
#define SHT11_INPUT    P3DIR &= ~0x10;

#define noACK 0
#define ACK   1
                            // adr  command  r/w
#define MEASURE_TEMP 0x03   // 000   0001    1
#define MEASURE_HUMI 0x05   // 000   0010    1

enum { TEMP, HUMI };

bool _Data_wait(uint32_t w) {
  volatile uint32_t i; 
  for(i=0;i<w;i++) nop(); 
  return true;
}

// If the A/D-Converter is enabled via _System_enableADC, the values of the analog
// inputs are continously converted
void _Data_initADC() {
  //Setup ADC12, ref., sampling time, REFON for DAC!!
  ADC12CTL0 = SHT0_15+SHT1_15+MSC;  
  ADC12CTL1 = SHP + CONSEQ_3 + CSTARTADD_0;   // Use sampling timer, repeat-sequence-of-channels
  ADC12MCTL0 = (INCH_0 + SREF_0);             // P60: AIN0
  ADC12MCTL1 = (INCH_0 + SREF_0);             // P60: AIN0
  ADC12MCTL2 = (INCH_0 + SREF_0);             // P60: AIN0
  ADC12MCTL3 = (INCH_1 + SREF_0);             // P61: AIN1
  ADC12MCTL4 = (INCH_1 + SREF_0);             // P61: AIN1
  ADC12MCTL5 = (INCH_1 + SREF_0);             // P61: AIN1
  ADC12MCTL6 = (INCH_2 + SREF_0);             // P62: AIN2
  ADC12MCTL7 = (INCH_2 + SREF_0);             // P62: AIN2
  ADC12MCTL8 = (INCH_2 + SREF_0);             // P62: AIN2
  ADC12MCTL9 = (INCH_2 + SREF_0);             // P62: AIN2
  ADC12MCTL10 = (INCH_3 + SREF_0);            // P63: AIN2
  ADC12MCTL11 = (INCH_4 + SREF_0);            // P64: AIN3
  ADC12MCTL12 = (INCH_5 + SREF_0);            // P65: AIN4
  ADC12MCTL13 = (INCH_10 + SREF_0) + EOS;     // Temp diode
}

// Enables and starts ADC.
// Enables the A/D-Converter, uses ~1 mA.
bool _Data_enableADC() {
  _Data_initADC();
  ADC12CTL0 |= ADC12ON;
  _Data_wait(600);      // Delay for reference start-up
  ADC12CTL0 |= ENC;     // enable conversion
  ADC12CTL0 |= ADC12SC; // Start conversion
  return true;
}

// Stops and disables ADC.
// Disables the A/D-Converter.
bool _Data_disableADC() {
  ADC12CTL0 &= ~ENC;       // Stop conversion
  _Data_wait(600);         // Wait, TODO exchange with test for ADC12BUSY &test
  ADC12CTL0 &= ~(ADC12ON); // turn off ADC12
  return true;
}

bool _Data_set3DSensitivity(uint16_t sen) {
  switch(sen) {
    case 2: SET_3D_2_0G; break;
    case 4: SET_3D_4_0G; break;
    case 6: SET_3D_6_0G; break;
    default: SET_3D_1_5G; break;
  }
  return true;
}

TD_t* _Data_get3D(TD_t* value) { 
  value->x = ADC12MEM0;
  value->y = ADC12MEM3;
  value->z = ADC12MEM6;
  return value;
}

int8_t _Data_writeByte(uint8_t value) { 
  uint8_t i, error = 0;  
  for(i=0; i<8; i++) {
    if(value & BIT7) SHT11_DATA_ON;
    else SHT11_DATA_OFF;
    SHT11_CLK_ON;         // clk for SENSI-BUS
    _NOP();
	_NOP();
	_NOP();               // pulswith approx. 5 us
    SHT11_CLK_OFF;
    value = value << 1;
  }
  SHT11_CLK_ON; 
  _NOP();
  _NOP();
  _NOP();                 // pulswith approx. 5 us
  SHT11_INPUT;            // SDA is input
  _NOP();                 // clk #9 for ack 
  SHT11_CLK_OFF;
  error = SHT11_DATA;     // check ack (DATA will be pulled down by SHT11)
  return error;           // error=1 in case of no acknowledge
}

int8_t _Data_readByte(uint8_t ack) {
  SHT11_INPUT;               // SDA is input
  uint8_t i, value = 0;
  for(i=0; i<8; i++) {
	value = value << 1;      // shift bit 
    SHT11_CLK_ON;            // clk for SENSI-BUS
    if (SHT11_DATA) value++; // read bit 
    SHT11_CLK_OFF;
  }
  SHT11_OUTPUT;              // SDA is output
  _NOP();
  SHT11_DATA_OFF;
  SHT11_CLK_ON;              // clk #9 for ack
  _NOP();
  _NOP();
  _NOP();                    // pulswith approx. 5 us
  SHT11_CLK_OFF;						    
  return value;
}

void _Data_transstart() {
   SHT11_OUTPUT;
   SHT11_DATA_ON;
   SHT11_CLK_OFF; //Initial state
   _NOP();
   SHT11_CLK_ON;
   _NOP();
   SHT11_DATA_OFF;
   _NOP();
   SHT11_CLK_OFF;  
   _NOP();
   _NOP();
   _NOP();
   SHT11_CLK_ON;
   _NOP();
   SHT11_DATA_ON;		   
   _NOP();
   SHT11_CLK_OFF;		   
}

void _Data_connectionreset() {  
  uint8_t i; 
  SHT11_DATA_ON;
  SHT11_CLK_OFF;                // Initial state
  for(i=0; i<9; i++) {          // 9 SCK cycles
    SHT11_CLK_ON;
    SHT11_CLK_OFF;
  }
}

 							   
uint8_t _Data_measure(uint8_t* p_value, uint8_t* p_checksum, uint8_t mode) {
  unsigned error = 0;
  _Data_transstart(); // transmission start
  switch(mode) {       // send command to sensor
    case TEMP	: error += _Data_writeByte(MEASURE_TEMP); break;
    case HUMI	: error += _Data_writeByte(MEASURE_HUMI); break;
    default     : break;	 
  }
  while(SHT11_DATA);                    // wait until sensor has finished the measurement
  if(SHT11_DATA) error += 1;            // or timeout (~2 sec.) is reached
  *(p_value+1) = _Data_readByte(ACK);  // read the first byte (MSB)
  *(p_value) = _Data_readByte(ACK);    // read the second byte (LSB)
  *p_checksum = _Data_readByte(noACK); // read checksum
  return error;
}


uint16_t _Data_getTemperature() { 
  uint8_t error, checksum;
  uint16_t temp;
  _Data_connectionreset();
  error = 0;
  error += _Data_measure((uint8_t*) &temp, &checksum, TEMP); // measure temperature
  if(error != 0) {
	  _Data_connectionreset(); // in case of an error: connection reset
	  return 0;
  }
  else return temp;
}

uint16_t _Data_getHumidity() { 
  uint8_t error, checksum;
  uint16_t hum;
  _Data_connectionreset();
  error = 0;
  error += _Data_measure((uint8_t*) &hum, &checksum, HUMI); // measure temperature
  if(error != 0) {
	  _Data_connectionreset(); // in case of an error: connection reset
	  return 0;
  }
  else return hum;
}

bool Data_init() {
  Data_boardID = 1;
  Data_channels = 3;
  SET_3D_1_5G;
  return true;
}

bool Data_enable(uint8_t channel, uint8_t parameter_length, uint8_t* parameter) {
  if(channel >= Data_channels) return false;
  if(channel == 0) {
    if(parameter_length != 2) return false;
    uint16_t temp = parameter[0];
	temp = temp << 8;
    temp = temp + parameter[1];
    _Data_set3DSensitivity(temp);
	_Data_enableADC();
  }
  return true;
}

bool Data_disable(uint8_t channel) {
  if(channel >= Data_channels) return false;
  if(channel == 0) _Data_disableADC();
  return true;
}

sdata_t Data_read(uint8_t channel) {
  Data_sensorValue.board = Data_boardID;
  Data_sensorValue.channel = 255;
  Data_sensorValue.timeStamp.seconds = 0;
  Data_sensorValue.timeStamp.miliseconds = 0;
  Data_sensorValue.value_length = 0;
  if(channel >= Data_channels) return Data_sensorValue;
  if(channel == 0) {
    Data_sensorValue.channel = 0;
	Data_sensorValue.value_length = 6;
	TD_t temp;
	_Data_get3D((TD_t*) &temp);
	RTC_GetRawTime(&(Data_sensorValue.timeStamp));
	Data_sensorStream[1] = temp.x;
	Data_sensorStream[0] = temp.x >> 8;
	Data_sensorStream[3] = temp.y;
	Data_sensorStream[2] = temp.y >> 8;
	Data_sensorStream[5] = temp.z;
	Data_sensorStream[4] = temp.z >> 8;
	Data_sensorValue.value = Data_sensorStream;
	return Data_sensorValue;
  }
  if(channel == 1) {
    Data_sensorValue.channel = 1;
	Data_sensorValue.value_length = 2;
	uint16_t temp = _Data_getTemperature();
	RTC_GetRawTime(&(Data_sensorValue.timeStamp));
	Data_sensorStream[1] = temp;
	Data_sensorStream[0] = temp >> 8;
	Data_sensorValue.value = Data_sensorStream;
	return Data_sensorValue;
  }
  if(channel == 2) {
    Data_sensorValue.channel = 2;
	Data_sensorValue.value_length = 2;
	uint16_t temp = _Data_getHumidity();
	RTC_GetRawTime(&(Data_sensorValue.timeStamp));
	Data_sensorStream[1] = temp;
	Data_sensorStream[0] = temp >> 8;
	Data_sensorValue.value = Data_sensorStream;
	return Data_sensorValue;
  }
  return Data_sensorValue;
}

uint8_t Data_getBoardID() {
  return Data_boardID;
}

uint8_t Data_getChannels() {
  return Data_channels;
}

COMMAND(rtd, COMMAND_SERIAL | COMMAND_REMOTE, cmd_args) {
  TD_t temp;
  _Data_get3D(&temp);
  snprintf(cmd_args->response, cmd_args->response_length, "x:%i|y:%i|z:%i", temp.x, temp.y, temp.z);
}

COMMAND(tem, COMMAND_SERIAL | COMMAND_REMOTE, cmd_args) {
  uint16_t temp = _Data_getTemperature();
  snprintf(cmd_args->response, cmd_args->response_length, "%i", temp);
}

COMMAND(hum, COMMAND_SERIAL | COMMAND_REMOTE, cmd_args) {
  uint16_t temp = _Data_getHumidity();
  snprintf(cmd_args->response, cmd_args->response_length, "%i", temp);
}
