/*
Copyright 2006, Freie Universitaet Berlin. All rights reserved.

These sources were developed at the Freie Universitt Berlin, Computer
Systems and Telematics group.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are
met:

- Redistributions of source code must retain the above copyright
notice, this list of conditions and the following disclaimer.

- Redistributions in binary form must reproduce the above copyright
notice, this list of conditions and the following disclaimer in the
documentation and/or other materials provided with the distribution.
 
- Neither the name of Freie Universitaet Berlin (FUB) nor the names of its
contributors may be used to endorse or promote products derived from
this software without specific prior written permission.

This software is provided by FUB and the contributors on an "as is"
basis, without any representations or warranties of any kind, express
or implied including, but not limited to, representations or
warranties of non-infringement, merchantability or fitness for a
particular purpose. In no event shall FUB or contributors be liable
for any direct, indirect, incidental, special, exemplary, or
consequential damages (including, but not limited to, procurement of
substitute goods or services; loss of use, data, or profits; or
business interruption) however caused and on any theory of liability,
whether in contract, strict liability, or tort (including negligence
or otherwise) arising in any way out of the use of this software, even
if advised of the possibility of such damage.

This implementation was developed by the CST group at the FUB.

For documentation and questions please use the web site
http://scatterweb.mi.fu-berlin.de and the mailinglist
scatterweb@lists.spline.inf.fu-berlin.de (subscription via the Website).
Berlin, 2008
*/

/** \file ScatterWeb.Data.h
 ** \ingroup System
 ** \brief The ScatterWeb.Data consists mostly of the fuctions that constitute
 ** the sensors architecture. The sensors architecture enables you to build
 ** components that efficiently manage data from multiple data sources. It
 ** provides the tools to initialize, enable, and disable data in this multiple
 ** tier system.
 ** 
 ** @author          Marco Hutta
 ** @version         1.2
 ** @since           1.0  
 **/

#ifndef DATA_H_
#define DATA_H_

/**
 * Holds sensor data record.
 * <ul>
 * <li>channel - channel of sensors
 * <li>timeStamp - timestamp of sensor reading.
 * <li>value_length - length of value.
 * <li>value - raw value of sensor reading.
 * </ul>
 *  
 * @since           1.0
 */
typedef struct {
  uint8_t board;
  uint8_t channel;
  rawtime_t timeStamp;
  uint8_t value_length;
  uint8_t* value;
} sdata_t;

/**
 * Copy of the sensor data record
 *
 * @since         1.0
 */
sdata_t Data_sensorValue;

/**
 * Copy of the sensor event data record
 *
 * @since         1.0
 */
sdata_t Data_sensorEventValue;

/**
 * Copy of the sensor data stream
 *
 * @since         1.0
 */
uint8_t Data_sensorStream[6];

/**
 * Copy of the sensor event data stream
 *
 * @since         1.0
 */
uint8_t Data_sensorEventStream[1];

/**
 * Copy of the board ID
 *
 * @since         1.0
 */
uint8_t Data_boardID;

/**
 * Copy of the channels
 *
 * @since         1.0
 */
uint8_t Data_channels;

/**
 * Initializes the sensors
 *
 * @return                  <code>True</code> always.
 * @since                   1.0
 */
bool Data_init();

/**
 * Returns board ID
 *
 * @return UINT8            ID of the sensor board
 *                          0  - this means that no sensor board is available
 *                          1  - this means that just on-board sensors are available
 *                          >1 - otherwise
 * @since                   1.0
 */
uint8_t Data_getBoardID();

/**
 * Returns channels
 *
 * @return UINT8            number of channels
 *                          0  - this means that no sensor is available
 *                          >0 - otherwise
 * @since                   1.0
 */
uint8_t Data_getChannels();

/**
 * Starts a sensor
 *
 * @param channel           sensor to start
 * @param parameter_length  length of paraemter
 * @param parameter         paraemter to pass
 * @return                  <code>True</code> if successful;
 *                          <code>False</code> otherwise.
 * @since                   1.0
 */
bool Data_enable(uint8_t channel, uint8_t parameter_length, uint8_t* parameter);

/**
 * Stops a sensor
 *
 * @param channel           sensor to stop
 * @return                  <code>True</code> if successful;
 *                          <code>False</code> otherwise.
 * @since                   1.0
 */
bool Data_disable(uint8_t channel);

/**
 * Reads a sensor
 *
 * @param channel           sensor to read
 * @return sdata_t          value of sensor reading.
 *                          error: sdata_t.channel = 255
 * @since                   1.0
 */
sdata_t Data_read(uint8_t channel);

#endif
