/**
 * Copyright (c) 2008, Freie Universitt Berlin, Tomasz Naumowicz
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:
 * 
 * Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
 * Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.
 * Neither the name of the Freie Universitt Berlin nor the names of its contributors may be used to endorse or promote products derived from this software without specific prior written permission.
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include "System.h"

APP(APPH_FLASHABLE, __APP_NAME, __APPVER_MAJOR, __APPVER_MINOR);


/*
 * Used to monitor system state and communicate with the main loop.
 * ISR modify the System_runtime_flag to report events
 * The main loop checks the state of the System_runtime_flag and executed aproperiate actions.
 */
volatile uint8_t System_runtime_flag;

extern bool Net_txHandler();
extern bool Net_rxHandler();

int main(void){
	////////////////////////////////////// boot sequence
	WDTCTL = WDTPW | WDTHOLD;			// stop watchdog timer for the initialization
	
	////////////////////////////////////// setup hardware ports
	MSB430H_InitPorts();				// init the MSB430H board (configure ports)

	////////////////////////////////////// Important: clock configuration
	////////////////////////////////////// 7.372 MHz are supplied via XT2: the external quartz (100uA power consumption)
	////////////////////////////////////// done.
	  
	////////////////////////////////////// setup the ACLK to be supplied by the XT: External 32kHz quartz is attached (32768Hz)
	#ifndef __ACLK_32KHZ__
	#error "Selected ACLK ist not supported by the firmware"
	#endif
	// --> it's preset already after POR/PUC
	////////////////////////////////////// done.
	// ADDED by Hutta
	extern bool DCO_set();
	DCO_set();
	// ADDED by Hutta
	////////////////////////////////////// basic initialization completed
	_EINT();							// make sure that interrupts are enabled after the initialization

	////////////////////////////////////// initialize pheripherials
	USART_Init();						// USART (defaults to UART 115200 bps)

	////////////////////////////////////// initialize extended firmware componentes and start printing output
	Timers_Init();						// Init the software timers (sets up the TimerA and serves the TimerA interrupts
	RTC_Init();							// Init the software real time clock  (requires Timers_Init)
	
	Logging_Init();						// Logging features
	
	Configuration_Init();				// read the configuration settings from the infomem area
	
	Network_Init(Configuration.NodeID);	// start the radio support
	// ADDED by Hutta
	extern bool DCO_start();
	DCO_start();
	// ADDED by Hutta
	Commands_Init();					// init commands support

	////////////////////////////////////// boot sequence complete.
	STATUS_LED_OFF;						// make sure the LED is off after the initialization
	
	////////////////////////////////////// Display application details
	Programming_PrintApplicationHeader();
	
	Configuration_Log();
	
	////////////////////////////////////// initialize SD card and FAT16 support
	
	//retry to init the SD card 10 times
	{
		uint8_t counter;
		for (counter = 0; counter < 10; counter++) {			//TODO: after the cold start the init has to be executed twice: this is not caused by power on time of the sd card, this was tested already - some registers are changed during the first run - this needs more time to be investigated. after the deadline.
			SD_Card_Init();
			if (SD_Card_Init_Result == SD_INIT_SUCCESS)
				break;
		}
	}
	
	switch (SD_Card_Init_Result) {
		case SD_INIT_NOCARD:
			LOG_CODE(1, "no SD");
			break;
		case SD_INIT_WRITEPROTECTED:
			LOG_CODE(2, "SD readonly");
			break;
		case SD_INIT_SUCCESS:
			LOG_CODE(3, "SD OK");
			break;
		default:
			LOG_CODE(4, "SD error %u", SD_Card_Init_Result);
			break;
	}
	
	if (SD_Card_Init_Result == SD_INIT_SUCCESS) {
		enum dfs_init_result ret;
		ret = DFS_Init(LOGLEVEL & LOGLEVEL_VERBOSE);
		
		if (ret == FAT_ERROR_NONE) 
			LOG_CODE(5, "FAT OK");
		else
			LOG_CODE(6, "FAT Error %u", ret);
	}
	else {
		LOG_CODE(7, "FAT disabled");
	}
	
	Programming_ProcessBootloaderInfo();
		
	LOG("App: initializing");
	Application_Init();
	LOG("App: initialized");

	
	WDTCTL = WDTPW | WDTHOLD;            	// Stop watchdog timer (the SMCLK runs in the LPM0 mode)
	
	
	// main loop
	while (1) {
		if (System_runtime_flag & FLAG_SOFTWARE_TIMER) {
			CLEAR(System_runtime_flag, FLAG_SOFTWARE_TIMER);
			
			Timers_ProcessPendingTimerHandler();
		}
		
		/* added by Hutta */
		if (System_runtime_flag & FLAG_RADIO_TX) {
			CLEAR(System_runtime_flag, FLAG_RADIO_TX);
					
			Net_txHandler();
		}
		
		if (System_runtime_flag & FLAG_RADIO_RX) {
					//CLEAR(System_runtime_flag, FLAG_SOFTWARE_TIMER);
					
			Net_rxHandler();
		}
		/* added by Hutta */
		
		if (System_runtime_flag & FLAG_UART_NEWLINE) {
			CLEAR(System_runtime_flag, FLAG_UART_NEWLINE);

			uint8_t* line = USART_ReadLine();				//this function call copies the line from the ringbuffer
			
			if (!Commands_ProcessSerialInput(line)) {
				//Core_ProcessReceivedLine(line);
			}
			
		}
		
		if (System_runtime_flag & FLAGS_USER) {
			Application_ProcessSystemFlags();
		}
	}

	return 0;
}
