/**
 * Copyright (c) 2008, Freie Universitt Berlin, Tomasz Naumowicz
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:
 * 
 * Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
 * Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.
 * Neither the name of the Freie Universitt Berlin nor the names of its contributors may be used to endorse or promote products derived from this software without specific prior written permission.
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

/**
 * USART control module:
 * The universal synchronous/asynchronous receive/transmit (USART)
 * peripheral interface supports two serial modes with one hardware module.
 * 
 * This file contains the code for the UART and the SPI mode.
 */

#include "System.h"

enum usart_mode			USART_CurrentMode;

volatile uint8_t		_usart_locked;

/**
 * UART SUPPORT
 */

/**
 * allows to preview the character received via the UART port
 * returning TRUE stops the further processing of the character by the USART logic
 * 	might by used to intercept and process data that shouldn't be stored in the receiver ringbuffer1
 * returning FALSE allows the rest of the USART logic to execute
 */
fp_previewCharacter_t USART_CharacterPreview;

/**
 * all received characters are stored in the ringbuffer.
 */
RINGBUFFER_NEW(_rx_ringbuffer, UART_RX_BUFFER_SIZE);

#define USART_WAIT_LOCK(mode)	( (USART_CurrentMode != mode ) && (_usart_locked) )

static uint8_t			_usart_speed_br0[2];
static uint8_t			_usart_speed_br1[2];
static uint8_t			_usart_speed_bmn[2];

/**
 * @brief	Baudrate settings for the UART mode
 * 
 * triple {UBR01, UBR11, UMCTL1}
 */
const uint8_t bps_settings[3][3] = {
	{0xFF, 0x02, 0xBF},		// BPS_9600		UBR01=0xFF; UBR11=0x02; UMCTL1=0xBF;	/* uart 7372000Hz 9600bps */
	{0x08, 0x01, 0x00},		// BPS_19200	BR01=0x80; UBR11=0x01; UMCTL1=0x00;		/* uart 7372000Hz 19197bps */
	{0x27, 0x00, 0x10}		// BPS_115200	UBR01=0x40; UBR11=0x00; UMCTL1=0x00;	/* uart 7372000Hz 115187bps */
};//added by hutta

// UTCTL1 = SSEL1; UBR01 = 0x27; UBR11 = 0x00; UMCTL1 = 0x10;

void _USART_Reconfigure(enum usart_mode mode) {
	//this is a critical section, we have to disable all interrupts.
	_DINT();
	
	//wait for all buffered data to be transmitted
	UART_WAIT_TXDONE();
	
	//set the configuration:
	switch (mode) {
		case USART_MODE_UART: 
				////////////////////////////////////// MSB430H Specific: Select UART function
				P5OUT |= 0x01;						// unselect SD CARD
				P3SEL |= 0xC0;						// select rs232

				// The required USART initialization/re-configuration process is:
				// 1) Set SWRST (BIS.B #SWRST,&UxCTL)
				U1CTL = SWRST;
				// 2) Initialize all USART registers with SWRST = 1 (including UxCTL)
				U1CTL |= CHAR;          			// 8-bit character
				U1TCTL |= SSEL1;					// UCLK = SCLK

				//set the speed according to the stored configuration
				UBR01 = _usart_speed_br0[mode];
				UBR11 = _usart_speed_br1[mode];
				UMCTL1 = _usart_speed_bmn[mode];
						
				// 3) Enable USART module via the MEx SFRs (URXEx and/or UTXEx)
				ME2 |= UTXE1 | URXE1;				// Enable USART1 TXD/RXD	

				// 4) Clear SWRST via software (BIC.B #SWRST,&UxCTL)
				U1CTL &= ~SWRST;

				// 5) Enable interrupts (optional) via the IEx SFRs (URXIEx and/or UTXIEx)
				IE2 |= URXIE1; //interrupt enable for RX
				IFG1 &= ~URXIE1;
				
				//IE2 |= UTXIE1; //interrupt enable for TX
				//IFG1 &= ~UTXIE1;
				
			break;
		case USART_MODE_SPI:
				////////////////////////////////////// MSB430H Specific: Select UART function
				P5OUT &= ~0x01;						// select SD CARD
				P3SEL &= ~0xC0;						// unselect rs232
	
				// The required USART initialization/re-configuration process is:
				// 1) Set SWRST (BIS.B #SWRST,&UxCTL)
				U1CTL = SWRST;
				// 2) Initialize all USART registers with SWRST = 1 (including UxCTL)
				U1CTL |= CHAR | SYNC | MM;		// 8-bit SPI Master
				UTCTL1 |= CKPH | SSEL1 | SSEL0 | STC;	// SMCLK, 3-pin mode, clock idle low, data valid on rising edge, UCLK delayed
	
				//set the speed according to the stored configuration
				UBR01 = _usart_speed_br0[mode];
				UBR11 = _usart_speed_br1[mode];
				UMCTL1 = _usart_speed_bmn[mode];
						
				// 3) Enable USART module via the MEx SFRs (URXEx and/or UTXEx)
				ME2 |= USPIE1;							// Enable USART1 SPI
	
				// 4) Clear SWRST via software (BIC.B #SWRST,&UxCTL)
				U1CTL &= ~SWRST;
	
				// 5) Enable/Disable interrupts (optional) via the IEx SFRs (URXIEx and/or UTXIEx)
				IE2 &= ~URXIE1; //disable interrupt for RX
				IFG1 &= ~URXIE1;
				
				//IE2 |= UTXIE1; //interrupt enable for TX
				//IFG1 &= ~UTXIE1;
				
			break;
		default:
			break;
	}
	
	//re-enable interrupts
	_EINT();
}

bool USART_Lock(enum usart_mode mode) {
	if ((USART_CurrentMode != mode ) && (_usart_locked))
		return false;										// USART is locked already in another mode
	
	// lock can be set:
	_usart_locked++;
	
	if (USART_CurrentMode != mode) {
		_USART_Reconfigure(mode);
		USART_CurrentMode = mode;
	}

	return true;
}

bool USART_LockWait(enum usart_mode mode) {
	while (USART_WAIT_LOCK(mode))
		_NOP();
	
	return USART_Lock(mode);
}

bool USART_Unlock(enum usart_mode mode) {
	if (USART_CurrentMode != mode)
		return false;
	
	if (_usart_locked > 0) {
		_usart_locked--;
		
		if (_usart_locked == 0)
			if (USART_CurrentMode != USART_MODE_UART) {
				_USART_Reconfigure(USART_MODE_UART);
				USART_CurrentMode = USART_MODE_UART;
			}
		
		return true;
	}
	
	return false;
}

/**
 * Sets the low level spped registers. If the USART is in the selected mode, the change will be processed immiedately
 * and the interrupts will be enabled. If not, the new speed settings will be cached and used as soon as the new
 * mode will be selected.
 * 
 * It is expected that this function will be used whilie using the SPI mode.
 * Use the USART_SetUartSpeed funcition for the UART mode.
 * 
 */
void USART_SetSpeedRegisters(enum usart_mode mode, uint8_t ubr0, uint8_t ubr1, uint8_t umctl) {
	_usart_speed_br0[mode] = ubr0;
	_usart_speed_br1[mode] = ubr1;
	_usart_speed_bmn[mode] = umctl;

	if (mode == USART_CurrentMode)
		_USART_Reconfigure(mode);
}

/**
 * Sets the speed for the UART mode. If the UART mode is active, the change will be processed immidately
 * and the interrupts will be enabled.
 */
void USART_SetUartSpeed(enum uart_bps uartSpeed){
	USART_SetSpeedRegisters(
			USART_MODE_UART,
			bps_settings[uartSpeed][0],
			bps_settings[uartSpeed][1],
			bps_settings[uartSpeed][2]
	);
}


/**
 * Initis the USART, sets the UART mode as default with the speed 115200.
 * This function call enables interrupts.
 */
void USART_Init() {
	USART_CurrentMode = USART_MODE_UART;
	
	_usart_locked = 0;
	USART_CharacterPreview = NULL;
	
	USART_SetSpeedRegisters(USART_MODE_SPI, 0x02, 0x00, 0x00);
	USART_SetUartSpeed(BPS_115200);
}

interrupt(USART1RX_VECTOR) UsartRX_ISR(void) {
	
	uint8_t input;								// used to hold the received character

	if (U1RCTL_ & RXERR) { 						//the receive error flag was set
		input = U1RXBUF;						//clear the error flag by reading the character and leaving the ISR
		return;									//and return!
	}
	
	// there was no error during the communication:
	
	input = U1RXBUF;							//read the rx buffer of the USART
	IFG1 &= ~URXIE1;							//make sure, the interrupt flag was cleared, so that new characters can be processed
	
	// allow to modify the behaviour of the ISR: the application can preview characters and interrupt its processing
	if (USART_CharacterPreview != NULL) {		//if the preview handler was definded
		if (USART_CharacterPreview(input))		//use the preview handler: the handler returns TRUE if the processing of the characters should be stopped
			return;
	}
	
	// regular processing of entered character: storage in the ring buffer and notifications via system flags after the NEW LINE character
	
	if (input == '\r')							//ingore carriage return	
		return;
	
	if (input == '\n') {						//process NEW LINE character
		ringbuffer_put(&_rx_ringbuffer, '\0');
		System_runtime_flag |= FLAG_UART_NEWLINE;
		return;
	}

	ringbuffer_put(&_rx_ringbuffer, input);		//store all other characters
}

uint8_t _lastReceivedLine[USART_MAX_RECEIVED_LINE_LENGTH]; 

uint8_t* USART_ReadLine() {
	int8_t index = -1;
	int8_t character;
	
	do {
		character = ringbuffer_get(&_rx_ringbuffer);
		if (character == -1)								// if the ringbuffer was empty: break the loop
			break;
		
		index++;
		_lastReceivedLine[index] = character;
	} while (_lastReceivedLine[index] != '\0' && index < USART_MAX_RECEIVED_LINE_LENGTH);
	
	/**
	 * the loop terminated: check if the data is correct
	 * the valid state is when:
	 * 		_lastReceivedLine[index] == '\0'
	 * it might be, that the received line was longer than the maximum allowed length
	 * in that case the loop breaks the execution and the '\0' would be missing.
	 * in that case the ring buffer could contain the rest of the message: the state of the ring buffer
	 * would be corrupted: it would still hold a message, that wouldn't be processed until the next NEW LINE
	 * is received.
	 * In order to make sure, the output is in the valid state, '\0' will be stored 
	 * at the end of the line.
	 */
	
	_lastReceivedLine[index] = '\0';
	
	/**
	 * in order to make sure that the ring buffer is in the correct state:
	 * if the character == -1: the buffer was empty, the buffer is in the correct state.
	 * the character != -1: there is still something in the buffer: set the fimrware flag to NEW LINE,
	 * so that the application is forced to pick up the data again.
	 * it might be, that the last character in the buffer was the '\0'
	 * this would mean, that the state of the ringbuffer is correct.
	 * in order to see, if data in the buffer is available, the length of the ringbuffer will be used:
	 */
	
	if (ringbuffer_len(&_rx_ringbuffer) > 0)
		System_runtime_flag |= FLAG_UART_NEWLINE;
	
	return _lastReceivedLine;
}

/**
 * Implementation of the putchar function from the stdio.h
 * This function is required for printf etc to work
 */
int putchar(int c) {
	if (USART_MODE_IS(USART_MODE_UART)) {
		UART_WAIT_TX();
		UART_TX = c;
	
		return c;
	}
	else {
		return -1;
	}
}

/**
 * Implementation of the _putchar function from the smallib.h
 * This function is required for sl_printf etc to work
int _putchar(int c) {
	if (USART_MODE_IS(USART_MODE_UART)) {
		UART_WAIT_TX();
		UART_TX = c;
	
		return c;
	}
	else {
		return -1;
	}	
}
 */


/**
 * SPI1 support
 */

#define	SPI_idle_SYMBOL		0xFF

uint8_t SPI_rx() {
  SPI_rxTX:
	if((IFG2 & UTXIFG1) == 0) goto SPI_rxTX;
	UART_TX = SPI_idle_SYMBOL;
  SPI_rx:
	if((IFG2 & URXIFG1) == 0) goto SPI_rx;
	return(UART_RX);
}

void SPI_tx(register const uint8_t c) {
  SPI_tx:
	if((IFG2 & UTXIFG1) == 0) goto SPI_tx;
	UART_TX = c;
  SPI_txRX:
	if((IFG2 & URXIFG1) == 0) goto SPI_txRX;
	UART_RX;
}

void SPI_read(void* pDestination, const uint16_t size, const bool incDest) {
	register uint8_t* p = (uint8_t*)pDestination;
	register uint16_t i;
	SPI_READTX:
	  if((IFG2 & UTXIFG1) == 0) goto SPI_READTX;
	  for(i = size; i > 0; i--) {
		UART_TX = SPI_idle_SYMBOL;
		SPI_READRX:
		if((IFG2 & URXIFG1) == 0) goto SPI_READRX;
		*p = UART_RX;
		if(incDest) p++;
	  }
}

void SPI_write(const void* pSource, const uint16_t size, const uint8_t startToken, const bool incSource) {
 	register unsigned char* p = (unsigned char*)pSource;
 	register uint16_t i;
 	SPI_tx(startToken);
	for( i = size; i > 0; i-- ) {
		SPI_tx(*p);
		if( incSource ) p++;
	}
}

void SPI_idle(register const uint16_t clocks) {
	register uint16_t i;
	for( i = 0; i < clocks; i++ ) {
		UART_RESET_RX();
		UART_TX = SPI_idle_SYMBOL;
		UART_WAIT_RX();
	}
}

uint8_t SPI_waitToken(const uint8_t token, const uint16_t timeout) {
	uint16_t	i;
	uint8_t	rx;
	for( i = 0; i < timeout; i++ ) {
		rx = SPI_rx();
		if( rx == token )
			return 1;
	}
	return 0;
}









