/**
 * Copyright (c) 2008, Freie Universitt Berlin, Tomasz Naumowicz
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:
 * 
 * Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
 * Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.
 * Neither the name of the Freie Universitt Berlin nor the names of its contributors may be used to endorse or promote products derived from this software without specific prior written permission.
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#ifndef USART_H_
#define USART_H_

#ifndef UART_RX_BUFFER_SIZE
	#define UART_RX_BUFFER_SIZE					40
#endif

#ifndef USART_MAX_RECEIVED_LINE_LENGTH
	#define USART_MAX_RECEIVED_LINE_LENGTH		40
#endif

/**
 * Available modes of the USART module
 */
enum usart_mode {
	USART_MODE_UART     = 0,
	USART_MODE_SPI      = 1
};

extern enum usart_mode		USART_CurrentMode;

typedef bool (*fp_previewCharacter_t) (uint8_t);

/**
 * allows to preview the character received via the UART port
 * returning TRUE stops the further processing of the character by the USART logic
 * 	might by used to intercept and process data that shouldn't be stored in the receiver ringbuffer1
 * returning FALSE allows the rest of the USART logic to execute
 */
extern fp_previewCharacter_t USART_CharacterPreview;

/**
 * Available baudrate configurations for RS232 communication
 */
enum uart_bps {
	//BPS_9600       = 0,//added by hutta
	//BPS_19200      = 1,
	BPS_115200     = 2
};

#define	UART_RX					RXBUF1
#define	UART_TX					TXBUF1
#define UART_RESET_RX()			{ U1IFG &= ~URXIFG1; }
#define	UART_RESET_RXTX()		{ U1IFG &= ~(URXIFG1 | UTXIFG1); }
#define	UART_WAIT_RX()			while( (U1IFG & URXIFG1) == 0 ) { _NOP(); }
#define	UART_WAIT_TX()			while( (U1IFG & UTXIFG1) == 0 ) { _NOP(); } 
#define UART_WAIT_TXDONE()		while( (UTCTL1 & TXEPT) == 0 ) { _NOP(); }
#define USART_MODE_IS(mode) 	(USART_CurrentMode == mode)



/**
 * Sets the speed for the UART mode. If the UART mode is active, the change will be processed immidately
 */
void USART_SetUartSpeed(enum uart_bps uartSpeed);

/**
 * Sets the low level spped registers. If the USART is in the selected mode, the change will be processed immiedately
 * It is expected that this function will be used whilie using the SPI mode.
 * Use the USART_SetUartSpeed funcition for the UART mode.
 */
void USART_SetSpeedRegisters(enum usart_mode mode, uint8_t ubr0, uint8_t ubr1, uint8_t umctl);

// int __attribute__((format (printf, 1, 2))) printf(const char *string, ...);
// int __attribute__((format (printf, 2, 3))) sprintf (char *buf, const char *fmt, ...);

bool USART_Lock(enum usart_mode mode);

bool USART_LockWait(enum usart_mode mode);

//bool USART_Unlock();

bool USART_Unlock(enum usart_mode mode);

uint8_t* USART_ReadLine();


/**
 * Initis the USART, sets the UART mode as default with the speed 115200
 */
void USART_Init();


/**
 * This member supports the ScatterWeb infrastructure
 * and is not intended to be used directly from your code.
 */ 
uint8_t SPI_rx();

/**
 * This member supports the ScatterWeb infrastructure
 * and is not intended to be used directly from your code.
 */ 
void SPI_tx(const uint8_t c);

/**
 * This member supports the ScatterWeb infrastructure
 * and is not intended to be used directly from your code.
 */ 
void SPI_read(void* pDestination, const uint16_t size, const bool incDest);

/**
 * This member supports the ScatterWeb infrastructure
 * and is not intended to be used directly from your code.
 */ 
void SPI_write(const void* pSource, const uint16_t size, const uint8_t startToken, const bool incSource);


/**
 * This member supports the ScatterWeb infrastructure
 * and is not intended to be used directly from your code.
 */ 
void SPI_idle(const uint16_t clocks);

/**
 * This member supports the ScatterWeb infrastructure
 * and is not intended to be used directly from your code.
 */ 
uint8_t SPI_waitToken(const uint8_t token, const uint16_t timeout);


#define SPI_INIT \
	st( \
		UCTL0  = SWRST; \
		UCTL0 |= CHAR + SYNC + MM; \
		UTCTL0 = CKPH + SSEL1 + SSEL0 + STC; \
		UBR00 = 0x02; \
		UBR10 = 0x00; \
		UMCTL0 = 0x00; \
		URCTL0 = 0x00; \
		ME1 |= USPIE0; \
		UCTL0  &= ~SWRST; \
	)

//----------------------------------------------------------------------------------
//  void spiInit(uint8_t clockrate)
//
//  DESCRIPTION:
//		Init UART0 for SPI mode.
//
//  ARGUMENTS:
//      uint8_t clockrate
//			Ignore clockrate parameter (no effect for now).
//
//----------------------------------------------------------------------------------
void spi0_init(uint8_t clockrate);

#endif /*USART_H_*/
