 /*
 * Copyright (c) 2008, Freie Universitt Berlin, Tomasz Naumowicz
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:
 * 
 * Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
 * Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.
 * Neither the name of the Freie Universitt Berlin nor the names of its contributors may be used to endorse or promote products derived from this software without specific prior written permission.
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */
 
#include "System.h"

/* dosfs: readme.txt

User-Supplied Functions
=======================
You must provide functions to read sectors into memory and write
them back to the target media. The demo suite includes an emulation
module that reads/writes a disk image file (#define HOSTVER pulls
in hostemu.h which wraps the prototypes for these functions).
There are various tools for UNIX, DOS, Windows et al, to create
images from storage media; my preferred utility is dd.

The functions you must supply in your embedded app are:

DFS_ReadSector(unit,buffer,sector,count)
DFS_WriteSector(unit,buffer,sector,count)

These two functions read and write, respectively, "count" sectors of
size SECTOR_SIZE (512 bytes; see below) from/to physical sector
#"sector" of device "unit", to/from the scratch buffer "buffer". They
should return 0 for success or nonzero for failure. In the current
implementation of DOSFS, count will always be 1.

The "unit" argument is designed to permit implementation of multiple
storage devices, for example multiple media slots on a single device,
or to differentiate between master and slave devices on an ATAPI bus.
		 
This code is designed for 512-byte sectors. Although the sector size
is a #define, you should not tinker with it because the vast majority
of FAT filesystems use 512-byte sectors, and the DOSFS code doesn't
support runtime determination of sector size. This will not affect the
vast majority of users.
*/

#if FATDEBUGGING
void _DFS_DisplayBlock(uint8_t* buffer) {
	uint16_t	i;

	printf("%u Bytes follow:\r\n", 512);
	for(i = 0; i < 512; ) {
		printf("%02X", (uint8_t) buffer[i]);
		i++;
		if( (i % 2) == 0 ) printf(" ");
		if( (i % 16) == 0 ) printf("\r\n");
	}	
}
#endif 

uint32_t DFS_ReadSector(uint8_t unit, uint8_t *buffer, uint32_t sector, uint32_t count) {
	/* Uses the following function from ScatterWeb.Sd.h
	 *   uint16_t Sd_read_block(void (*const pBuffer), const uint32_t address)
	 * 
	 * sector contains the number of the sector to read
	 * SECTOR_SIZE is the sector size
	 * the underlaying Sd_read_block expects absolute address of the data to read
	 * address = sector * SECTOR_SIZE 
	 */
	if (count != 1)
		return 1;
	
	uint32_t address = sector * SECTOR_SIZE;
	uint16_t ret = Sd_read_block(buffer, address);
	
	if (ret == 0) {
		//try to reset the card
		uint8_t retries = 10;
		while (retries > 0) {
						#if FATDEBUGGING
								printf("[r.retry]");
						#endif		
			USART_LockWait(USART_MODE_SPI);
			_Sd_reset();
			USART_Unlock(USART_MODE_SPI);
			
			ret = Sd_read_block(buffer, address);
			if (ret > 0)
				break;
			
			retries--;
		}
	}
	
	
	
#if FATDEBUGGING
		printf("DFS_ReadSector: reading #%lu @%lu success: %u\r\n", sector, address, ret);
		//if (sector == 737) 
		//	_DFS_DisplayBlock(buffer);
#endif
	if (ret == 0)
		return 1;
	
	return 0;
}

uint32_t DFS_WriteSector(uint8_t unit, uint8_t *buffer, uint32_t sector, uint32_t count) {
	/* Uses the following function from ScatterWeb.Sd.h
	 *   Sd_write_block(const uint32_t address, void const (*const pBuffer))
	 * 
	 * sector contains the number of the sector to read
	 * SECTOR_SIZE is the sector size
	 * the underlaying Sd_read_block expects absolute address of the data to read
	 * address = sector * SECTOR_SIZE 
	 */
	if (count != 1)
		return 1;
	
	uint32_t address = sector * SECTOR_SIZE;	

#if FATDEBUGGING
		printf("DFS_WriteSector: writing #%lu\r\n", sector);
#endif
		
		enum sd_write_ret ret = Sd_write_block(address, buffer); 
		

		if (ret != SD_WRITE_SUCCESS) {
			//try to reset the card
			uint8_t retries = 10;
			while (retries > 0) {
							#if FATDEBUGGING
									printf("[w.retry]");
							#endif		
				USART_LockWait(USART_MODE_SPI);
				_Sd_reset();
				USART_Unlock(USART_MODE_SPI);
				
				ret = Sd_write_block(address, buffer); 
				if (ret == SD_WRITE_SUCCESS)
					break;
				
				retries--;
			}
		}

#if FATDEBUGGING
		//if (sector == 737) {
		//	_DFS_DisplayBlock(buffer);
		//}
#endif

		return ret;
}

void DSF_ToFatDate(uint16_t* date, uint16_t year, uint8_t month, uint8_t day) {
	*date = (year - 1980) & 0x7F;
	*date <<= 4;
	*date += month & 0x0F;
	*date <<= 5;
	*date += day & 0x1F;	
}

void DSF_ToFatTime(uint16_t* time, uint8_t hours, uint8_t minutes, uint8_t seconds) {
	*time = hours & 0x1F;
	*time <<= 6;
	*time += minutes & 0x3F;
	*time <<= 5;
	*time += (seconds >> 1) & 0x1F;
}

void DSF_FromFatDate(uint16_t date, uint16_t* year, uint8_t* month, uint8_t* day) {
	*day = date & 0x001F;	
	
	date >>= 5;
	*month = date & 0x000F;
	
	date >>= 4;
	*year = (date & 0x007F) + 1980;
}

void DSF_FromFatTime(uint16_t time, uint8_t* hours, uint8_t* minutes, uint8_t* seconds) {
	*seconds = (time & 0x001F) << 1;

	time >>= 5;
	*minutes = time & 0x003F;

	time >>= 6;
	*hours = time & 0x001F;
}



void DFS_UpdateTimestamp(PDIRENT dirent, bool creation, bool access, bool modification) {
	uint16_t fat_date, fat_time;
	
	time_t now;
	RTC_GetTime(&now);
	
	DSF_ToFatDate(&fat_date, now.year + 1970, now.month, now.day);
	DSF_ToFatTime(&fat_time, now.hour, now.minute, now.second);
	
	if (creation) {
		*((uint16_t*) &(dirent->crtdate_l)) = fat_date;
		*((uint16_t*) &(dirent->crttime_l)) = fat_time;
	}
	
	if (access) {
		*((uint16_t*) &(dirent->lstaccdate_l)) = fat_date;
	}
	
	if (modification) {
		*((uint16_t*) &(dirent->wrtdate_l)) = fat_date;
		*((uint16_t*) &(dirent->wrttime_l)) = fat_time;
	}
}


enum dfs_init_result DFS_Init() {
	uint32_t pstart, psize;
	uint8_t pactive, ptype;
	
	// Obtain pointer to first partition on first (only) unit
	pstart = DFS_GetPtnStart(0, _fat_buffer, 0, &pactive, &ptype, &psize);
	if (pstart == 0xffffffff) {
		#if (LOGLEVEL & LOGLEVEL_VERBOSE)
			printf(" FAT.ERR: Cannot find first partition\n");
		#endif
		DFS_Init_Result = FAT_ERROR_PARTITION;
		
		return FAT_ERROR_PARTITION;
	}

	//printf("Partition 0 start sector 0x%lX active 0x%X type 0x%X size 0x%lX\r\n", pstart, pactive, ptype, psize);

	// well sometimes the sector 0 contains the partition table already, so the data read vi DFS_GetPtnStart is useless,
	// since some 'random' values are returned. there must be a way to determine, which system is used.
	// it should be included into the DFS_GetPtnStart --> the function ASSUMES there is an MBR
	// for now, a simple workaround:
			if (pstart > 4096)
				pstart = 0;
	
	LOG("NO MBR detected, using workaround.");
	if (DFS_GetVolInfo(0, _fat_buffer, pstart, &DFS_VolumeInfo)) {
		#if (LOGLEVEL & LOGLEVEL_VERBOSE)
			printf(" FAT.ERR: Error getting volume information\n");
		#endif
		DFS_Init_Result = FAT_ERROR_VOLUME;
		return FAT_ERROR_VOLUME;
	}

	#if (LOGLEVEL & LOGLEVEL_VERBOSE) 
		printf(" FAT: Label '%-11.11s'\r\n", DFS_VolumeInfo.label);
		printf(" FAT: %d sectors per cluster, %d reserved , volume total %lu\r\n", DFS_VolumeInfo.secperclus, DFS_VolumeInfo.reservedsecs, DFS_VolumeInfo.numsecs);
		printf(" FAT: %lu sectors per FAT, first FAT at sector #%lu, root dir at #%lu.\r\n",DFS_VolumeInfo.secperfat,DFS_VolumeInfo.fat1,DFS_VolumeInfo.rootdir);
		printf(" FAT: %u root dir entries, data area starts at sector #%lu.\r\n",DFS_VolumeInfo.rootentries,DFS_VolumeInfo.dataarea);
		printf(" FAT: %lu clusters (%lu bytes) in data area.\r\n", DFS_VolumeInfo.numclusters, DFS_VolumeInfo.numclusters * DFS_VolumeInfo.secperclus * SECTOR_SIZE);
		printf(" FAT: ");
		if (DFS_VolumeInfo.filesystem == FAT12)
			printf("FAT12.\r\n");
		else if (DFS_VolumeInfo.filesystem == FAT16)
			printf("FAT16.\r\n");
		else if (DFS_VolumeInfo.filesystem == FAT32)
			printf("FAT32.\r\n");
		else
			printf("[unknown]\n");
	#endif

	if (DFS_VolumeInfo.filesystem != FAT16) {
		DFS_Init_Result = FAT_ERROR_TYPE;
		return FAT_ERROR_TYPE;
	}
	
	DFS_Init_Result = FAT_ERROR_NONE;

	return FAT_ERROR_NONE;
}

#if DFS_DIR_COMMAND
COMMAND(dir, COMMAND_SERIAL, cmd_args) {
	printf("[dir:start] (%s) (Name) (Size) (Type)\r\n", cmd_args->arguments);
	
	if (SD_Card_Init_Result != SD_INIT_SUCCESS || DFS_Init_Result != FAT_ERROR_NONE) {
		printf("[dir:end] error\r\n");
		return;
	}

	DIRINFO _currentDirectory;
	_currentDirectory.scratch = _fat_buffer;
	
	if (DFS_OpenDir(&DFS_VolumeInfo, cmd_args->arguments, &_currentDirectory) > 0) {
		printf("[dir:end] error\r\n");
		return;
	}
	
	DIRENT		_directoryEntry;
	uint8_t		count = 0;
	
	while (!DFS_GetNext(&DFS_VolumeInfo, &_currentDirectory, &_directoryEntry)) {
		if (_directoryEntry.name[0]) {
			if (_directoryEntry.attr & ATTR_DIRECTORY)
				printf("[dir:cont] (%-11.11s) (0) (dir)\r\n",
								_directoryEntry.name);
			else
				printf("[dir:cont] (%-11.11s) (%lu) (file)\r\n",
								_directoryEntry.name, *((uint32_t*) &(_directoryEntry.filesize_0)));		
			
			count++;
		}
	}
	
	printf("[dir:end] (%u) entries\r\n", count);
}
#endif

#if DFS_READ_COMMAND
COMMAND(read, COMMAND_SERIAL, cmd_args) {
	
	//parse input, navigate to the target directory
	// read \test\test2\a.txt 1000
	// first parameter is the path:
	//		a.txt - only as filename
	//		\dir\a.txt - or with the path
	//		\a.txt - root directory
	

	
	//pointer to the path:
	//skip the command and space
	char* pchar;
	
	char* path = NULL;
	uint32_t readfrom, readto;
	readfrom = 0;
	readto = 0;
	
	//tokenize the command --> should be moved to general 'command' engine
	pchar = strtok(cmd_args->arguments, " ");
	uint8_t parameterIndex = 1;
	while (pchar != NULL) {
		switch (parameterIndex) {
			case 1: 
				//copy the path
				path = pchar;
				break;
			case 2:
				//prosess 'from byte'
				readfrom = strtoul(pchar, NULL, 10);
				break;
			case 3:
				//process 'to byte'
				readto = strtoul(pchar, NULL, 10);
				break;
		}
		
		pchar = strtok(NULL, " ");
		parameterIndex++;
	}
	
	printf("[read:start] path(%s)from(%lu)to(%lu)\r\n", path, readfrom, readto);
	
	if (SD_Card_Init_Result != SD_INIT_SUCCESS || DFS_Init_Result != FAT_ERROR_NONE) {
		printf("[read:end] error\r\n");
		return;
	}

	

	FILEINFO fileInfo;
	if (DFS_OpenFile(&DFS_VolumeInfo, path, DFS_READ, _fat_buffer, &fileInfo)) {
		printf("[read:end] error\r\n");
		return;
	}
	
	if (fileInfo.filelen < readfrom || fileInfo.filelen < readto) {
		printf("[read:end] error: crossed EOF\r\n");
		return;
	}
	
	if (readfrom > 0) {
		DFS_Seek(&fileInfo, readfrom, _fat_buffer);
	}
	
	uint8_t buffer[32];
	
	uint32_t bytesRead, bytesReadTotal, bytesLeft;
	uint16_t bytesToRead;
	uint16_t tempIndex; //used for printing data
	
	if (readto == 0)
		readto = fileInfo.filelen;
	
	bytesLeft = readto - readfrom; //bytes left to read
	bytesReadTotal = 0;

	do {
		if (bytesLeft >= 32)
			bytesToRead = 32;
		else
			bytesToRead = (uint16_t) bytesLeft;

		DFS_ReadFile(&fileInfo, _fat_buffer, buffer, &bytesRead, bytesToRead);

		//print the content of from the buffer
		for (tempIndex = 0; tempIndex < bytesRead; tempIndex++) {
			putchar(buffer[tempIndex]);
		}

		bytesLeft = bytesLeft - (uint32_t) bytesRead;

		bytesReadTotal += bytesRead;
		
	} while (bytesLeft > 0);	
	
	printf("\r\n[read:end] read(%lu)bytes|from pos(%lu) to pos(%lu)|filesize(%lu)\r\n", bytesReadTotal, readfrom, readto, fileInfo.filelen);
	
	
}
#endif

#if DFS_APPEND_COMMAND
COMMAND(append, COMMAND_SERIAL, cmd_args) {
	FILEINFO fileInfo;

	if (SD_Card_Init_Result != SD_INIT_SUCCESS || DFS_Init_Result != FAT_ERROR_NONE) {
		snprintf(cmd_args->response, cmd_args->response_length, "error");
		return;
	}


	
	char* pchar;
	
	char* path = NULL;
	char* content = NULL;
	
	//tokenize the command --> should be moved to general 'command' engine
	pchar = strtok(cmd_args->arguments, " ");
	//uint8_t parameterIndex = 1;
	if (pchar == NULL)
		return;
	path = pchar;
	
	pchar = strtok(NULL, "\r");

	if (pchar == NULL)
		return;

	content = pchar;

	//printf("path: %s\r\ncontent: %s\r\n", path, content);

	
	if (DFS_OpenFile(&DFS_VolumeInfo, path, DFS_WRITE, _fat_buffer, &fileInfo)) {
		snprintf(cmd_args->response, cmd_args->response_length, "error openning file.");
		return;
	}
	
	DFS_Seek(&fileInfo, fileInfo.filelen, _fat_buffer);
	uint32_t bytesWritten;
	DFS_WriteFile(&fileInfo, _fat_buffer, content, &bytesWritten, strlen(content));
	//snprintf(cmd_args->response, cmd_args->response_length, "(%lu) bytes written (new file size %lu)\r\n", bytesWritten, fileInfo.filelen);
	DFS_WriteFile(&fileInfo, _fat_buffer, "\r\n", &bytesWritten, 2);
	//snprintf(cmd_args->response, cmd_args->response_length, "(new line) (%lu) bytes written (new file size %lu)\r\n", bytesWritten, fileInfo.filelen);
}
#endif

#if DFS_FILEINFO_COMMAND
COMMAND(fileinfo, COMMAND_SERIAL, cmd_args) {
	if (SD_Card_Init_Result != SD_INIT_SUCCESS || DFS_Init_Result != FAT_ERROR_NONE) {
		printf("[fileinfo] error\r\n");
		return;
	}	
	
	FILEINFO fileInfo;
	if (DFS_OpenFile(&DFS_VolumeInfo, cmd_args->arguments, DFS_READ, _fat_buffer, &fileInfo)) {
		snprintf(cmd_args->response, cmd_args->response_length, "error openning file.");
		return;
	}
	
	//fileInfo contains data about the file
	printf("Filename: %s\r\n"
			" Size: %lu\r\n"
			" First Cluster: %lu\r\n", 
			cmd_args->arguments, 
			fileInfo.filelen,
			fileInfo.firstcluster);
	
	if (!DFS_ReadSector(fileInfo.volinfo->unit, _fat_buffer, fileInfo.dirsector, 1)) {
		PDIRENT de = &(((PDIRENT) _fat_buffer)[fileInfo.diroffset]);
		
		uint16_t fat_time;
		uint16_t fat_date;
		uint16_t year;
		uint8_t month, day, hour, minute, second;
		
		// creation:
		fat_time = *((uint16_t*) &(de->crttime_l));
		fat_date = *((uint16_t*) &(de->crtdate_l));
		
		DSF_FromFatDate(fat_date, &year, &month, &day);
		DSF_FromFatTime(fat_time, &hour, &minute, &second);

		printf(" Created: %.2u:%.2u:%.2u %.2u-%.2u-%.4u\r\n",
				hour, minute, second, day, month, year);
		
		
		// modification:
		fat_time = *((uint16_t*) &(de->wrttime_l));
		fat_date = *((uint16_t*) &(de->wrtdate_l));
		
		DSF_FromFatDate(fat_date, &year, &month, &day);
		DSF_FromFatTime(fat_time, &hour, &minute, &second);

		printf(" Modified: %.2u:%.2u:%.2u %.2u-%.2u-%.4u\r\n",
				hour, minute, second, day, month, year);
	}
}

#endif


